(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Omega.

Require Import rel_utils.
Require Import list_utils.
Require Import good_base.

Set Implicit Arguments.

(* 

   This file contains the basic "prop-bounded" definitions of the "Almost Full"
   AF predicate as defined in "Stop when your almost full" by Vytiniotis, Coquand
   and Wahlstedt.

   Some proofs and results of this file are largely inspired by that work
   and its "Accompanying development" which is Coq code made available by
   the autors. 

   In my opinion, the really non-trivial part of what I borrowed from those
   authors is the *very nice* proof of Ramsey's theorem (originally by Coquand), 
   i.e. stability of the AF predicate by intersection (or isomorphically 
   by products).

   From my point of view, a quite trivial but however VERY usefull addition
   to that code is the af_relmap theorem which says that the AF is stable
   under surjective *relational* maps. Hence, it works very well for subtypes
   { x | P x } which is not really the case with  af_cofmap, af_fmap or
   af_fmap_gen. 

   Another addition is the dependently typed induction principle for AF
   named af_dep_ind where the ground case is NOT when the relation is
   full BUT when the type is EMPTY (a particular case of full relation).
   That concept plays a crucial role in the proof of both Higman's theorem
   and Kruskal's theorem.

   There is also a development for lexicographic products but there is
   more powerfull stuff in afs_lex.

*)

Section Almost_Full.

  Variables (X : Type).

  Implicit Type R S : X -> X -> Prop.

  Inductive af : (X -> X -> Prop) -> Prop := 
    | in_af_0 : forall R, full R -> af R
    | in_af_1 : forall R, (forall a, af (lift_rel R a)) -> af R.

  Fact af_inc R S : R inc2 S -> af R -> af S.
  Proof.
    intros H1 H; revert S H1; induction H as [ R HR | R HR IH ]; intros S H1.
    constructor 1; auto.
    constructor 2; auto.
    intro x; apply (IH x).
    intros y z [|]; [ left | right ]; auto.
  Defined.

  Fact af_eq R S : R ~eq2 S -> (af R <-> af S).
  Proof.
    intros []; split; apply af_inc; auto.
  Qed.

  Fact sec_binary_infinite_chain R f : af R -> forall k, exists i j, k <= i < j /\ R (f i) (f j).
  Proof.
    induction 1 as [ R' HR' | R' HR' IH ]; intros k.
    exists k, (S k); split; auto; apply HR.
    destruct (IH (f k) (S k)) as ( i & j & Hij & [ H | H ] ); 
      [ exists i, j | exists k, i ]; split; auto; omega.
  Qed.

  Corollary af_inf_chain R f : af R -> exists i j, i < j /\ R (f i) (f j).
  Proof.
    intros H.
    destruct (sec_binary_infinite_chain f H 0) as ( i & j & ( _ & ? ) & ? ).
    exists i, j; auto.
  Qed.

  Corollary af_includes_eq R : af R -> @eq _ inc2 R.
  Proof.
    intros H1 x y ?; subst.
    destruct (af_inf_chain (fun _ => y) H1) as ( _ & _ & _ & ?); auto.
  Qed.    

  Definition dec_rel R := forall x y, R x y \/ ~ R x y.

  Section af_from_wf.

    Variable (R : X -> X -> Prop) (decR : dec_rel R).

    Fact af_iter z : Acc R z -> af (fun x y => ~ R x z \/ ~ R y x).
    Proof.
      induction 1 as [ z Hz IH ].
      constructor 2.
      intros u.
      destruct (decR u z) as [ Hzu | Hzu ].
      generalize (IH _ Hzu). 
      apply af_inc.
      intros ? ?; red; tauto.
      constructor 1; intros; red; tauto.
    Defined.

    Fact af_from_wf : well_founded R -> af (fun x y => ~ R y x).
    Proof.
      intros HR.
      constructor 2.
      intros x.
      generalize (af_iter (HR x)).
      apply af_inc.
      intros ? ?; red; tauto.
    Defined.
      
  End af_from_wf.

  Section wf_from_af.
  
    Variable (R T : X -> X -> Prop).

    Lemma acc_from_af y : 
       (forall x z,  clos_refl_trans _ T z y
                  -> clos_trans_1n _ T x z 
                  -> R z x 
                  -> False)
       -> af R -> Acc T y.
    Proof.
      intros H1 H2; revert y H1.
      induction H2 as [ R HR | R HR IH ]; intros y H.
      
      constructor; intros x Hx.
      exfalso; apply (@H x y); auto.
      constructor 2.
      constructor 1; auto.

      constructor; intros z Hz.
      
      apply IH with (a := y).
      intros x' z' H1 H2 [ H3 | H3 ].
      apply H with (3 := H3); auto.
      constructor 3 with z; auto.
      constructor 1; auto.
      apply H with (3 := H3); auto.
      constructor 2.
      rewrite <- clos_trans_t1n_iff, -> clos_trans_tn1_iff.
      rewrite clos_trans_n1_clos_rt.
      exists z; auto.
    Defined.

    Hypothesis HT : forall x y, clos_trans_1n _ T x y -> R y x -> False.
  
    Lemma wf_from_af : af R -> well_founded T.
    Proof.      
      intros ? ?.
      apply acc_from_af; auto.
      intros ? ? ?; apply HT.
    Defined.
      
  End wf_from_af.

  Section wf_from_wqo.
 
    Variable (R : X -> X -> Prop).

    Theorem strict_transitive : transitive _ R -> transitive _ (fun x y => R x y /\ ~ R y x).
    Proof.
      intros H x y z (H1 & H2) (H3 & H4); split.
      apply H with (1 := H1); auto.
      contradict H4; apply H with (1 := H4); auto.
    Qed.
  
    Definition wqo := transitive _ R /\ af R.

    Theorem wf_from_wqo : wqo -> well_founded (fun x y => R x y /\ ~ R y x).
    Proof.
      intros [ ? ? ].
      apply wf_from_af with (R := R); auto.
      intros ? ?; apply trans_strict_prop; auto.
    Qed.

  End wf_from_wqo.

End Almost_Full.


(* a bar like induction principle for af
   this allows to reason by induction on lists with
   a fixed relation that is lifted at the induction step

 *)

Section af_lrlift_ind.

  Variable (X : Type) (R : X -> X -> Prop) (Q : list X -> Prop).

  Hypothesis H1 : good R inc1 Q.
  Hypothesis H2 : forall ll, (forall a, af (R lrlift (a::ll))) -> (forall a, Q (a::ll)) -> Q ll.
  
  Let af_li_rec (S : X -> X -> Prop) ll : S inc2 R lrlift ll -> af S -> Q ll.
  Proof.
    intros G1 G2; revert G2 ll G1.
    induction 1 as [ S HS | S HS IHS ]; intros ll Hll.

    assert (af (R lrlift ll)) as H3.
      constructor 1; intros; auto.
    apply H2.
    intros a; revert H3; apply af_inc; simpl; unfold lift_rel; tauto.
    intros a.
    apply H2.
    intros b; revert H3; apply af_inc; simpl; unfold lift_rel; tauto.
    intros b.
    apply H1.
    specialize (HS a b).
    apply Hll in HS.
    apply lift_rel_list_ctxt in HS.
    red in HS.
    destruct HS as [ HS | [ HS | HS ] ].
    do 2 constructor 2; auto.
    destruct HS as ( c & H4 & H5).
    constructor 2; constructor 1 with c; auto.
    constructor 1 with a; simpl; auto.
    
    apply H2.
    intros a.
    generalize (HS a); apply af_inc.
    simpl lift_rel_list; unfold lift_rel.
    intros x y [ | ]; [ left | right]; auto.

    intros a.
    apply (IHS a).
    simpl lift_rel_list; unfold lift_rel.
    intros x y [ | ]; [ left | right]; auto.
  Qed.

  Fact af_lrlift_ind ll : af (R lrlift ll) -> Q ll.
  Proof.
    apply af_li_rec; auto.
  Qed.

End af_lrlift_ind.

Fact af_True X : @af X (fun _ _ => True).
Proof.
  constructor 1; auto.
Qed.

Lemma af_bool : af (@eq bool).
Proof.
  constructor 2.
  intros [ | ]; constructor 2.
  intros [ | ]. constructor 1; firstorder.
  constructor 2.
  intros [ | ].
  constructor 1; firstorder.
  constructor 1; firstorder.
  constructor 2; intros [ | ].    
  constructor 2; intros [ | ].
  constructor 1; firstorder.
  constructor 1; firstorder.
  constructor 1; firstorder.
Qed.

Lemma leq_af : af le.
Proof.
  generalize (af_from_wf dec_lt lt_wf).
  apply af_inc.
  intros ? ?; abstract omega.
Defined.

Section af_induction.

  Variables (X : Type)
            (T R : X -> X -> Prop) 
            (HR : af R) 
            (HTR : forall x y, clos_trans_1n _ T x y -> R y x -> False) 
            (P : X -> Type) (HP : forall x, (forall y, T y x -> P y) -> P x).

  Theorem af_induction_type: forall a, P a.
  Proof. 
    apply well_founded_induction_type with (R := T); auto.
    apply wf_from_af with (R := R); auto.
  Defined.

End af_induction.

Section af_union.

   Variables (X : Type) (R S : X -> X -> Prop).

   Fact af_union : af R -> af (fun x y => R x y \/ S x y).
   Proof.
     apply af_inc.
     intros ? ?; auto.
   Qed.

End af_union.

Section af_cofmap.

  Variables (X Y : Type) (f : Y -> X) (R : X -> X -> Prop).
  
  Corollary af_cofmap : af R -> af (fun x y => R (f x) (f y)).
  Proof.
    induction 1 as [ R HR | R HR IH ].
    constructor 1; firstorder.
    constructor 2; intros; apply IH.
  Qed.

End af_cofmap.

(* the most general projection principle with a surjective relation from R to S *)

Section af_relmap.

  Variables (X Y : Type) (f : X -> Y -> Prop) 
            (R : X -> X -> Prop) (S : Y -> Y -> Prop)
            (Hf : forall y, exists x, f x y)
            (HRS : forall x1 x2 y1 y2, f x1 y1 -> f x2 y2 -> R x1 x2 -> S y1 y2).

  Theorem af_relmap : af R -> af S.
  Proof.
    intros H; revert S HRS.
    induction H as [ R HR | R HR IH ]; intros S H.

    constructor 1.
    intros y1 y2.
    destruct (Hf y1) as (x1 & Hx1).
    destruct (Hf y2) as (x2 & Hx2).
    apply H with (3 := HR x1 x2); auto.
    
    constructor 2; intros y0.
    destruct (Hf y0) as (x0 & Hx0).
    apply IH with (a := x0).
    intros x1 x2 y1 y2 H1 H2.
    generalize (H _ _ _ _ H1 H2).
    unfold lift_rel.
    firstorder.
  Qed.
    
End af_relmap.

(* two instances where the relation is a function *)

Section af_fmap_gen.

  Variables (X Y : Type) (f : X -> Y) (Hf : forall y, exists x, y = f x).

  Implicit Types (R : X -> X -> Prop) (S : Y -> Y -> Prop).  

  Fact af_fmap_gen R S : (forall x y, R x y -> S (f x) (f y)) -> af R -> af S.
  Proof.
    intros H.
    apply af_relmap with (f := fun x y => y = f x); auto.
    intros; subst; auto.
  Qed.

  Definition rel_fmap R (y1 y2 : Y) := exists x1 x2, y1 = f x1 /\ y2 = f x2 /\ R x1 x2.
  
  Fact af_fmap R : af R -> af (rel_fmap R).
  Proof.
    apply af_fmap_gen; auto.
    intros x1 x2; intros; subst; exists x1, x2; auto.
  Qed.
 
End af_fmap_gen.

Section af_intersection.

  (* Borrowed from "Stop when you are almost full" and the corresponding source code
     called "Accompanying development" *)

  Variable X : Type.

  Lemma oplus_nullary (A B : Prop) (R : X -> X -> Prop) : 
    af R -> forall C, (forall x y, R x y <-> C x y \/ A) 
                  -> af (fun x y => C x y \/ B) 
                  -> af (fun x y => C x y \/ A /\ B).
  Proof.
    induction 1 as [ R HR | R HR IH ]; intros C H1 H2.

    apply af_inc with (2 := H2).
    intros x y [ Hxy | Hxy ]; auto.
    specialize (HR x y).
    rewrite H1 in HR.
    tauto.
    
    constructor 2.
    intros x.
    apply af_inc with (R := fun y z => (C y z \/ C x y) \/ A /\ B).
    intros ? ?; unfold lift_rel; tauto.
    apply IH with (a := x).
    intros ? ?; unfold lift_rel; do 2 rewrite H1; tauto.
    apply af_inc with (2 := H2).
    intros ? ?; tauto.
  Qed.

  Lemma oplus_nullary_cor (A B : Prop) (C : X -> X -> Prop) :  
    af (fun x y => C x y \/ A) -> 
    af (fun x y => C x y \/ B) -> 
    af (fun x y => C x y \/ (A /\ B)).
  Proof.
    intros.
    apply oplus_nullary with (R := fun x y => C x y \/ A); auto.
    split; auto.
  Qed.

  Lemma oplus_unary (A B : X -> Prop):
         forall R, af R 
      -> forall T, af T 
      -> forall C, (forall x y, R x y -> C x y \/ A x) 
                -> (forall x y, T x y -> C x y \/ B x) 
      -> af (fun x y => C x y \/ (A x /\ B x)).
  Proof.
    induction 1 as [ R HR | R HR IHR ].

    intros T HT C Req Teq.
    eapply af_inc with (2 := HT).
    intros x y.
    generalize (Req x y) (Teq x y) (HR x y); tauto.
    
    induction 1 as [ T HT | T HT IHT ].

    intros C Req Teq.
    apply af_inc with (R := R).
    intros x y.
    generalize (HT x y) (Req x y) (Teq x y); tauto.
    constructor 2 with (1 := HR).
  
    intros C Req Teq.

    constructor 2; intros x.
    generalize (in_af_1 HT); intros HT'.
    apply af_inc with (R := fun y z : X => ((C y z \/ A y /\ B y) \/ C x y) \/ A x /\ B x).
    intros ? ?; unfold lift_rel; tauto.
    apply oplus_nullary_cor.
    generalize (IHR x _ HT' (fun y z => C y z \/ C x y \/ A x)); intros G0.
    eapply af_inc.
    Focus 2.
    apply G0.
    intros y z; generalize (Req y z) (Req x y); unfold lift_rel; tauto.
    intros y z Hyz; generalize (Teq y z); tauto.
    intros ? ?; simpl; tauto.
    generalize (IHT x (fun y z => C y z \/ C x y \/ B x)); intros G0.
    eapply af_inc.
    Focus 2.
    apply G0.
    intros y z; generalize (Req y z) (Req x y); tauto.
    intros y z [ H1 | H1 ]; apply Teq in H1; tauto.
    intros ? ?; simpl; tauto.
  Qed.
  
  Lemma oplus_unary_cor (A B : X -> Prop) (C : X -> X -> Prop):
            af (fun x y => C x y \/ A x) 
         -> af(fun x y => C x y \/ B x) 
         -> af (fun x y => C x y \/ (A x /\ B x)).

  Proof.
    intros.
    apply oplus_unary with (R := fun x y => C x y \/ A x) 
                           (T := fun x y => C x y \/ B x); auto.
  Qed.

  Theorem af_intersection (A B : X -> X -> Prop):
        af A 
     -> af B 
     -> af (fun x y => A x y /\ B x y).
 
  Proof.
    intros HA HB.
    revert A HA B HB.
    induction 1 as [ A HA | A HA IHA ].
    
    intros B HB.
    apply af_inc with (2 := HB).
    intros x y; specialize (HA x y); tauto.
    
    induction 1 as [ B HB | B HB IHB ].

    generalize (in_af_1 HA); intros HA'.
    apply af_inc with (2 := HA').
    intros ? ?; firstorder.
    
    constructor 2; intros x.
    apply oplus_unary_cor.
    generalize (in_af_1 HB); intros HB'.
    apply af_inc with (2 := IHA x _ HB').
    intros ? ?; unfold lift_rel; tauto.
    apply af_inc with (2 := IHB x).
    intros ? ?; unfold lift_rel; tauto.
  Qed.

End af_intersection.

Section af_product.
  
  Variable (X Y : Type).
  Variable (R : X -> X -> Prop) (S : Y -> Y -> Prop).
  
  Theorem af_product : af R
                    -> af S
                    -> af (rel_prod R S).
  Proof.
    intros HR HS.
    apply af_intersection; apply af_cofmap; auto.
  Qed.

End af_product.

Section af_sum.

  Variable (X Y : Type). 

  Lemma af_sum_left R : af R -> af (@sum_left X Y R).
  Proof.
    induction 1 as [ R HR | R HR IH ] using af_ind.
    
    apply in_af_1; intros x.
    apply in_af_1; intros y.
    apply in_af_0.
    intros [|] [|]; destruct x; destruct y; simpl; firstorder.
    
    apply in_af_1; intros [ x | x ].
    apply af_inc with (2 := IH x).
    intros [|] [|]; unfold lift_rel, sum_left; tauto.
    apply in_af_1; intros [ y | y ].
    apply af_inc with (2 := IH y).
    intros [|] [|]; unfold lift_rel, sum_left; tauto.
    apply in_af_0.
    intros [|] [|]; unfold lift_rel, sum_left; tauto.
  Defined.

  Lemma af_sum_right R : af R -> af (@sum_right X Y R).
  Proof.
    induction 1 as [ R HR | R HR IH ] using af_ind.
    
    apply in_af_1; intros x.
    apply in_af_1; intros y.
    apply in_af_0.
    intros [|] [|]; destruct x; destruct y; simpl; firstorder.
    
    apply in_af_1; intros [ x | x ].
    apply in_af_1; intros [ y | y ].
    apply in_af_0.
    intros [|] [|]; unfold lift_rel, sum_right; tauto.
    apply af_inc with (2 := IH y).
    intros [|] [|]; unfold lift_rel, sum_right; tauto.
    apply af_inc with (2 := IH x).
    intros [|] [|]; unfold lift_rel, sum_right; tauto.
  Defined.
  
  Lemma af_sum_both R S : af R -> af S -> af (@sum_both X Y R S).
  Proof.
    intros HR HS.
    apply af_inc with (2 := af_intersection (af_sum_left HR) (af_sum_right HS)).
    intros [|] [|] (H1 & H2); simpl in H1, H2 |- *; tauto.
  Defined.

End af_sum.

Section af_lsum.

  Variable (X Y : Type).

  Definition rel_lsum (ll : list X) : (forall x, In_t x ll -> relation Y) -> relation { x : _ & (In_t x ll * Y)%type }.
  Proof.
    induction ll as [ | x ll IH ]; intros H.
    intros ? ?; apply False.
    intros (a & [ Ha | Ha ] & y1) (b & [ Hb | Hb ] & y2).
    subst x.
    apply (H _ (inl eq_refl) y1 y2).
    apply False.
    apply False.
    apply (IH (fun u Hu => H _ (inr Hu)) (existT _ a (Ha,y1)) (existT _ b (Hb,y2))).
  Defined.
  
  Fact af_rel_lsum ll R : (forall x Hx, af (R x Hx)) -> af (rel_lsum ll R).
  Proof.
    revert R.
    induction ll as [ | x ll IH ]; intros R HR.
    constructor 1; intros (? & [] & _).
    set (S := fun u Hu => R u (inr Hu)).
    assert (HS : forall u Hu, af (S u Hu)).
      intros; apply HR.
    set (S0 := R x (inl eq_refl)).
    assert (HS0 : af S0).
      apply HR.    
    generalize (af_sum_both HS0 (IH S HS)).
    set (f (c : Y + { x : _ & (In_t x ll * Y)%type })  
               := match c with 
               | inl y                 => existT _ x (inl eq_refl,y)
               | inr (existT x (px,y)) => existT _ x (inr px,y)
               end : { a : _ & (In_t a (x::ll) * Y)%type }).
    apply af_fmap_gen with f.

    intros (a & [ Ha | Ha ] & y).
    exists (inl y); simpl; subst; auto.
    exists (inr (existT _ a (Ha, y))); simpl; auto.
    
    intros [ y1 | (a & Ha & y1) ] [ y2 | (b & Hb & y2) ]; simpl; auto.
  Qed.

End af_lsum.

Section af_lsum'.
  
  Variable Y : Type.
  
  Fact af_rel_fsum n R : (forall k Hk, af (R k Hk)) -> af (@rel_fsum Y n R).
  Proof.
    revert R.
    induction n as [ | n IH ]; intros R HR.
    constructor 1; simpl.
    intros ((? & ?) & ?); exfalso; omega.

    simpl in R, HR.

    set (S0 := R 0 (lt_0_Sn n)).
    set (Sn k Hk := R _ (lt_n_S k n Hk)).
    assert (HS0 : af S0). 
      apply HR. 
    assert (HSn : forall k Hk, af (Sn k Hk)). 
      intros; apply HR.
      
    generalize (af_sum_both HS0 (IH Sn HSn)).
    
    set (f (c : Y + { k | k < n } * Y) (d : { k | k < S n } * Y) :=
      match c, d with
        | inl y1              , (exist 0 _,y2)     => y1 = y2    
        | inr (exist k _,y1)  , (exist (S l) _,y2) => k = l /\ y1 = y2
        | _                   , _                  => False
      end).   
    apply af_relmap with f.
    
    intros (([ | a] & Ha) & y).
    exists (inl y); simpl; auto.
    exists (inr (exist _ a (lt_S_n _ _ Ha), y)); simpl; auto.
    
    intros [ y1 | ((k1 & Hk1) & y1) ] 
           [ y2 | ((k2 & Hk2) & y2) ]
           (([|k3] & Hk3) & y3) 
           (([|k4] & Hk4) & y4); simpl; try tauto.
    intros ? ?; subst; auto.
    unfold S0; rewrite (lt_pirr _ Hk3); auto.
    intros (? & ?) (? & ?); subst k3 k4 y3 y4.
    destruct (eq_nat_dec k1 k2); simpl.
    unfold Sn; rewrite (lt_pirr _ Hk3); auto.
    auto.
  Qed.

End af_lsum'.

Section wqo_inter.

  Variables (X : Type) (R S : X -> X -> Prop) (HR : wqo R) (HS : wqo S).
  
  Theorem wqo_intersection : wqo (fun x y => R x y /\ S x y).
  Proof.
    destruct HR as [ H1 H2 ].
    destruct HS as [ H3 H4 ].
    split.
    intros x y z.
    generalize (H1 x y z) (H3 x y z); tauto.
    apply af_intersection; auto.
  Qed.
  
End wqo_inter.

Section rel_list_prod.

  Variable (X Y : Type).
  
  Implicit Type R : X -> Y -> Prop.

  Inductive rlp : list (X -> Y -> Prop) -> list X -> list Y -> Prop :=
    | rlp_nil  : rlp nil nil nil
    | rlp_cons : forall R lR x ll y mm, R x y -> rlp lR ll mm -> rlp (R::lR) (x::ll) (y::mm).
    
End rel_list_prod.

Require Import Eqdep_dec.

Fact UIP_nat (n m : nat) (H1 H2 : n = m) : H1 = H2.
Proof.
  apply UIP_dec, eq_nat_dec.
Qed.

Section af_rlp.

  Variable X : Type.

  Let P U n (l : list U) := length l = n.

  Definition rlp_n lR (l m : sig (P (length lR))) := @rlp X X lR (proj1_sig l) (proj1_sig m).

  Let S_inj (n m : nat) : S n = S m -> n = m.
  Proof.
    intros H.
    apply f_equal with (f := pred) in H.
    apply H.
  Defined.

  Let f n (c : X * sig (@P X n)) : sig (@P X (S n)).
  Proof.
    refine(
    match c with
      | (x,exist l H) => exist _ (x::l) _
    end).
    red in H |- *.
    subst n; reflexivity.
  Defined.

  Let f_surj n l : exists c, l = @f n c.
  Proof.
    destruct l as [ [ | x l ] Hl ]; red in Hl.
    discriminate Hl.
    simpl in Hl.
    exists (x,exist _ l (S_inj Hl)).
    simpl; f_equal.
    injection Hl; intros; subst.
    apply UIP_nat.
  Qed.

  Fact af_rlp lR : Forall (@af _) lR -> af (rlp_n lR).
  Proof.
    induction 1 as [ | R lR H1 H2 IH ].

    constructor 1.
    unfold P;
    intros (x & Hx) (y & Hy).
    destruct x; try discriminate Hx.
    destruct y; try discriminate Hy.
    constructor 1.

    generalize (af_fmap _ (@f_surj (length lR)) (af_product H1 IH)).
    apply af_inc.
    intros [ [ | x l ] Hl ] [ [ | y m ] Hm ]; try discriminate Hl; try discriminate Hm.
    simpl in Hl, Hm.
    unfold rel_fmap.
    intros ( (u & ll & Hll) & (v & mm & Hmm) & H3 & H4 & H5 & H6 ).
    simpl in H3, H4, H5, H6 |- *.
    injection H3.
    injection H4.
    clear H3 H4.
    intros; subst.
    constructor; auto.
  Qed.
    
End af_rlp.

Section af_list_product.

  Variable (X : Type).
  
  Implicit Type R : X -> X -> Prop.

  Fixpoint rel_list_prod (lR : list (relation X)) :=
    match lR with 
      | nil   => fun _ _ => True
      | R::lR => fun ll mm => 
      match ll  , mm with
        | nil   , nil   => True
        | nil   ,   _   => False
        | _     , nil   => False
        | x::ll , y::mm => R x y /\ rel_list_prod lR ll mm
      end
    end.

  Let f (c : unit + X * list X) := 
    match c with
      | inl _      => nil  
      | inr (x,l) => x::l
    end.

  Let f_surj l : exists c, l = f c.
  Proof.
    destruct l as [ | x l ].
    exists (inl tt); auto.
    exists (inr (x,l)); auto.
  Qed.

  Let R0 : relation unit := fun  _ _ => True.
  
  Let af_R0 : af R0.
  Proof. constructor 1; red; auto. Qed.

  Theorem af_rel_list_prod lR : Forall (@af _) lR -> af (rel_list_prod lR).
  Proof.
    induction 1 as [ | R lR H1 H2 IH2 ].
    simpl; constructor 1; auto.

    generalize (af_fmap _ f_surj (af_sum_both af_R0 (af_product H1 IH2))).
    apply af_inc.
    unfold rel_fmap.

    intros [ | x l ] [ | y m ]; simpl; auto.

    intros ( [ [] | (x',l') ] & [ [] | (y',m') ] & H3 & H4 & H5 ); (discriminate H3 || discriminate H4 || idtac).
    simpl in H4; injection H4; clear H4; intros; subst; red in H5; auto.
    intros ( [ [] | (x',l') ] & [ [] | (y',m') ] & H3 & H4 & H5 ); (discriminate H3 || discriminate H4 || idtac).
    simpl in H3; injection H3; clear H3; intros; subst; red in H5; auto.
    
    intros ( [ [] | (x',l') ] & [ [] | (y',m') ] & H3 & H4 & H5 ); (discriminate H3 || discriminate H4 || idtac).
    simpl in H3, H4; injection H3; injection H4; clear H3 H4; intros; subst; red in H5; auto.
  Qed.

End af_list_product.


Section af_dec.

  Variables (X : Type) (R : X -> X -> Prop).

  Implicit Types (A B : X -> Prop).
  
  Theorem af_dec A B : A inc1 B -> af (R <# B #>) -> af (R <# A #>).
  Proof.
    intros H H1.
    set (f := fun x => match x with exist x Hx => exist _ x (H _ Hx) end).
    generalize (af_cofmap f H1).
    apply af_inc.
    intros (x & Hx) (y & Hy); auto.
  Qed.
  
End af_dec.

Section af_subset_union.

  Variables (X : Type) (R : X -> X -> Prop) (A B : X -> Prop).

  Let AincAB : A inc1 A cup1 B. Proof. intros ?; left; auto. Defined.
  Let BincAB : B inc1 A cup1 B. Proof. intros ?; right; auto. Defined.

  Let f (x : sig A + sig B) := match x with
                                 | inl (exist a Ha) => exist _ a (AincAB Ha)
                                 | inr (exist b Hb) => exist _ b (BincAB Hb)
                               end.                      

  Let f_surj : forall y : {z : X | A z \/ B z}, exists x : {x | A x} + {x | B x}, y = f x.
  Proof.
    intros (y & [ Hy | Hy ]).
    exists (inl (exist _ y Hy)); auto.
    exists (inr (exist _ y Hy)); auto.
  Qed.

  Theorem af_subset_union : af (R <# A #>) -> af (R <# B #>) -> af (R <# A cup1 B #>).
  Proof.
    intros HA HB. 
    generalize (af_fmap _ f_surj (af_sum_both HA HB)).
    apply af_inc.
    intros (x & Hx) (y & Hy) (Hu & Hv & H1 & H2 & H3).
    red; simpl.
    destruct Hu as [ (u & Hu) | (u & Hu) ];
    destruct Hv as [ (v & Hv) | (v & Hv) ];
    red in H3; simpl in H1, H2, H3;
    injection H1; injection H2; intros; subst; auto; tauto.
  Qed.

End af_subset_union.

Section af_subset_product.

  Variables (X: Type) (R : X -> X -> Prop).
  Variable pX : X -> X -> X.

  Hypothesis HpX : forall x1 x2 y1 y2, R x1 x2 -> R y1 y2 -> R (pX x1 y1) (pX x2 y2).

  Variables (A B : X -> Prop).
  
  Notation prodX := (fun u => exists a b, A a /\ B b /\ u = pX a b).

  Let ABinc a b : A a -> B b -> prodX (pX a b).
  Proof. intros H1 H2; exists a, b; split; auto. Defined.

  Let f (x : sig A * sig B) := match x with (exist a Ha, exist b Hb) => exist _ (pX a b) (ABinc Ha Hb) end.

  Let f_surj : forall (z : sig prodX), exists x : sig A * sig B, z = f x.
  Proof.
    intros (z & a & b & Ha & Hb & H).
    exists (exist _ _ Ha, exist _ _ Hb); subst; simpl; auto.
  Qed.
  
  Theorem af_subset_product : af (R <# A #>) -> af (R <# B #>) -> af (R <# prodX #>).
  Proof.
    intros HA HB.
    generalize (af_fmap _ f_surj (af_product HA HB)).
    apply af_inc.
    intros x y.
    intros (((a1 & Ha1),(b1 & Hb1)) 
           &((a2 & Ha2),(b2 & Hb2)) & H1 & H2 & H3 & H4).
    simpl in H1, H2, H3, H4; subst; red; simpl.
    red in H3, H4; simpl in H3, H4.
    apply HpX; auto.
  Qed.

End af_subset_product.

Section af_subset_true.

  Variables (X : Type) (R : X -> X -> Prop).

  Theorem af_subset_true : af (R <# fun _ => True #>) -> af R.
  Proof.
    intros H.
    set (f := fun x : { x : X | True } => proj1_sig x).
    assert (forall z, exists x, z = f x) as Hf.
      intros.
      exists (exist _ z I); auto.
    generalize (af_fmap _ Hf H).
    apply af_inc.
    intros u v ((x & Hx) & (y & Hy) & H1 & H2 & H3).
    subst; apply H3.
  Qed.

End af_subset_true.

Section af_sub_inc.

  Variable (X : Type) (P Q : X -> Prop).
  
  Hypothesis HPQ : P inc1 Q.

  Let f (x : sig P) := match x with exist x Hx => exist _ x (HPQ Hx) end.

  Implicit Type R : X -> X -> Prop.  

  Fact af_sub_inc R : af (fun u v : sig Q => R (proj1_sig u) (proj1_sig v)) 
                   -> af (fun u v : sig P => R (proj1_sig u) (proj1_sig v)).
  Proof.
    intros H.
    generalize (af_cofmap f H).
    apply af_inc.
    intros [] []; simpl; auto.
  Qed.

End af_sub_inc.

Section af_cofmap_sub.

  Variables (X Y : Type).

  Implicit Types (P : X -> Prop) (R : X -> X -> Prop).

  Fact af_cofmap_sub P S : af S -> forall R, (forall x y, P x -> S x y -> R x y) -> af (fun a b : sig P => R (proj1_sig a) (proj1_sig b)).
  Proof.
   induction 1 as [ S HS | S HS IH ]; intros R HR.
   constructor 1; intros [ a Ha ] [ b Hb ]; simpl.
   apply HR; auto.
   constructor 2.
   intros [ u Hu ].
   assert (forall x y : X, P x -> (S rlift u) x y -> (R rlift u) x y) as H.
     intros x y Hx; generalize (HR x y Hx) (HR u x Hu); cbv; tauto.
   generalize (IH u (R rlift u) H).
   apply af_inc.
   intros [ a Ha ] [ b Hb ]; cbv; auto.
  Qed.

  Variable R : X -> X -> Prop.

  Hypothesis Rdec : forall x y, R x y \/ ~ R x y.

  Let K ll x := Forall (fun u => ~ R u x) ll.
  
  Let Kdecide ll : forall x, K ll x \/ exists u, In u ll /\ R u x.
  Proof.
    induction ll as [ | y ll IH ]; intros x.
    left; constructor.
    destruct (Rdec y x) as [ Hy | Hy ].
    right; exists y; simpl; tauto.
    destruct (IH x) as [ H | (u & H1 & H2) ].
    left; constructor; auto.
    right; exists u; simpl; tauto.
  Qed.

  Implicit Arguments K [].

  Let af_lrlift_sub_rec ll mm : 
                   af ((fun u v : {x | K ll x} => R (proj1_sig u) (proj1_sig v)) lrlift mm) 
                -> af ((R lrlift ll) lrlift map (@proj1_sig _ _) mm).
  Proof.
    revert mm; apply af_lrlift_ind.

    intros mm Hmm.
    constructor 1.
    intros x y.
    apply lift_rel_list_ctxt; red.
    left; revert Hmm.
    apply good_map.
    intros ? ?; apply lift_rel_list_incr.
    
    intros mm Hmm IHmm.
    constructor 2.
    intros a.
    destruct (Kdecide ll a) as [ Ha | (u & H1 & H2) ].
    
    apply (IHmm (exist _ a Ha)).
    
    constructor 2; intros b.
    constructor 1.
    intros x y.
    unfold lift_rel.
    right; right.
    apply lift_rel_list_incr.
    apply lift_rel_list_ctxt; red.
    right; left.
    exists u; auto.
  Qed.
  
  Fact af_lrlift_sub ll : af (fun u v : sig (K ll) => R (proj1_sig u) (proj1_sig v)) -> af (R lrlift ll).
  Proof.
    apply af_lrlift_sub_rec with (mm := nil).
  Qed.  

End af_cofmap_sub.

Section af_sig_not.

  Variables (X : Type) (R : X -> X -> Prop) (x : X).
  
  Fact af_rlift_sub : af (R rlift x) -> af (fun u v : { y | ~ R x y } => R (proj1_sig u) (proj1_sig v)).
  Proof.
    intros H.
    apply af_cofmap_sub with (S := R rlift x); auto.
    intros u v H1; cbv; tauto.
  Qed.

  Hypothesis Rdec : forall x y, R x y \/ ~ R x y.
  
  Fact af_sub_rlift : af (fun u v : { y | ~ R x y } => R (proj1_sig u) (proj1_sig v)) -> af (R rlift x).
  Proof.
    intros H.
    apply af_lrlift_sub with (ll := x::nil); auto.
    revert H.
    apply af_sub_inc.
    intros u Hu.
    apply Forall_cons_inv in Hu; tauto.
  Qed.    

End af_sig_not.

Section af_lex_1.

  Variable (X : Type) (P : relation X -> Prop).

  Hypothesis HP : forall R, (  full R 
                             \/ (forall x, P (R rlift x)) 
                             ) -> P R.

  Fact af_lex1_ind : @af _ inc1 P.
  Proof.
    induction 1 as [ R HR | R HR IHR ].
    apply HP; left; auto.
    apply HP; right; auto.
  Qed.

End af_lex_1.

Section af_lex_2.

  Variable (X Y : Type) (P : relation X -> relation Y -> Prop).

  Hypothesis HP0 : forall R S, full S -> full R -> P R S.
  Hypothesis HP1 : forall R S, full S -> (forall x, P (R rlift x) S) -> P R S.
  Hypothesis HP2 : forall R S, (forall x R', af R' -> P R' (S rlift x)) -> P R S.

  Fact af_lex2_ind S : af S -> forall R, af R -> P R S.
  Proof.
    induction 1 as [ S HS | S HS IHS ].
    induction 1 as [ R HR | R HR IHR ].
    apply HP0; auto.
    apply HP1; auto.
    induction 1 as [ R HR | R HR IHR ]; apply HP2; auto.
  Qed.

End af_lex_2.

Section af_lex_3.

  Variable (X Y Z : Type) (P : relation X -> relation Y -> relation Z -> Prop).

  Hypothesis HP0 : forall R S T, full T -> full S -> full R -> P R S T.
  Hypothesis HP1 : forall R S T, full T -> full S -> (forall x, P (R rlift x) S T) -> P R S T.
  Hypothesis HP2 : forall R S T, full T -> (forall y R', af R' -> P R' (S rlift y) T) -> P R S T.
  Hypothesis HP3 : forall R S T, (forall z R' S', af R' -> af S' -> P R' S' (T rlift z)) -> P R S T.

  Fact af_lex3_ind R S T : af R -> af S -> af T -> P R S T.
  Proof.
    intros H1 H2 H3; revert T H3 S H2 R H1.
    induction 1 as [ T HT | T HT IHT ].
    induction 1 as [ S HS | S HS IHS ].
    induction 1 as [ R HR | R HR IHR ].
    apply HP0; auto.
    apply HP1; auto.
    induction 1 as [ R HR | R HR IHR ].
    apply HP2; auto.
    apply HP2; auto.
    induction 1 as [ S HS | S HS IHS ].
    induction 1 as [ R HR | R HR IHR ].
    intros; apply HP3; auto.
    intros; apply HP3; auto.
    induction 1 as [ R HR | R HR IHR ].
    apply HP3; auto.
    apply HP3; auto.
  Qed.

End af_lex_3.

Section af_dep_ind.

  Variable (P : forall X, relation X -> Prop).

  Hypothesis HP0 : P ER.
  Hypothesis HP1 : forall X R, full R -> P ER -> @P X R.
  Hypothesis HP2 : forall X R, (forall x, af (R rlift x))-> (forall x, P (R rlift x)) -> @P X R.

  Theorem af_dep_ind : forall X R, af R -> @P X R.
  Proof.
    intros X.
    induction 1 as [ R HR | R HR IHR ].
    apply HP1; auto.
    apply HP2; auto.
  Qed.

End af_dep_ind.

Section af_lex.

  Definition afr := { X : Type & { R | @af X R } }.

  Definition emp_afr : afr.
  Proof.
    exists Empty_set, ER; constructor 1; intros [].
  Defined.

  Definition full_afr : forall X (R : relation X), full R -> afr.
  Proof.
    intros X R HR.
    exists X, R; constructor 1; apply HR.
  Defined.

  Definition af_afr X R : @af X R -> afr.
  Proof.
    intros HR; exists X, R; apply HR.
  Defined.

  Fact af_lift (aR : afr) (x : projT1 aR) : afr.
  Proof.
    destruct aR as (X & R & HR); simpl in x.
    exists X, (R rlift x).
    induction HR; auto.
    constructor 1; red; firstorder.
  Qed.

  Inductive af_lex : list afr -> list afr -> Prop :=
    | in_af_lex_0 : forall ll mm X S HS, length ll = length mm -> af_lex (emp_afr::ll) (@full_afr X S HS::mm)
    | in_af_lex_1 : forall ll mm aR x, length ll = length mm -> af_lex (af_lift aR x ::ll) (aR::mm)
    | in_af_lex_2 : forall aR ll mm, af_lex ll mm -> af_lex (aR::ll) (aR::mm).

  Let af' (sX : sigT relation) := af (projT2 sX).

End af_lex.

Section af_lex_2_ext.

  Variable (P : forall X Y, relation X -> relation Y -> Prop).

  Hypothesis HP_0 : forall X R, (   X = (Empty_set : Type)
                                 \/ full R /\ P ER ER
                                 \/ forall x, P (R rlift x) ER
                                ) -> @P X _ R ER.

  Hypothesis HP0_ : forall Y S, (   full S /\ (forall X R, af R -> @P X _ R ER)
                                 \/ (forall X' R' x, af R' -> @P X' Y R' (S rlift x))
                                ) -> P ER S.

  Hypothesis HP_F : forall X Y R S, full S -> (   full R /\ P ER S /\ (forall X' R', af R' -> @P X' _ R' ER)
                                               \/ (forall x, @P X Y (R rlift x) S) 
                                               /\ (forall X' R', af R' -> @P X' _ R' ER) 
                                              ) -> @P X Y R S.

 
  Hypothesis HP__ : forall X Y R S, (   full R /\ @P _ Y ER S /\ (forall X' R' x, af R' -> @P X' Y R' (S rlift x))
                                     \/ (forall x, @P X Y (R rlift x) S)
                                     /\ (forall X' R' x, af R' -> @P X' _ R' (S rlift x))
                                    ) -> P R S.
 
  Fact af_ind_lex2 X Y R S : af R -> af S -> @P X Y R S.
  Proof.
    intros HR HS; revert Y S HS X R HR.
    induction 1 as [ | Y S HS H1 | Y S IHS ] using af_dep_ind.

    induction 1 as [ | X R HR H2 | X R IHR ] using af_dep_ind.
    apply HP_0; tauto.
    apply HP_0; tauto.
    apply HP_0; tauto.
    
    induction 1 as [ | X R HR H2 | X R IHR ] using af_dep_ind.
    apply HP0_; tauto.
    apply HP_F; tauto.
    apply HP_F; tauto.
    
    induction 1 as [ | X R HR H2 | X R IHR ] using af_dep_ind.
    apply HP0_; auto.
    apply HP__; left; repeat split; auto.
    apply HP__; right; repeat split; auto.
  Qed.

End af_lex_2_ext.

Section af_lex_k.

  Definition isub (X : Type) := nat -> X -> Prop. 

  Variables (k : nat).

  Variable P : forall X, relation X -> isub X -> Prop.
  
(*  Hypothesis HP1 : forall i, i < k -> forall X RX SX Y RY SY S1 S2, empty (SX i) -> full (RY i) -> (forall j, i < j < k -> 


  Fact af_lex_induction : forall X R S, (forall i, i < k -> af (R <# @S i #>)) -> @P X R S. *)

End af_lex_k.



