(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREER SOFTWARE LICENSE AGREEMENT         *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import wft.
Require Import good_base.
Require Import af_t.

Set Implicit Arguments.

(* Beware that the following results only work when P is a decidable subset of X *)

Section Almost_Full_sub_type.

  Variable X : Type.
  
  Implicit Type R S : X -> X -> Prop.
  
  Variable P : X -> Prop.
  
  Hypothesis Pdec : forall x, { P x } + { ~ P x }.

  (* Beware that R restricted to P (R <# P #>) is secured 
     by a well founded tree wft over X and not over { x | P x }
     
     This works well only when P is a decidable subset
  *)  

  Fixpoint afs_securedby R t :=
    match t with
      | in_wft_zt    => forall x y, P x -> P y -> R x y
      | in_wft_sup f => forall x, P x -> afs_securedby (R rlift x) (f x)
    end.

  Definition afs_t R := { t | afs_securedby R t }.

  Fact in_afs_t0 R : (forall x y, P x -> P y -> R x y) -> afs_t R.
  Proof.
    exists in_wft_zt; auto.
  Qed.

  (* even for the second constructor we need a decidable subset *)

  Fact in_afs_t1 R : (forall a, P a -> afs_t (R rlift a)) -> afs_t R.
  Proof.
    intros g.
    exists (in_wft_sup (fun a => match Pdec a with left Ha => proj1_sig (g a Ha) | _ => in_wft_zt end)); simpl.
    intros a Ha.
    destruct (Pdec a) as [ C | C ].
    apply (proj2_sig (g a C)).
    contradict C; auto.
  Qed.

  Section afs_t_rect.

    Variables (Q : (X -> X -> Prop) -> Type) 
              (HP1 : forall R, (forall x y, P x -> P y -> R x y) -> Q R)
              (HP2 : forall R, (forall a, P a -> afs_t (lift_rel R a)) 
                            -> (forall a, P a -> Q (lift_rel R a)) 
                            -> Q R).
    
    Theorem afs_t_rect : forall R, afs_t R -> Q R.
    Proof.
      intros R (t & HR).
      revert R HR Q HP1 HP2.
      induction t as [ | f Hf ]; simpl; intros R HR Q HP1 HP2.
      apply HP1, HR.
      apply HP2.
      intros a Ha; exists (f a); apply (HR _ Ha).
      intros a Ha; apply (Hf a); auto.
    Qed.

  End afs_t_rect.

  Fact afs_t_inc R S : R inc2 S -> afs_t R -> afs_t S.
  Proof.
    intros H0 H1; revert S H0.
    induction H1 as [ R HR | R HR1 HR2 ] using afs_t_rect; intros S H.
    apply in_afs_t0.
    intros; apply H; auto.
    apply in_afs_t1.
    intros a Ha; apply (HR2 a), lift_rel_inc; auto.
  Qed.

  Fact afs_t_impl_af_t R : afs_t R -> af_t (R <# P #>).
  Proof.
    induction 1 as [ R H | R HR1 HR2 ] using afs_t_rect.
    apply in_af_t0; intros (?&?) (?&?); cbv; auto.
    apply in_af_t1.
    intros (a&Ha); generalize (HR2 a Ha).
    apply af_t_inc.
    intros (?&?) (?&?); cbv; auto.
  Qed.

  (* for the converse, we need a decidable subset P *)

  Let af_t_impl_afs_t_rec (R : relation (sig P)) : af_t R 
                                                -> (forall x y x' y', proj1_sig x = proj1_sig x' 
                                                                   -> proj1_sig y = proj1_sig y'
                                                                   -> R x y -> R x' y')
                                                -> afs_t (fun x y => match Pdec x, Pdec y with 
                                                       | left Hx, left Hy => R (exist _ x Hx) (exist _ y Hy)
                                                       | _     , _        => False
                                                     end).
  Proof.
    induction 1 as [ R HR | R HR1 HR2 ] using af_t_rect; intros H.
    apply in_afs_t0.
    intros x y Hx Hy.
    destruct (Pdec x); destruct (Pdec y); auto; tauto.
    apply in_afs_t1.
    intros a Ha.
    specialize (HR2 (exist _ a Ha)).
    spec HR2 H'.
    intros x y x' y' ? ? [ | ]; [ left | right ].
    apply (H x y); auto.
    apply (H (exist _ a Ha) x); auto.
    revert HR2.
    apply afs_t_inc.
    intros x y.
    unfold lift_rel.
    destruct (Pdec x); destruct (Pdec y); destruct (Pdec a); cbv; try tauto.
    intros [ | H1 ]; [ left | right ]; auto.
    revert H1; apply H; auto.
  Qed.
  
  Fact af_t_impl_afs_t R : af_t (R <# P #>) -> afs_t R.
  Proof.
    intros H.
    specialize (af_t_impl_afs_t_rec H).
    spec af_t_impl_afs_t_rec H'.
    intros (?&?) (?&?) (?&?) (?&?); cbv; intros ? ?; subst; auto.
    revert af_t_impl_afs_t_rec.
    apply afs_t_inc.
    intros x y; destruct (Pdec x); destruct (Pdec y); auto; tauto.
  Qed.

End Almost_Full_sub_type.

Fact af_t_any_afs_t X P R : (forall x : X, { P x } + { ~ P x }) -> af_t R -> afs_t P R.
Proof.
  intros HP.
  induction 1 as [ R G1 | R G1 G2 ] using af_t_rect.
  apply in_afs_t0; intros; auto.
  apply in_afs_t1; auto.
Qed.

Fact af_t_afs_t_True X R : @af_t X R -> afs_t (fun _ => True) R.
Proof.
  induction 1 using af_t_rect.
  apply in_afs_t0; intros; auto.
  apply in_afs_t1; intros; auto; tauto.
Qed.

Fact afs_t_True_af_t X R : afs_t (fun _ => True) R -> @af_t X R.
Proof.
  induction 1 using afs_t_rect.
  apply in_af_t0; auto.
  apply in_af_t1; auto.
Qed.

Section afs_t_sb_restr.

  (* if the wft t secures (P,R) then the SAME tree secures (P,R cap2 P x P) *)

  Variable (X : Type) (P : X -> Prop) (R S : X -> X -> Prop).
  Hypothesis HRS : forall x y, P x -> P y -> R x y -> S x y.

  Fact afs_t_sb_restr t : afs_securedby P R t -> afs_securedby P S t.
  Proof.
    revert R S HRS.
    induction t as [ | f IHf ]; intros R S HRS; simpl.
    intros; auto.
    intros H x Hx.
    generalize (H x Hx); apply IHf.
    intros ? ? ? ? [|]; [ left | right ]; auto.
  Qed.

  Fact afs_t_restr : afs_t P R -> afs_t P S.
  Proof. intros (t & Ht); exists t; revert Ht; apply afs_t_sb_restr. Qed.

End afs_t_sb_restr.

Section afs_t_sb_incl.

  (* if Q inc1 P and R inc2 S then if a wft t secures (P,R) then
     the SAME tree secures (Q,S) *)

  Variable (X : Type) (P Q : X -> Prop) (HPQ : Q inc1 P) 
                 (R S : X -> X -> Prop) (HRS : R inc2 S).

  Fact afs_t_sb_incl t : afs_securedby P R t -> afs_securedby Q S t.
  Proof.
    revert R S HRS.
    induction t as [ | f IHf ]; intros R S HRS; simpl.
    intros; auto.
    intros H x Hx; generalize (H x (HPQ Hx)); apply IHf.
    cbv; firstorder.
  Qed.

  Fact afs_t_incl : afs_t P R -> afs_t Q S.
  Proof. intros (t & Ht); exists t; revert Ht; apply afs_t_sb_incl. Qed.

End afs_t_sb_incl.

Fact afs_t_sb_cap X P R t : @afs_securedby X P R t <-> afs_securedby P (fun x y => R x y /\ P x /\ P y) t.
Proof.
  split.
  apply afs_t_sb_restr; intros; auto.
  apply afs_t_sb_incl; intros; tauto.
Qed.

Fact afs_t_cap1 X P R : @afs_t X P R -> afs_t P (fun x y => R x y /\ P x /\ P y).
Proof. intros (t & Ht); exists t; revert Ht; apply afs_t_sb_cap. Qed.

Fact afs_t_cap2 X P R : afs_t P (fun x y => R x y /\ P x /\ P y) -> @afs_t X P R. 
Proof. intros (t & Ht); exists t; revert Ht; apply afs_t_sb_cap. Qed.

Section afs_t_relmap.

  Variables (X Y : Type) (f : X -> Y -> Prop) 
            (P : X -> Prop) (Q : Y -> Prop) (Qdec : forall x, { Q x } + { ~ Q x }) 
            (R : X -> X -> Prop) (S : Y -> Y -> Prop)
            (Hf : forall y, Q y -> { x | f x y /\ P x })
            (HRS : forall x1 x2 y1 y2, P x1 -> P x2 -> Q y1 -> Q y2 -> f x1 y1 -> f x2 y2 -> R x1 x2 -> S y1 y2).

  Fact afs_t_relmap : afs_t P R -> afs_t Q S.
  Proof.
    intros H; revert S HRS.
    induction H as [ R HR | R HR IH ] using afs_t_rect; intros S H.

    apply in_afs_t0.
    intros y1 y2 Hy1 Hy2.
    destruct (Hf Hy1) as (x1 & H1 & Hx1).
    destruct (Hf Hy2) as (x2 & H2 & Hx2).
    apply H with (7 := HR _ _ Hx1 Hx2); auto.
    
    apply in_afs_t1; auto.
    intros y Hy.
    destruct (Hf Hy) as (x & Hx & H1).
    apply IH with (1 := H1).
    intros x1 x2 y1 y2 H2 H3 H4 H5 H6 H7.
    unfold lift_rel.
    intros [ | ]; [ left | right ].
    apply H with (5 := H6) (6 := H7); auto.
    apply H with (5 := Hx) (6 := H6); auto.
  Qed.
    
End afs_t_relmap.

Section afs_t_sum.

  Variable (X Y : Type) 
           (P : X -> Prop) (Q : Y -> Prop) 
           (R : X -> X -> Prop) (S : Y -> Y -> Prop).

  Hypothesis HX : afs_t P R.
  Hypothesis HY : afs_t Q S.
  
  Variable Pdec : forall x, { P x } + { ~ P x }.
  Variable Qdec : forall y, { Q y } + { ~ Q y }.

  Let f : sig P + sig Q -> sig (pred_sum P Q).
  Proof. 
    intros [ (x & Hx) | (y & Hy) ].
    exists (inl x); auto.
    exists (inr y); auto.
  Defined.

  Let f_surj : forall y, { x | y = f x }.
  Proof. 
    intros ( [ x | y ] & H ); simpl in H.
    exists (inl (exist _ _ H)); auto.
    exists (inr (exist _ _ H)); auto.
  Qed.

  Fact afs_t_sum_both : afs_t (pred_sum P Q) (sum_both R S).
  Proof.
    apply afs_t_impl_af_t in HX.
    apply afs_t_impl_af_t in HY.
    generalize (af_t_sum_both HX HY); intros H.
    apply af_t_impl_afs_t.
    intros [ x | y ]; simpl; auto.
    revert H.
 
    apply af_t_relmap with (f := fun c d => match c, proj1_sig d with 
                                              | inl x, inl y => proj1_sig x = y 
                                              | inr x, inr y => proj1_sig x = y
                                              | _    , _     => False
                                            end).
    intros ( [ x | y ] & H1); simpl in H1.
    exists (inl (exist _ x H1)); simpl; auto.
    exists (inr (exist _ y H1)); simpl; auto.
    
    intros [ (x1 & H1) | (y1 & H1) ]
           [ (x2 & H2) | (y2 & H2) ]
           [ [x3 | y3] H3 ]
           [ [x4 | y4] H4 ]; simpl; try tauto; intros ?  ?; subst; cbv; auto.
  Qed.

End afs_t_sum.

Section afs_t_prod.

  Variable (X Y : Type) 
           (P : X -> Prop) (Q : Y -> Prop) 
           (R : X -> X -> Prop) (S : Y -> Y -> Prop).
  
  Hypothesis HR : afs_t P R.
  Hypothesis HQ : afs_t Q S.

  Variable Pdec : forall x, { P x } + { ~ P x }.
  Variable Qdec : forall y, { Q y } + { ~ Q y }.

  Let pred_prod_dec c : { pred_prod P Q c } + { ~ pred_prod P Q c }.
  Proof.
    unfold pred_prod.
    destruct c as (x & y); simpl.
    destruct (Pdec x); destruct (Qdec y); tauto.
  Qed.

  Fact afs_t_product : afs_t (pred_prod P Q) (rel_prod R S).
  Proof.
    apply afs_t_impl_af_t in HR.
    apply afs_t_impl_af_t in HQ.
    apply af_t_impl_afs_t; auto.
    generalize (af_t_prod HR HQ).
    apply af_t_relmap with (f := fun c d => proj1_sig (fst c) = fst (proj1_sig d)
                                         /\ proj1_sig (snd c) = snd (proj1_sig d)).
    intros ((x,y) & Hx & Hy); simpl in Hx, Hy.
    exists (exist _ _ Hx, exist _ _ Hy); auto.
    intros ((x1 & Hx1),(y1 & Hy1)) ((x2 & Hx2),(y2 & Hy2))
           ((x3,y3) & Hx3 & Hy3) ((x4,y4) & Hx4 & Hy4); red; simpl.
    intros (? & ?) (? & ?); subst; auto.
  Qed.

End afs_t_prod.

Section afs_t_idx_rel.
  
  Variables (I X : Type) (P : X -> Prop) (Pdec : forall x, { P x } + { ~ P x }).

  Fact afs_t_idx_rel R (li : list I) : (forall i, In_t i li -> afs_t P (R i))
                         -> afs_t (fun l => Forall P l /\ length l = length li) (idx_rel R li).
  Proof.
    induction li as [ | i li IHli ]; intros H.

    apply in_afs_t0.
    intros [ | ] [ | ] (Hx1 & Hx2) (Hy1 & Hy2).
    constructor.
    discriminate Hy2.
    discriminate Hx2.
    discriminate Hx2.
    
    spec_all IHli.
    intros; apply H; right; auto.
    specialize (H i); spec_all H.
    left; auto.
    
    assert (forall n l, { Forall P l /\ length l = n } + { ~ (Forall P l /\ length l = n) }) as H'.
      intros n l.
      destruct (eq_nat_dec (length l) n);
      destruct (Forall_dec P l); try tauto; intros; auto.
    
    generalize (afs_t_product H IHli); intros H2; spec_all H2; auto.
    revert H2.
    apply afs_t_relmap with (f := fun c m => fst c::snd c = m); auto.

    intros [ | y ly ] (H1 & H2).
    discriminate H2.
    injection H2; clear H2; intros H2.
    apply Forall_cons_inv in H1; destruct H1.
    exists (y,ly); cbv; auto.
    
    intros (x1 & lx1) (x2 & lx2) 
           [ | y1 ly1 ] [ | y2 ly2 ]; simpl;
    intros (H1 & H2 & H3) (H4 & H5 & H6) (H7 & H8) (H9 & H10); 
      try discriminate 1; try discriminate 2.
    intros G1 G2; injection G1; injection G2; clear G1 G2; intros ? ? ? ?; subst.
    intros (? & ?); constructor; auto.
  Qed.

End afs_t_idx_rel.

Section afs_t_union.

  Variable (X : Type) (P Q : X -> Prop) (R S : X -> X -> Prop).
  
  Hypothesis HR : afs_t P R.
  Hypothesis HS : afs_t Q S.
  
  Hypothesis Pdec : forall x, { P x } + { ~ P x }.
  Hypothesis Qdec : forall x, { Q x } + { ~ Q x }.
  
  Fact afs_t_union : afs_t (P cup1 Q) (R cup2 S).
  Proof.
    generalize (afs_t_sum_both HR HS Pdec Qdec).
    apply afs_t_relmap with (f := fun x y => y = match x with inl x => x | inr x => x end).
    intros x; destruct (Pdec x); destruct (Qdec x); tauto.
    intros x Hx.
    destruct (Pdec x) as [ H | H ].
    exists (inl x); auto.
    exists (inr x); split; auto; red; tauto.    
    intros [|] [|]; simpl; intros; subst; tauto.
  Qed.

End afs_t_union.

Section afs_sig_not.

  Variables (X : Type) (R : X -> X -> Prop) (x : X).
  
  Hypothesis Rdec : forall y, { R x y } + { ~ R x y }.

  Section afs_t_rlift_afs_t_cap.

    Variables (P : X -> Prop) (Pdec : forall x, { P x } + { ~ P x }).
    
    Fact afs_t_rlift_afs_t_cap: afs_t P (R rlift x) -> afs_t (P cap1 (fun y => ~ R x y)) R.
    Proof.
      apply afs_t_relmap with (f := fun a b => b = a /\ ~ R x a).
      intros y; destruct (Rdec y); destruct (Pdec y); tauto.
      intros y (H1 & H2); exists y; auto.
      cbv; intros x1 x2 y1 y2 _ _ _ _ (? & H) (? & _); subst; tauto.
    Qed.
    
    Fact afs_t_cap_afs_t_rlift : afs_t (P cap1 (fun y => ~ R x y)) R -> afs_t P (R rlift x). 
    Proof.

      (* we use afs_union with P = (P cap ~ R x) cup (P cap R x) *)
    
      intros H0.
      assert (afs_t (P cap1 R x) (R rlift x)) as H1.
        apply in_afs_t0; cbv; firstorder.
      generalize (afs_t_union H0 H1); clear H1; intros H1.
      spec_all H1.
      intros y; destruct (Rdec y); destruct (Pdec y); tauto.
      intros y; destruct (Rdec y); destruct (Pdec y); tauto.
      
      revert H1; apply afs_t_relmap with (f := fun a b => b = a); auto.
      intros y Hy; exists y; split; auto.
      specialize (Rdec y); tauto.
      cbv; intros; subst; tauto.
    Qed.

  End afs_t_rlift_afs_t_cap.

  Fact af_t_rlift_afs_t : af_t (R rlift x) -> afs_t (fun y => ~ R x y) R.
  Proof.
    intros H.
    apply af_t_afs_t_True, afs_t_rlift_afs_t_cap in H.
    revert H.
    apply afs_t_incl; auto.
    tauto.
  Qed.
  
  Fact afs_t_af_t_rlift : afs_t (fun y => ~ R x y) R -> af_t (R rlift x).
  Proof.
    intros H.
    apply afs_t_True_af_t, afs_t_cap_afs_t_rlift.
    tauto.
    revert H; apply afs_t_incl; tauto.
  Qed.

End afs_sig_not.

Section afs_owft_sec.

  Variable X : Type.

  Implicit Types P Q : X -> Prop.
  Implicit Types R S : X -> X -> Prop.

  (* (empty,R) can be secured by None whereas (P,full) can be secured by Some in_wft_zt 
     where None is strictly less than Some _ for some well founded relation 
     on owft (option (wft X)), ie either easier or more facile *)
  
  Definition afs_owft_sec (t : owft X) P R :=
    match t with
      | None   => forall x, ~ P x 
      | Some t => afs_securedby P R t
    end.

  Fact afs_owft_sec_incl t P Q R S : Q inc1 P -> R inc2 S -> afs_owft_sec t P R -> afs_owft_sec t Q S.
  Proof.
    intros H1 H2.
    destruct t as [ t | ]; simpl.
    apply afs_t_sb_incl; auto.
    intros H x ?; apply (H x); auto.
  Qed.

  Fact afs_owft_afs_t t P R : afs_owft_sec t P R -> afs_t P R.
  Proof.
    destruct t as [ t | ]; simpl.
    exists t; auto.
    intros H; apply in_afs_t0; intros x ? ?; exfalso; apply (H x); auto. 
  Qed.
  
  Fact afs_t_afs_owft P R : afs_t P R -> { t : _ | afs_owft_sec t P R }.
  Proof.
    intros (t & Ht); exists (Some t); simpl; auto.
  Qed.

End afs_owft_sec.
