(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Omega.

Require Import list_utils.
Require Import rel_utils.
Require Import good_base.
Require Import wft.

Set Implicit Arguments.

Section Bar_Prop.

  Variables (A : Type) (P : list A -> Prop).

  Inductive bar : list A -> Prop :=
    | in_bar_0 : forall ll, P ll -> bar ll
    | in_bar_1 : forall ll, (forall a, bar (a::ll)) -> bar ll.

  Implicit Type f : nat -> A. 

  Let bar_recursion ws : bar ws -> forall f n, ws = pfx_rev f n -> exists m, P (pfx_rev f m).
  Proof.
    induction 1 as [ ll Hll | ll Hll IH ]; intros f n ?; subst.
    exists n; auto.
    apply (IH (f n) f (S n)); auto.
  Defined.
  
  Theorem bar_inv : bar nil -> forall f, exists m, P (pfx_rev f m).
  Proof.
    intros H f; apply bar_recursion with (1 := H) (n := 0); auto.
  Defined.

  Hypothesis P_cons : forall x ll, P ll -> P (x::ll).

  Theorem bar_app ll mm : bar ll -> bar (mm++ll).
  Proof.
    intros H.
    induction mm as [ | x mm IH ]; simpl; auto.
    inversion_clear IH; auto.
    constructor 1; apply P_cons; auto.
  Qed.

  Theorem bar_nil_all : bar nil -> forall ll, bar ll.
  Proof.
    intros H ll.
    rewrite app_nil_end.
    apply bar_app, H.
  Qed.

End Bar_Prop.

Section Bar_inc.

  Variable (A : Type) (P Q : list A -> Prop).

  Fact bar_inc : P inc1 Q -> bar P inc1 bar Q.
  Proof.
    intros H ?; induction 1 as [ ws Hws | ws IH ].
    constructor 1; apply H, Hws.
    constructor 2; auto.
  Qed.

End Bar_inc.

Section Bar_lift.

  Variables (A : Type) (P : list A -> Prop).

  Let bar_lift1 k : bar P k -> forall u v, k = v++u -> bar (fun v => P (v++u)) v.
  Proof.
    induction 1 as [ k Hk | k Hk IH ]; intros u v H.
    
    constructor 1; subst; auto.
    
    constructor 2; intros a.
    apply (IH a u (a::v)).
    simpl; f_equal; auto.
  Qed.
  
  Let bar_lift2 u v : bar (fun v => P (v++u)) v -> bar P (v++u).
  Proof.
    induction 1 as [ v Hv | v Hv IH ].
    constructor 1; auto.
    constructor 2; intros a.
    apply (IH a).
  Qed.

  Theorem bar_lift u : bar P u <-> bar (fun v => P (v++u)) nil.
  Proof.
    split.
    intros H; apply bar_lift1 with (1 := H); auto.
    apply bar_lift2.
  Qed.

End Bar_lift.

Section bar_subst_eq.

  Variable (A : Type) (P : list A -> Prop) (a b : A).
  
  Hypothesis Hab : forall mm ll, P (mm++a::ll) -> P (mm++b::ll).
  
  Let bar_subst_eq_rec l : bar P l -> forall mm ll, l = mm++a::ll -> bar P (mm++b::ll).
  Proof.
    induction 1 as [ l Hl | l Hl IH ]; intros; subst.
    constructor 1; auto.
    constructor 2; intros c.
    apply (IH c (c::mm)); auto.
  Qed.
  
  Fact bar_subst_eq mm ll : bar P (mm++a::ll) -> bar P (mm++b::ll).
  Proof.
    intros H; apply bar_subst_eq_rec with (1 := H); auto.
  Qed.

End bar_subst_eq.

Section Bar_inter.

  Variables (A : Type) (P Q : list A -> Prop).

  Hypothesis HP : forall x ll, P ll -> P (x::ll).
  Hypothesis HQ : forall x ll, Q ll -> Q (x::ll).

  Theorem bar_intersection u : bar P u -> bar Q u -> bar (fun x => P x /\ Q x) u.
  Proof.
    induction 1 as [ u Hu | u Hu IHu ].
    induction 1 as [ v Hv | v Hv IHv ].
    constructor 1; auto.
    constructor 2; intros a.
    apply IHv, HP; auto.
    induction 1 as [ v Hv | v Hv IHv ].
    constructor 2; intros a.
    apply IHu.
    constructor 1.
    apply HQ; auto.
    constructor 2; intros a.
    apply IHu; auto.
  Qed.

End Bar_inter.

Section Bar_map.

  Variables (A B : Type) (f : A -> B) (Hf : forall b, exists a, b = f a)
            (P : list A -> Prop) 
            (Q : list B -> Prop)
            (HPQ1 : forall ll, P ll -> Q (map f ll))
            (HPQ2 : forall ll, Q (map f ll) -> P ll).

  Lemma bar_map_inv ll : bar Q (map f ll) -> bar P ll.
  Proof.
    set (ll' := map f ll); generalize (eq_refl ll').
    unfold ll' at 2.
    generalize ll'; clear ll'; intros ll'.
    intros H3 H4.
    revert H4 ll H3.  
    induction 1 as [ mm Hmm | mm Hmm IH ]; intros ll ?; subst.
    constructor 1; apply HPQ2, Hmm.
    constructor 2; intros a.
    apply IH with (f a); auto.
  Qed.

  Lemma bar_map ll : bar P ll -> bar Q (map f ll).
  Proof.
    induction 1 as [ ll Hll | ll Hll IH ]; simpl.
    constructor 1; apply HPQ1; auto.
    constructor 2; intros b.
    destruct (Hf b) as (a & Ha); subst.
    apply IH.
  Qed.

End Bar_map.
