(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Wf.
Require Import Wellfounded.
Require Import Relations.

Require Omega.

Require Import rel_utils.
Require Import list_utils.

Require Import good_base.
Require Import af.
Require Import bar.
Require Import af_bar.
Require Import afs.

Set Implicit Arguments.

Section Bars_Prop.

  Variables (A : Type) (P : A -> Prop) (Q : list A -> Prop).

  Inductive bars : list A -> Prop :=
    | in_bars_0 : forall ll, Forall P ll -> Q ll -> bars ll
    | in_bars_1 : forall ll, Forall P ll -> (forall a, P a -> bars (a::ll)) -> bars ll.

  Implicit Type f : nat -> A. 

  Let bars_recursion ws : bars ws -> forall f n, (forall n, P (f n)) -> ws = pfx_rev f n -> exists m, Q (pfx_rev f m).
  Proof.
    induction 1 as [ ll H1 H2 | ll H1 H2 IH ]; intros f n Hf Hws; subst.
    exists n; auto.
    apply (IH (f n) (Hf n) f (S n)); auto.
  Defined.
  
  Theorem bars_inv : bars nil -> forall f, (forall n, P (f n)) -> exists m, Q (pfx_rev f m).
  Proof.
    intros H f Hf; apply bars_recursion with (1 := H) (2 := Hf) (n := 0); auto.
  Defined.

  Fact bars_Forall ws : bars ws -> Forall P ws.
  Proof.
    induction 1; auto.
  Qed.

  Hypothesis Q_cons : forall x ll, P x -> Forall P ll -> Q ll -> Q (x::ll).

  Theorem bars_app ll mm : Forall P mm -> bars ll -> bars (mm++ll).
  Proof.
    induction 1 as [ | x mm Hx Hmm IH ]; simpl; auto; intros H.
    apply IH in H.
    inversion_clear H; auto.
    constructor 1.
    constructor; auto.
    apply Q_cons; auto.
  Qed.

  Theorem bars_nil_all : bars nil -> forall ll, Forall P ll -> bars ll.
  Proof.
    intros H ll Hll.
    rewrite app_nil_end.
    apply bars_app; auto.
  Qed.

End Bars_Prop.

Section Bar_Bars.

  Variable (X : Type) (P : X -> Prop) (Q : list X -> Prop).

  Let Q' (ll : list (sig P)) := Q (map (@proj1_sig _ _) ll).
  
  Let Forall_sig (ll : list (sig P)) : Forall P (map (@proj1_sig _ _) ll).
  Proof.
    induction ll as [ | (? & ?) ]; simpl; constructor; auto.
  Qed.
  
  Let bars_bar_rec ll : bars P Q ll -> forall ll', ll = map (@proj1_sig _ _) ll' -> bar Q' ll'.
  Proof.
    induction 1 as [ ll H1 H2 | ll H1 H2 IH ]; intros ll' ?; subst.
    constructor 1; auto.
    constructor 2; intros (a & Ha).
    apply (IH _ Ha); auto.
  Qed.
    
  Fact bar_bars_eq ll : bar Q' ll <-> bars P Q (map (@proj1_sig _ _) ll).
  Proof.
    split.

    induction 1 as [ | ll H1 IH ].
    constructor 1; auto.
    constructor 2; auto.
    intros a Ha.
    apply (IH (exist _ a Ha)).
    
    intros H.
    apply bars_bar_rec with (1 := H); auto.
  Qed.

  Fact bar_eq_bars : bar Q' nil <-> bars P Q nil.
  Proof.
    apply bar_bars_eq.
  Qed.

End Bar_Bars.

Fact afs_eq_bars X P R : @afs X P R <-> bars P (good R) nil.
Proof.
  rewrite <- bar_eq_bars, afs_af_rel_restr_eq.
  rewrite af_eq_bar_good; split.
  apply bar_inc.
  intros x; apply good_map; auto.
  apply bar_inc.
  intros x; apply good_map_inv; auto.
Qed.

Section Bars_mono.

  Variable (A : Type) (P1 P2 : A -> Prop) (Q1 Q2 : list A -> Prop).

  Fact bars_mono : P2 inc1 P1 -> Q1 inc1 Q2 -> forall ws, Forall P2 ws -> bars P1 Q1 ws -> bars P2 Q2 ws.
  Proof.
    intros H1 H2 lws Hws; induction 1 as [ ws H3 H4 | ws H3 H4 IH ].
    constructor 1; auto.
    constructor 2; auto.
  Qed.

End Bars_mono.

Section Bars_lift.

  Variables (A : Type) (P : A -> Prop) (Q : list A -> Prop).

  Let bars_lift1 k : bars P Q k -> forall u v, k = v++u -> bars P (fun v => Q (v++u)) v.
  Proof.
    induction 1 as [ k H1 H2 | k H1 H2 IH ]; intros u v H; subst; 
      apply Forall_app in H1.
    constructor 1; auto; tauto.
    constructor 2; try tauto.
    intros a Ha.
    apply (IH a Ha u (a::v)); auto.
  Qed.
  
  Let bars_lift2 u v : Forall P u -> bars P (fun v => Q (v++u)) v -> bars P Q (v++u).
  Proof.
    intros Hu.
    induction 1 as [ v H H2 | v H1 H2 IH ].
    constructor 1; auto.
    apply Forall_app; auto.
    constructor 2.
    apply Forall_app; auto.
    intros a Ha; apply (IH a Ha).
  Qed.

  Theorem bars_lift u : bars P Q u <-> Forall P u /\ bars P (fun v => Q (v++u)) nil.
  Proof.
    split.
    intros H.
    generalize (bars_Forall H); intros H'.
    split; auto.
    apply bars_lift1 with (1 := H); auto.
    intros (H1 & H2).
    apply bars_lift2 in H2; auto.
  Qed.

End Bars_lift.

Section bar_bars_eq.

  Variable (X : Type) (Q : list X -> Prop).

  Let Forall_True (ll : list X) : Forall (fun _ => True) ll.
  Proof. induction ll; constructor; auto. Qed.

  Fact bar_bars_True_eq ll : bar Q ll <-> bars (fun _ => True) Q ll.
  Proof.
    split; (induction 1; [ constructor 1 | constructor 2 ]; auto ).
  Qed.

End bar_bars_eq.

Section Bars_map.

  Variables (A B : Type) 
            (PA : A -> Prop)
            (PB : B -> Prop)
            (f : A -> B) 
            (Hf0 : forall a, PA a -> PB (f a))
            (Hf1 : forall a, PB (f a) -> PA a)
            (Hf2 : forall b, PB b -> exists a, PA a /\ b = f a)
            (QA : list A -> Prop) 
            (QB : list B -> Prop)
            (HPQ1 : forall ll, Forall PA ll -> QA ll -> QB (map f ll))
            (HPQ2 : forall ll, Forall PA ll -> QB (map f ll) -> QA ll).

  Let Forall_prop1 ll : Forall PA ll -> Forall PB (map f ll).
  Proof.
    rewrite Forall_map; revert ll.
    apply Forall_impl; auto.
  Qed.

  Let Forall_prop2 ll : Forall PB (map f ll) -> Forall PA ll.
  Proof.
    rewrite Forall_map; revert ll.
    apply Forall_impl; auto.
  Qed.

  Let bars_map_inv_rec mm : bars PB QB mm -> forall ll, mm = map f ll -> bars PA QA ll.
  Proof.
    induction 1 as [ mm H1 H2 | mm H1 H2 IH ]; intros ll ?; subst.
    constructor 1; auto.
    constructor 2; auto.
    intros a Ha.
    apply Hf0 in Ha.
    apply (IH _ Ha); auto.
  Qed.

  Lemma bars_map_inv ll : bars PB QB (map f ll) -> bars PA QA ll.
  Proof.
    intros H; apply bars_map_inv_rec with (1 := H); auto.
  Qed.

  Lemma bars_map ll : bars PA QA ll -> bars PB QB (map f ll).
  Proof.
    induction 1 as [ ll H1 H2 | ll H1 H2 IH ]; simpl.
    constructor 1; auto.
    constructor 2; auto; intros b Hb.
    destruct (Hf2 Hb) as (a & Ha & ?); subst.
    apply IH; auto.
  Qed.

End Bars_map.

Section bars_ramsey.

  Variables (X Y : Type) (P : X -> Prop) (Q : Y -> Prop) (R : X -> X -> Prop) (S : Y -> Y -> Prop).

  Fact bars_ramsey : bars P (good R) nil 
                  -> bars Q (good S) nil 
                  -> bars (pred_prod P Q) (good (fun x y => 
                                     match x,y with 
                                      | (x1,x2), (y1,y2) => R x1 y1 
                                                         /\ S x2 y2 
                                     end)) nil.
  Proof.
    do 3 rewrite <- afs_eq_bars.
    intros H1 H2.
    generalize (afs_product H1 H2).
    apply afs_mono; auto.
    intros [] [] ?; auto.
  Qed.

End bars_ramsey.

