(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Min.
Require Import List.
Require Import Relations.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import subword.

Require Import af.
Require Import afs.
Require Import afs_lex.

Require Import tree.
Require Import tree_embeddings.

Require Import tree_kruskal_fix.
Require Import tree_kruskal_ev3.
Require Import tree_kruskal_ev4.

Set Implicit Arguments.

Section higman_ktree_af_rec.

  Variable X : Type.
  Notation T := (ktree_fix X).

  Variable k : nat.
  
  Let Q P R :=  (forall n, ~ P n (@in_ktf_u X))  
             -> (forall n x, { P n x } + { ~ P n x })
             -> (forall n, k < n -> P n = fun _ => False)
             ->  afs (wfptree P) (emb_tree_higman R).

  Theorem higman_ktree_af_rec P R : (forall n, afs (P n) (R n)) -> Q P R.
  Proof.
    revert P R.
    apply afs_easier_ind with (k := S k); unfold Q; clear Q.

    (* Q is extensional *)
    
    intros P1 P2 R1 R2 H1 H2 H3 H4 H5 H6.
    spec_all H3.
    intro; rewrite H1; auto.
    intros ? ?; rewrite H1; auto.
    intros ? ?; rewrite H1; auto.
    revert H3.
    apply afs_mono.
    apply tree_fall_inc.
    intros ? ?; rewrite H1; auto.
    apply emb_tree_higman_inc.
    intro; rewrite H2; auto.
    
    (* **** *)

    intros w P R HR IHs HP1 HP2 HP3.
    assert (forall t, { wfptree P t } + { ~ wfptree P t }) as HP4.
      apply wfptree_dec, HP2.
    assert (forall i, afs (P i) (R i)) as HPR.
      revert HR; apply afs_correct_afs.

(* now starts the nested induction over the well-formed tree *)

    apply in_afs_1; auto.    
    induction 1 as [ m lt H_m_lt IHt ] using wfptree_ind.
    
    case_eq (length lt).

(* case where lt is empty, ie we have a leaf *)
    
    intros Hl_lt.
    destruct lt.
    2: discriminate Hl_lt.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in H_m_lt.
    destruct H_m_lt as [ Hm1 _ ]; simpl in Hm1.
    generalize (HR 0); intros HR0.

    case_eq (w 0); intros Hs0; 
      rewrite Hs0 in HR0; simpl in HR0.
   
    (* w 0 = af_empty, then P 0 is empty *)
    
    contradict Hm1; apply HR0; omega.

    (* w 0 = af_full, then R 0 is full on P 0 *)
    
    apply in_afs_0.
    intros x y Hx Hy; right.
    clear y Hy.
    destruct (tree_leaf x) as (l & Hl).
    unfold wfptree in Hx.
    rewrite tree_fall_sub_tree in Hx.    
    generalize Hl.
    apply emb_sub_tree_higman.
    constructor 2; simpl.
    apply HR0; auto; try omega.
    apply Hx with (1 := Hl).
    constructor.
    
    (* w 0 = af_rlift, then R rlift x is secured by f x *)
    
    spec_all HR0; try omega.    
    set (w' n := match n with 0 => af_rlift | S n => w (S n) end).
    set (R' n := match n with 0 => R 0 rlift m | S n => R (S n) end).

    specialize (IHs w' P R'); spec_all IHs; auto.

    intros [ | j ].
    simpl; revert HR0; apply afs_inc; left; auto. 
    apply HR; auto.
    
    exists 0.
    split; auto.
    omega.
    split; auto.
    unfold w', R'; rewrite Hs0.
    apply lt_afsr_rlift; auto.
    intros [ | j ] ?; simpl; auto; omega.
    
    revert IHs.   
    apply afs_inc.
    intros t1 t2.
    induction 1 as [ t1 t2 x ll G1 G2 [ G3 | G3 ] 
                   | x y ll mm G1 G2 G3].
    left; constructor 1 with t2; auto.
    right; auto.
    apply Forall2_decide_ind in G3.
    destruct G3 as [ G3 | G3 ].

    destruct ll; simpl in G1.
    destruct G1 as [ G1 | G1 ].
    left; constructor 2; auto.
    right; constructor 2; auto.
    left; constructor 2; simpl; auto.
    
    rewrite Exists2_exists_left in G3.
    destruct G3 as (_ & t & ? & ?).
    right; constructor 1 with t; auto.
    
(* case where lt is NOT empty, ie NOT a leaf *)
    
    intros n Hn.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in Hm0.
    destruct Hm0 as [ Hm1 Hm2 ].
    rewrite Hn in Hm1.
    destruct (le_lt_dec (S n) k) as [ Hnk | Hnk ].
    2: rewrite HP3 in Hm1; tauto.
    assert (S n < S k) as Hnk'.
      omega.
    generalize (HR (S n)); intros HRSn.
    
    case_eq (w (S n)); intros Hs0; 
      rewrite Hs0 in HRSn; simpl in HRSn.

  (* s (length lt) = None, then P (S n) is empty *)
        
    contradict Hm1; apply HRSn.
    
  (* w (length lt) = af_full, then R (S n) is full on P (S n) *)
    
    set (P' := @ktf_prop4 _ _ P _ Hnk').
    assert (forall j, k < j -> P' j = fun _ => False) as HP3'.
      intros j Hj.
      unfold P'.
      rewrite ktf_prop4_neq; try omega.
      rewrite HP3; auto.
    assert (forall j x, k < j -> ~ P' j x) as HP3''.
      intros i x Hi.
      rewrite HP3'; auto.

    set (R' := @ktf_rel4 X _ _ Hnk' lt R).
    assert (afs (P' n) (R' n)) as G0.
      unfold P', R'.
      rewrite ktf_rel4_eq_1; auto.
      apply afs_ktf_rel4_i; auto.
      intros t Ht. 
      apply afs_af_rel_restr_eq.
      specialize (IHt t Ht).
      apply afs_af_rel_restr_eq in IHt.
      revert IHt; apply af_eq.
      apply litf_rel_restr_eq; auto.
      apply eq2_sym.
      apply emb_tree_higman_emb_tree_upto_eq.
      intros; rewrite HP3; auto.
    set (w' j := match (ktf_pos4 Hnk' j) with
                   | inleft (inleft (left _))   => w j
                   | inleft (inleft (right _))  => af_rlift
                   | inleft (inright _)         => af_empty 
                   | inright _                  => w j
                 end).
                 
    specialize (@IHs w' P' R'); spec_all IHs; auto.
    
    unfold w', P', R'.
    intros j.
    destruct (ktf_pos4 Hnk' j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_0; try omega. 
    auto.
    subst j; apply G0.
    subst j.
    rewrite ktf_prop4_eqS; try omega.
    simpl; auto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_3; try omega.
    auto.

    exists (S n); split; auto.
    unfold w', P', R'; split.
    destruct (ktf_pos4 Hnk' (S n)) as [[[|]|]|]; try omega; auto.
    rewrite Hs0; apply lt_afsr_empty_full.
    intros j Hj.
    destruct (ktf_pos4 Hnk' j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop4_neq; try omega.
    rewrite ktf_rel4_eq_3; try omega.
    auto.
    
    apply ktf_prop4_u; auto.
    apply ktf_prop4_dec; auto. 
    
    (* now IHs is proof that emb_tree_higman is af on wfptree P' *)
    
    apply afs_af_rel_restr_eq in IHs.
    generalize (emb_tree_higman_emb_tree_upto_eq P' R' HP3''); intros H.
    apply af_eq, proj1 in H; auto.
    apply ktf_ev4_af with (H_m_lt := H_m_lt) in H; auto.
    apply afs_af_rel_restr_eq.
    revert H; apply af_eq.
    apply eq2_trans with (1 := restr_lift_rel_eq _ (exist _ _ H_m_lt)); simpl.
    apply litf_rel_restr_eq; auto.
    apply emb_tree_higman_emb_tree_upto_eq; auto.
    intros; rewrite HP3; auto.
    
  (* s (length lt) = Some (in_wft_sup f), then R rlift x is secured by f x *)

    set (P' := @ktf_prop3 _ P n).
    assert (forall j, k < j -> P' j = fun _ => False) as HP3'.
      intros j Hj.
      unfold P'.
      rewrite ktf_prop3_neq; try omega.
      rewrite HP3; auto.
    assert (forall j x, k < j -> ~ P' j x) as HP3''.
      intros i x Hi.
      rewrite HP3'; auto.    
   set (R' := @ktf_rel3 _ _ _ Hnk' m lt R).
   assert (afs (P' n) (R' n)) as G0.
      unfold P', R'.
      rewrite ktf_rel3_eq_1; auto.
      apply afs_ktf_rel3_i; auto.
      intros t Ht.
      apply afs_af_rel_restr_eq.
      specialize (IHt t Ht).
      apply afs_af_rel_restr_eq in IHt.
      revert IHt; apply af_eq.
      apply litf_rel_restr_eq; auto.
      apply eq2_sym.
      apply emb_tree_higman_emb_tree_upto_eq.
      intros; rewrite HP3; auto.
    set (w' j := match (ktf_pos3 Hnk' j) with
                   | inleft (inleft (left _))   => w j
                   | inleft (inleft (right _))  => af_rlift
                   | inleft (inright _)         => af_rlift
                   | inright _                  => w j
                 end).
    specialize (IHs w' P' R').
    spec_all IHs; auto.

    unfold w', P', R'.
    intros j.
    destruct (ktf_pos3 Hnk' j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_0; try omega. 
    auto.
    subst j; apply G0.
    subst j.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_2; try omega.
    simpl; apply afs_inv; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_3; try omega.
    auto.

    exists (S n); split; auto.
    unfold w', P', R'; split.
    destruct (ktf_pos3 Hnk' (S n)) as [[[|]|]|]; try omega; auto.
    rewrite Hs0.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_2; try omega.
    apply lt_afsr_rlift; auto.
    intros j Hj.
    destruct (ktf_pos3 Hnk' j) as [[[|]|]|]; try omega; auto.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_3; try omega.
    auto.
  
    apply ktf_prop3_u; auto.
    apply ktf_prop3_dec; auto.
    
    (* now IHs is proof that emb_tree_higman is af on wfptree P' *)
    
    apply afs_af_rel_restr_eq in IHs.
    generalize (emb_tree_higman_emb_tree_upto_eq P' R' HP3''); intros H.
    apply af_eq, proj1 in H; auto.
    apply ktf_ev3_af with (H_m_lt := H_m_lt) in H; auto.
    apply afs_af_rel_restr_eq.
    revert H; apply af_eq.
    apply eq2_trans with (1 := restr_lift_rel_eq _ (exist _ _ H_m_lt)); simpl.
    apply litf_rel_restr_eq; auto.
    apply emb_tree_higman_emb_tree_upto_eq; auto.
    intros; rewrite HP3; auto.
  Qed.
  
End higman_ktree_af_rec.

Section higman_wfptree_afs.
  
  Variable (X : Type) (k : nat) (P : nat -> X -> Prop) (R : nat -> X -> X -> Prop).
  Hypothesis HP1 : forall n x, { P n x } + { ~ P n x }. 
  Hypothesis HP2 : forall n x, k <= n -> ~ P n x.
  Hypothesis HR  : forall n, n < k -> afs (P n) (R n).
  
  Theorem higman_wfptree_afs : afs (wfptree P) (emb_tree_higman R).
  Proof.
    destruct k as [ | k' ].
    
    apply in_afs_0.
    intros [ x ll ] y H.
    rewrite wfptree_fix in H.
    apply proj1, HP2 in H.
    destruct H.
    omega.
 
    set (P' n := match le_lt_dec n k' with
                   | left _  => fun t => match t with
                                           | in_ktf_0 x => P n x
                                           | _          => False
                                         end
                   | right _ => fun _ => False
                 end).

    assert (forall n x, { P' n x } + { ~ P' n x }) as P'dec.
      intros n x; unfold P'.
      destruct (le_lt_dec n k'); auto.
      destruct x; auto.      
    set (R' n (t1 t2 : ktree_fix X) := 
      match t1, t2 with
        | in_ktf_0 u , in_ktf_0 v => R n u v
        | _          , _          => False
      end).
    assert (forall n, afs (P' n) (R' n)) as Hs.
      intros n.
      unfold P', R'.
      destruct (le_lt_dec n k') as [ H | ].
      apply le_n_S in H.
      generalize (@HR n H).
      apply afs_relmap with (f := fun x y => y = in_ktf_0 x).
      intros [ | x | | ]; try tauto; exists x; auto.
      intros x1 x2 [ | y1 | | ] [ | y2 | | ]; try tauto.
      intros ? ? ? ? H1 H2; injection H1; injection H2; intros ? ?; subst; auto.
      constructor 1; tauto.

    generalize (@higman_ktree_af_rec _ k' P' R').
    intros Hig; spec_all Hig; auto.
    
    intros n; unfold P'.
    destruct (le_lt_dec n k'); auto.
    
    intros n Hn; unfold P'.
    destruct (le_lt_dec n k'); auto; omega.
    
    revert Hig.
    apply afs_relmap with (f := fun x y => x = tree_map (@in_ktf_0 _) y); auto.

    intros y Hy.
    exists (tree_map (@in_ktf_0 _) y); split; auto.

    revert Hy.
    induction 1 as [ a ll Hll IH ] using wfptree_rect.
    rewrite wfptree_fix in Hll; destruct Hll as [ H1 H2 ].
    rewrite tree_map_fix.
    rewrite wfptree_fix; split.
    rewrite map_length; unfold P'; simpl.
    destruct (le_lt_dec (length ll) k'); auto.
    apply HP2 in H1; auto.
    intros t Ht.
    apply in_map_iff in Ht.
    destruct Ht as (u & ? & Ht); subst t.
    revert Ht; apply In_t_P; auto.
 
    intros x1 x2 y1 y2 _ _ _ _ G1 G2 H.
    revert H y1 y2 G1 G2.    
    induction 1 as [ x1 t a ll G1 G2 G3 
                   | a b ll mm G1 G2 G3 ].

    intros y1 [ c kk ] G4 G5.
    rewrite tree_map_fix in G5.
    injection G5; clear G5; intros ? ?; subst ll a.
    rewrite in_map_iff in G1.
    destruct G1 as (u & G1 & G5); subst t.
    constructor 1 with u; auto.
    
    intros [ c kk ] [ d pp ] G4 G5.
    rewrite tree_map_fix in G4.
    rewrite tree_map_fix in G5.
    injection G4; injection G5.
    clear G4 G5; intros ? ? ? ?; subst.
    rewrite map_length in G1; red in G1.
    constructor 2; auto.
    rewrite Forall2_map_both in G3.
    revert G3.
    apply Forall2_impl.
    intros x y H; apply H; auto.
  Qed.

End higman_wfptree_afs.

