(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Min.
Require Import List.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import subword.

Require Import wft.
Require Import af_t.
Require Import afs_t.

Require Import tree.
Require Import tree_embeddings.

Require Import tree_kruskal_fix.
Require Import tree_kruskal_ev3.
Require Import tree_kruskal_ev4.

Set Implicit Arguments.

Section higman_ktree_rec.

  Variable X : Type.
  Notation T := (ktree_fix X).

  Variable k : nat.

  Theorem higman_ktree_rec s :
                     forall P, 
                     (forall n, ~ P n (@in_ktf_u X))  
                  -> (forall n x, { P n x } + { ~ P n x })
                  -> (forall n, k < n -> P n = fun _ => False)
                  -> forall R,
                     (forall n, n <= k -> afs_owft_sec (s n) (P n) (R n))
                  -> afs_t (wfptree P) (emb_tree_higman R).
  Proof.
(* the outermost induction is over (k+1) sequences of owft's with the easier
   well founded relation *)

    induction s as [ s IHs ] using (@owft_easier_rect _ (S k)).
    intros P HP1 HP2 HP3 R HR.
    assert (forall t, { wfptree P t } + { ~ wfptree P t }) as HP4.
      apply wfptree_dec, HP2.

(* now starts the nested induction over the well-formed tree *)

    apply in_afs_t1; auto.    
    induction 1 as [ m lt H_m_lt IHt ] using wfptree_rect.
    
    case_eq (length lt).

(* case where lt is empty, ie we have a leaf *)
    
    intros Hl_lt.
    destruct lt.
    2: discriminate Hl_lt.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in H_m_lt.
    destruct H_m_lt as [ Hm1 _ ]; simpl in Hm1.
    generalize (HR 0); intros HR0.

    case_eq (s 0); [ intros [ | f ] | ]; intros Hs0; 
      rewrite Hs0 in HR0; simpl in HR0.
    
    Focus 3.
    (* s 0 = None, then P 0 is empty *)
    
    contradict Hm1; apply HR0; omega.

    (* s 0 = Some in_wft_zt, then R 0 is full on P 0 *)
    
    apply in_afs_t0.
    intros x y Hx Hy; right.
    clear y Hy.
    destruct (tree_leaf x) as (l & Hl).
    unfold wfptree in Hx.
    rewrite tree_fall_sub_tree in Hx.    
    generalize Hl.
    apply emb_sub_tree_higman.
    constructor 2; simpl.
    apply HR0; auto; try omega.
    apply Hx with (1 := Hl).
    constructor.
    
    (* s 0 = Some (in_wft_sup f), then R rlift x is secured by f x *)
    
    set (s' n := match n with 0 => Some (f m) | S n => s (S n) end).
    set (R' n := match n with 0 => R 0 rlift m | S n => R (S n) end).
    specialize (IHs s'); spec_all IHs.
    rewrite owft_easier_eq.
    exists 0.
    repeat split; auto.
    omega.
    unfold s'; rewrite Hs0; apply in_wft_lt.
    intros [ | j ]; unfold s'; auto; omega.
    specialize (IHs P); spec_all IHs; auto.
    specialize (IHs R'); spec_all IHs; auto.
    intros [ | n ]; unfold s', R'; auto.
    intros; apply HR0; auto.

    revert IHs.   
    apply afs_t_incl.
    intros; auto.
    intros t1 t2.
    induction 1 as [ t1 t2 x ll G1 G2 [ G3 | G3 ] 
                   | x y ll mm G1 G2 G3].
    left; constructor 1 with t2; auto.
    right; auto.
    apply Forall2_decide_ind in G3.
    destruct G3 as [ G3 | G3 ].

    destruct ll; simpl in G1.
    destruct G1 as [ G1 | G1 ].
    left; constructor 2; auto.
    right; constructor 2; auto.
    left; constructor 2; simpl; auto.
    
    rewrite Exists2_exists_left in G3.
    destruct G3 as (_ & t & ? & ?).
    right; constructor 1 with t; auto.
    
(* case where lt is NOT empty, ie NOT a leaf *)
    
    intros n Hn.
    generalize H_m_lt.
    intros Hm0.
    rewrite wfptree_fix in Hm0.
    destruct Hm0 as [ Hm1 Hm2 ].
    rewrite Hn in Hm1.
    destruct (le_lt_dec (S n) k) as [ Hnk | Hnk ].
    2: rewrite HP3 in Hm1; tauto.
    generalize (HR _ Hnk); intros HRSn.
    assert (S n < S k) as Hnk'.
      omega.
    
    case_eq (s (S n)); [ intros [ | f ] | ]; intros Hs0; 
      rewrite Hs0 in HRSn; simpl in HRSn.
    Focus 3.

  (* s (length lt) = None, then P (S n) is empty *)
        
    contradict Hm1; apply HRSn.
    
  (* s (length lt) = Some in_wft_zt, then R (S n) is full on P (S n) *)
    
    set (P' := @ktf_prop4 _ _ P _ Hnk').
    assert (HP'1 : forall n x, { P' n x } + { ~ P' n x }).
      apply ktf_prop4_dec; auto.
    assert (HP'2 : forall t, { wfptree P' t } + { ~ wfptree P' t }).
      apply wfptree_dec; auto.
    assert (HP'3 : forall n x, k < n -> ~ P' n x).
      intros i x Hi.
      unfold P'.
      destruct (eq_nat_dec i (S n)).
      rewrite ktf_prop4_eqS; auto.
      rewrite ktf_prop4_neq; try omega.
      rewrite HP3; auto.
    
    set (R' := @ktf_rel4 X _ _ Hnk' lt R).
    assert (afs_t (P' n) (R' n)) as Hs'n.
      unfold P', R'.
      rewrite ktf_rel4_eq_1; auto.
      apply afs_t_ktf_rel4_i; auto.
      specialize (HR n); spec_all HR.
      omega.
      revert HR; apply afs_owft_afs_t.
      intros t Ht.
      apply af_t_impl_afs_t; auto.
      specialize (IHt t Ht).
      apply In_t_In in Ht.
      apply afs_t_impl_af_t in IHt; auto.
      revert IHt; apply af_t_eq.
      apply litf_rel_restr_eq; auto.
      apply emb_tree_higman_emb_tree_upto_eq.
      intros; rewrite HP3; auto.
    apply afs_t_afs_owft in Hs'n.
    destruct Hs'n as (s'n & Hs'n).
    set (s' i := if eq_nat_dec i (S n) then None else
                 if eq_nat_dec i n then s'n
                 else s i).

    specialize (IHs s'); spec_all IHs.
    
    rewrite owft_easier_eq.
    exists (S n).
    repeat split; auto.
    unfold s'; rewrite Hs0.
    destruct (eq_nat_dec (S n) (S n)); try omega.
    red; constructor.
    intros j Hj; unfold s'.
    destruct (eq_nat_dec j (S n)); try omega.
    destruct (eq_nat_dec j n); try omega; auto.

    specialize (IHs P'); spec_all IHs; auto.
    
    apply ktf_prop4_u; auto.
    unfold P'; intros j Hj; rewrite ktf_prop4_neq; try omega.
    apply HP3; auto.
    
    specialize (IHs R'); spec_all IHs.

    intros i Hi.
    unfold s'.
    destruct (eq_nat_dec i (S n)).
    subst i; red.
    unfold P'; intros x; rewrite ktf_prop4_eqS; try omega.
    destruct (eq_nat_dec i n); auto.
    subst; auto.
    unfold P'; rewrite ktf_prop4_neq; auto.
    unfold R'.
    destruct (le_lt_dec i (S n)).
    rewrite ktf_rel4_eq_0; auto; omega.
    rewrite ktf_rel4_eq_3; auto.
    
    (* now IHs is proof that emb_tree_higman is af on wfptree P' *)
    
    apply afs_t_impl_af_t in IHs.
    generalize (emb_tree_higman_emb_tree_upto_eq P' R' HP'3); intros H.
    apply af_t_eq in H; auto.
    apply ktf_ev4_af_t with (H_m_lt := H_m_lt) in H; auto.
    apply af_t_impl_afs_t; auto.
    revert H; apply af_t_eq.
    apply eq2_trans with (1 := restr_lift_rel_eq _ (exist _ _ H_m_lt)); simpl.
    apply litf_rel_restr_eq; auto.
    apply eq2_sym, emb_tree_higman_emb_tree_upto_eq; auto.
    intros; rewrite HP3; auto.
    
  (* s (length lt) = Some (in_wft_sup f), then R rlift x is secured by f x *)
    
    set (P' := @ktf_prop3 _ P n).
    assert (HP'1 : forall n x, { P' n x } + { ~ P' n x }).
      apply ktf_prop3_dec; auto.
    assert (HP'2 : forall t, { wfptree P' t } + { ~ wfptree P' t }).
      apply wfptree_dec; auto.
    assert (HP'3 : forall n x, k < n -> ~ P' n x).
      intros i x Hi.
      unfold P'.
      rewrite ktf_prop3_neq; try omega.
      rewrite HP3; auto.
   
    set (R' := @ktf_rel3 _ _ _ Hnk' m lt R).
    assert (afs_t (P' n) (R' n)) as Hs'n.
      unfold P', R'.
      rewrite ktf_rel3_eq_1; auto.
      apply afs_t_ktf_rel3_i; auto.
      specialize (HR n); spec_all HR; try omega.
      revert HR; apply afs_owft_afs_t.
      specialize (HR (S n)); spec_all HR; try omega.
      revert HR; apply afs_owft_afs_t.
      intros t Ht.
      apply af_t_impl_afs_t; auto.
      specialize (IHt t Ht).
      apply In_t_In in Ht.
      apply afs_t_impl_af_t in IHt; auto.
      revert IHt; apply af_t_eq.
      apply litf_rel_restr_eq; auto.
      apply emb_tree_higman_emb_tree_upto_eq.
      intros; rewrite HP3; auto.
    apply afs_t_afs_owft in Hs'n.
    destruct Hs'n as (s'n & Hs'n).
    set (s' i := if eq_nat_dec i (S n) then Some (f m) else
                 if eq_nat_dec i n then s'n
                 else s i).

    specialize (IHs s'); spec_all IHs.
    
    rewrite owft_easier_eq.
    exists (S n).
    repeat split; auto.
    unfold s'; rewrite Hs0.
    destruct (eq_nat_dec (S n) (S n)); try omega.
    apply in_wft_lt.
    intros j Hj; unfold s'.
    destruct (eq_nat_dec j (S n)); try omega.
    destruct (eq_nat_dec j n); try omega; auto.

    specialize (IHs P'); spec_all IHs; auto.
    
    apply ktf_prop3_u; auto.
    unfold P'; intros j Hj; rewrite ktf_prop3_neq; try omega.
    apply HP3; auto.
    
    specialize (IHs R'); spec_all IHs.

    intros i Hi.
    unfold s'.
    destruct (eq_nat_dec i (S n)).
    subst i; red.
    unfold P', R'.
    rewrite ktf_prop3_neq; try omega.
    rewrite ktf_rel3_eq_2; auto.
    
    destruct (eq_nat_dec i n); auto.
    subst; auto.
    unfold P'; rewrite ktf_prop3_neq; auto.
    unfold R'; destruct (le_lt_dec i (S n)).
    rewrite ktf_rel3_eq_0; auto; omega.
    rewrite ktf_rel3_eq_3; auto.
    
    (* now IHs is proof that emb_tree_upto k R' is af on wfptree P' *)
    
    apply afs_t_impl_af_t in IHs.
    generalize (emb_tree_higman_emb_tree_upto_eq P' R' HP'3); intros H.
    apply af_t_eq in H; auto.
    apply ktf_ev3_af_t with (H_m_lt := H_m_lt) in H; auto.
    apply af_t_impl_afs_t; auto.
    revert H; apply af_t_eq.
    apply eq2_trans with (1 := restr_lift_rel_eq _ (exist _ _ H_m_lt)); simpl.
    apply litf_rel_restr_eq; auto.
    apply eq2_sym, emb_tree_higman_emb_tree_upto_eq; auto.
    intros; rewrite HP3; auto.  
  Qed.
  
End higman_ktree_rec.

Section higman_wfptree.
  
  Variable (X : Type) (k : nat) (P : nat -> X -> Prop) (R : nat -> X -> X -> Prop).
  Hypothesis HP1 : forall n x, { P n x } + { ~ P n x }. 
  Hypothesis HP2 : forall n x, k <= n -> ~ P n x.
  Hypothesis HR  : forall n, n < k -> afs_t (P n) (R n).
  
  Theorem higman_wfptree : afs_t (wfptree P) (emb_tree_higman R).
  Proof.
    destruct k as [ | k' ].
    
    apply in_afs_t0.
    intros [ x ll ] y H.
    rewrite wfptree_fix in H.
    apply proj1, HP2 in H.
    destruct H.
    omega.
 
    set (P' n := match le_lt_dec n k' with
                   | left _  => fun t => match t with
                                           | in_ktf_0 x => P n x
                                           | _          => False
                                         end
                   | right _ => fun _ => False
                 end).

    assert (forall n x, { P' n x } + { ~ P' n x }) as P'dec.
      intros n x; unfold P'.
      destruct (le_lt_dec n k'); auto.
      destruct x; auto.      
    set (R' n (t1 t2 : ktree_fix X) := 
      match t1, t2 with
        | in_ktf_0 u , in_ktf_0 v => R n u v
        | _          , _          => False
      end).
    assert (forall n, n < S k' -> { t | afs_owft_sec t (P' n) (R' n) }) as Hs.
      intros n Hn.
      apply afs_t_afs_owft.
      generalize (HR Hn).
      apply afs_t_relmap with (f := fun x y => y = in_ktf_0 x).
      auto.
      unfold P'.
      destruct (le_lt_dec n k'); try tauto.
      intros [ | x | | ] H; try tauto; exists x; auto.
      intros; subst; auto.
    set (s' n := match le_lt_dec n k' with 
                   | right _ => None
                   | left H  => proj1_sig (Hs _ (le_n_S _ _ H))
                 end).                                            
    generalize (@higman_ktree_rec _ k' s' P').

    intros Hig; spec_all Hig; auto.
    
    intros n; unfold P'; destruct (le_lt_dec n k'); auto.
    intros n Hn; unfold P'; destruct (le_lt_dec n k'); auto; omega.
    
    specialize (Hig R'); spec_all Hig; auto.
    
    unfold s'.
    intros n Hn.
    destruct (le_lt_dec n k') as [ H | ]; try omega.
    apply (proj2_sig (Hs _ (le_n_S _ _ H))).
    
    revert Hig.
    apply afs_t_relmap with (f := fun x y => x = tree_map (@in_ktf_0 _) y); auto.
    apply wfptree_dec; auto.
    intros y Hy.
    exists (tree_map (@in_ktf_0 _) y); split; auto.

    revert Hy.
    induction 1 as [ a ll Hll IH ] using wfptree_rect.
    rewrite wfptree_fix in Hll; destruct Hll as [ H1 H2 ].
    rewrite tree_map_fix.
    rewrite wfptree_fix; split.
    rewrite map_length; unfold P'; simpl.
    destruct (le_lt_dec (length ll) k'); auto.
    apply HP2 in H1; auto.
    intros t Ht.
    apply in_map_iff in Ht.
    destruct Ht as (u & ? & Ht); subst t.
    revert Ht; apply In_t_P; auto.
 
    intros x1 x2 y1 y2 _ _ _ _ G1 G2 H.
    revert H y1 y2 G1 G2.    
    induction 1 as [ x1 t a ll G1 G2 G3 
                   | a b ll mm G1 G2 G3 ].

    intros y1 [ c kk ] G4 G5.
    rewrite tree_map_fix in G5.
    injection G5; clear G5; intros ? ?; subst ll a.
    rewrite in_map_iff in G1.
    destruct G1 as (u & G1 & G5); subst t.
    constructor 1 with u; auto.
    
    intros [ c kk ] [ d pp ] G4 G5.
    rewrite tree_map_fix in G4.
    rewrite tree_map_fix in G5.
    injection G4; injection G5.
    clear G4 G5; intros ? ? ? ?; subst.
    rewrite map_length in G1; red in G1.
    constructor 2; auto.
    rewrite Forall2_map_both in G3.
    revert G3.
    apply Forall2_impl.
    intros x y H; apply H; auto.
  Qed.

End higman_wfptree.

