(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.
Require Import list_fan.
Require Import sublist.
Require Import subword.
Require Import finite.

Set Implicit Arguments.

Section list_splits.

  Variables (X : Type).
  
  Implicit Types (l : list X).

  Fixpoint list_splits l := 
    match l with 
      | nil =>  ((nil,nil)::nil)
      | x::l => (nil,x::l)::map (fun (c : list X * list X) => let (u,v) := c in (x::u,v)) (list_splits l) 
    end.
    
  Fact list_splits_spec l x y : In (x,y) (list_splits l) <-> x++y = l.
  Proof.
    revert x y.
    induction l as [ | a l IH ]; intros x y; simpl; split.
    intros [ H | [] ]; injection H; intros; subst; auto.
    destruct x; destruct y; try discriminate 1; tauto.
    intros [ H | H ].
    injection H; intros; subst; auto.
    apply in_map_iff in H.
    destruct H as ((u,v) & H1 & H2).
    injection H1; clear H1; intros ? ?; subst x y.
    rewrite IH in H2.
    rewrite <- H2; auto.
    intros H.
    destruct x as [ | a' x ].
    simpl in H; subst y; left; auto.
    injection H; clear H; intros ? ?; subst a' l; right.
    apply in_map_iff.
    exists (x,y).
    rewrite IH; auto.
  Qed.

  Definition lmr l := flat_map (fun c => match c with (_,nil) => nil | (l,x::r) => (l,(x,r))::nil end) (list_splits l).

  Fact lmr_spec ll w : In w (lmr ll) <-> exists l x r, ll = l++x::r /\ w = (l,(x,r)).
  Proof.
    unfold lmr; rewrite in_flat_map.
    split.

    intros ((l,[ | x r ]) & H1 & H2).
    destruct H2.
    destruct H2 as [ H2 | [] ]; subst.
    apply list_splits_spec in H1; subst.
    exists l, x, r; auto.
    
    intros (l & x & r & H1 & H2); subst.
    exists (l,x::r); split; simpl; auto.
    rewrite list_splits_spec; auto.
  Qed.
 
  Let list_msplits k l : { ll | forall m, In m ll <-> flat_map (fun x => x) m = l /\ length m = S k }.
  Proof.
    revert l; induction k as [ | k IH ]; intros l.
    
    exists ((l::nil)::nil); split.
    intros [ ? | [] ]; subst; simpl; rewrite <- app_nil_end; auto.
    intros (H1 & H2).
    destruct m as [ | a [ | ] ]; try discriminate H2; simpl in H1.
    rewrite <- app_nil_end in H1; subst a; left; auto.
    
    set (h l := proj1_sig (IH l)).
    assert (Hh : forall l m, In m (h l) <-> flat_map (fun x => x) m = l /\ length m = S k).
      intros u; apply (proj2_sig (IH u)).
    generalize h Hh; clear h Hh IH; intros h Hh.
    
    set (g (c : list X * list X) := let (w1,w2) := c in map (fun d => w1::d) (h w2)).
    
    exists (flat_map g (list_splits l)).
    intros m; split.
    
    rewrite in_flat_map.
    intros ((u,v) & H1 & H2).
    unfold g in H2.
    rewrite list_splits_spec in H1.
    rewrite in_map_iff in H2.
    destruct H2 as (ll & H2 & H3).
    rewrite Hh in H3.
    destruct H3 as (H3 & H4).
    subst m l v; split; simpl; auto.
    
    intros (H1 & H2). 
    rewrite in_flat_map.
    destruct m as [ | a m ]; try discriminate H2.
    simpl in H2; injection H2; clear H2; intros H2.
    simpl in H1; subst l.
    exists (a, flat_map (fun x => x) m); split.
    rewrite list_splits_spec; auto.
    unfold g.
    rewrite in_map_iff.
    exists m; split; auto.
    rewrite Hh; split; auto.
  Qed.
  
  Definition list_multi_splits k l := proj1_sig (list_msplits k l).
  
  Fact list_multi_splits_spec k l mm : In mm (list_multi_splits k l) <-> flat_map (fun x => x) mm = l /\ length mm = S k.
  Proof.
    apply (proj2_sig (list_msplits k l)).
  Qed.

End list_splits.

Section intercalate.

  Variable (X : Type).

  Implicit Types (w : list X) (lw : list (list X)).

  Fixpoint intercalate w lw :=
    match w, lw with
      | nil  ,  lw   => flat_map (fun x => x) lw
      | w    , nil   => w
      | x::w , u::lw => u++x::intercalate w lw
    end.

  Fact intercalate_nil w : intercalate w nil = w.
  Proof.
    destruct w; auto.
  Qed.

  Fact intercalate_sublist w lw : w <sl intercalate w lw.
  Proof.
    revert lw; induction w as [ | x w IH ]; intros [ | v lw ]; simpl.
    constructor 1.
    constructor 1.
    apply sl_refl.
    apply sl_trans with (2 := sl_app_left _ _).
    constructor 2; auto.
  Qed.

  Fact intercalate_length w lw : length w <= length (intercalate w lw).
  Proof.
    apply sublist_length, intercalate_sublist.
  Qed.
  
  Fact intercalate_length_eq w lw : length (intercalate w lw) = length w + lsum (map (@length _) lw).
  Proof.
    revert lw; induction w as [ | x w IH ]; intros [ | v lw ]; simpl; auto.
    rewrite app_length, flat_map_length; auto.
    rewrite app_length; simpl.
    rewrite IH; omega.
  Qed.

  Fact intercalate_nils w mm : Forall (eq nil) mm <-> intercalate w mm = w.
  Proof.
    split.

    intros H; revert w.
    induction H as [ | n mm H1 H2 IH ]; intros w.
    apply intercalate_nil.
    subst n.
    destruct w as [ | x w ]; simpl.
    clear IH.
    induction H2; subst; simpl; auto.
    rewrite IH; auto.
    
    revert w.
    induction mm as [ | x mm ].
    constructor.
    intros [ | a w ]; simpl.
    destruct x as [ | ]; try discriminate 1; simpl.
    intros H; constructor; auto.
    apply IHmm with (w := nil); auto.
    destruct x as [ | ]; simpl.
    intros H; injection H; clear H; intros H; constructor; auto.
    apply IHmm with (1 := H).
    intros; exfalso.
    apply f_equal with (f := @length _) in H.
    simpl in H; rewrite app_length in H; simpl in H.
    generalize (intercalate_length w mm).
    omega.   
  Qed.

  Fact intercalate_nil_inv w lw : intercalate w lw = nil -> w = nil /\ Forall (eq nil) lw.
  Proof.
    destruct w as [ | x w ].
    rewrite <- intercalate_nils; auto.
    destruct lw as [ | [ | ] ?]; discriminate 1.
  Qed.

  Fact flat_map_nil U V l : flat_map (fun _ : U => @nil V) l = nil.
  Proof. induction l; simpl; auto. Qed.

  Fact intercalate_map_nil U (mm : list X)  (ll : list U) : mm = intercalate mm (map (fun _ => nil) ll).
  Proof.
    revert ll.
    induction mm as [ | x u IHu ]; intros [ | y v ]; simpl; auto.
    rewrite flat_map_map, flat_map_nil; auto.
    f_equal; apply IHu.
  Qed.
  
  Fact intercalate_head_map_nil U w1 w2 (lw : list U) : intercalate w1 (w2::map (fun _ => nil) lw) = w2++w1.
  Proof.
    destruct w1; simpl.
    rewrite flat_map_map, flat_map_nil; auto.
    rewrite <- intercalate_map_nil; auto.
  Qed.

  Fact sublist_intercalate w1 w2 : w1 <sl w2 -> exists lw, w2 = intercalate w1 lw.
  Proof.
    induction 1 as [ | a w1 w2 _ (lw & H) | a w1 w2 H1 (lw & H2) ].
    exists (ll::nil); apply app_nil_end.
    exists (nil::lw); simpl; subst; auto.
    destruct lw as [ | x lw ].
    rewrite intercalate_nil in H2; subst.
    exists ((a::nil)::nil); destruct w1; simpl;auto.
    rewrite intercalate_nil; auto.
    exists ((a::x)::lw); subst.
    destruct w1; simpl; auto.
  Qed.

  (* intercalate can reverse the sup-list relation *)

  Fact intercalate_spec w1 w2 : w1 <sl w2 <-> exists lw, w2 = intercalate w1 lw.
  Proof.
    split.
    apply sublist_intercalate.
    intros (lw & ?); subst; apply intercalate_sublist.
  Qed.

  Fact In_t_intercalate w lw x : In_t x (intercalate w lw) 
                             -> (In_t x w 
                              + { w' : _ &  In_t x w' * In_t w' lw })%type.
  Proof.
    revert lw x; induction w as [ | a w IH ]; intros [ | w1 lw ] x; simpl.
    tauto.
    intros H; right; apply In_t_app_plus in H.
    destruct H as [ H | H ].
    exists w1; tauto.
    apply In_t_flat_map in H.
    destruct H as (y & H1 & H2).
    exists y; tauto.
    tauto.
    intros H; apply In_t_app_plus in H.
    destruct H as [ H | [ H | H ] ].
    right; exists w1; tauto.
    tauto.
    apply IH in H.
    destruct H as [ H | (w' & H) ].
    tauto.
    right; exists w'; tauto.
  Qed.
  
  Fact intercalate_In_t w lw x : (In_t x w + { w' : _ &  In_t x w' * In_t w' lw })%type 
                               -> In_t x (intercalate w lw).
  Proof.
    intros [ H | (w' & H1 & H2) ].
    
    revert lw x H; induction w as [ | a w IH ]; intros [ | w1 lw ] x; simpl; try tauto.
    intros [ H | H ].
    subst; apply In_t_plus_app; right; left; auto.
    apply In_t_plus_app; right; right; apply IH; auto.
    
    revert lw x w' H1 H2; induction w as [ | a w IH ]; intros [ | w1 lw ] x w'; simpl; try tauto;
    intros H1 [ H2 | H2 ].    
    subst; apply In_t_plus_app; left; auto.
    apply In_t_plus_app; right; apply flat_map_In_t with (y := w'); auto.
    subst; apply In_t_plus_app; left; auto.
    apply In_t_plus_app; right; right; apply IH with w'; auto.
  Qed.

  Fact intercalate_In_spec w lw x : In x (intercalate w lw) <-> In x w \/ exists w', In x w' /\ In w' lw.
  Proof.
    split.
    
    apply In_t_P.
    intros H; apply In_t_intercalate in H.
    destruct H as [ H | (u & H1 & H2) ].
    left; apply In_t_In; auto.
    right; exists u; split; apply In_t_In; auto.
    
    intros [ H | (u & H1 & H2) ].
    revert H; apply In_t_P; intros H; apply In_t_In, intercalate_In_t; tauto.
    revert H1; apply In_t_P; intros H1.
    revert H2; apply In_t_P; intros H2.
    apply In_t_In, intercalate_In_t; right; exists u; auto.
  Qed.
 
  Section intercalate_inv.

    Let intercalate_inv_rec k ly : { ll | forall w lw, In (w,lw) ll <-> intercalate w lw = ly /\ length lw = k }.
    Proof.
      revert ly; induction k as [ | k IHk ].

      intros ly; exists ((ly,nil)::nil); intros w lw; split.
      intros [ H | [] ]; injection H; clear H; intros ? ?; subst; split; auto.
      apply intercalate_nil.
      intros (H1 & H2).
      destruct lw; try discriminate H2.
      rewrite intercalate_nil in H1.
      left; f_equal; auto.
      
      set (h ly := proj1_sig (IHk ly)).
      assert (Hh : forall ly w lw, In (w,lw) (h ly) <-> intercalate w lw = ly /\ length lw = k).
        intros ly; apply (proj2_sig (IHk ly)).
      generalize h Hh; clear h Hh IHk; intros h Hh.
      
      intros ly.
      
      set (f (c : list X * (X * list X)) := match c with (w1,(x,w2)) => map (fun d => (x::fst d,w1::snd d)) (h w2) end).
      
      exists (    map (fun x => (nil,x)) (list_multi_splits k ly)
              ++ flat_map f (lmr ly)).
      intros w lw; split.
 
      intros H; apply in_app_or in H; destruct H as [ H | H ]; revert H.
      
      rewrite in_map_iff.
      intros (ll & H1 & H2).
      injection H1; clear H1; intros ? ?; subst w ll; simpl.
      rewrite list_multi_splits_spec in H2; auto.

      rewrite in_flat_map.
      intros ((w1,(x,w2)) & H1 & H2); simpl in H2.
      rewrite lmr_spec in H1.
      destruct H1 as (lr & x0 & rr & ? & H1); subst.
      injection H1; clear H1; intros ? ? ?; subst lr x0 rr.
      apply in_map_iff in H2.
      destruct H2 as ((a, b) & H1 & H2); simpl in H1.
      injection H1; clear H1; intros ? ?; subst w lw.
      apply Hh in H2.
      destruct H2 as (H1 & H2).
      simpl; split; try omega.
      rewrite H1; auto.
      
      intros (H1 & H2).
      
      destruct lw as [ | w1 lw ]; try discriminate H2.
      simpl in H2; injection H2; clear H2; intros H2.
      destruct w as [ | x w ].
      
      simpl in H1.
      apply in_or_app; left.
      rewrite in_map_iff.
      exists (w1::lw); split; auto.
      rewrite list_multi_splits_spec; simpl; split; auto.
      
      simpl in H1.
      subst ly.
      apply in_or_app; right.
      rewrite in_flat_map.
      exists (w1,(x,intercalate w lw)); split.
      rewrite lmr_spec.
      exists w1, x, (intercalate w lw); auto.
      unfold f.
      rewrite in_map_iff.
      exists (w,lw); simpl; split; auto.
      rewrite Hh; auto.
    Qed.
    
    Definition intercalate_inv k l := proj1_sig (intercalate_inv_rec k l).
    
    Fact intercalate_inv_spec k l w lw : In (w,lw) (intercalate_inv k l) <-> intercalate w lw = l /\ length lw = k.
    Proof. apply (proj2_sig (intercalate_inv_rec k l)). Qed.

  End intercalate_inv.
  
  Definition intercalate_inv_lt k l := flat_map (fun i => intercalate_inv i l) (list_n k).
  
  Fact intercalate_inv_lt_spec k l w lw : In (w,lw) (intercalate_inv_lt k l) <-> intercalate w lw = l /\ length lw < k.
  Proof.
    unfold intercalate_inv_lt.
    rewrite in_flat_map.
    split.
    intros (i & H1 & H2).
    rewrite <- list_n_prop in H1.
    rewrite intercalate_inv_spec in H2.
    split; try tauto; omega.
    intros (H1 & H2).
    exists (length lw); split.
    rewrite <- list_n_prop; auto.
    rewrite intercalate_inv_spec; auto.
  Qed.

End intercalate.

Section intercalate_Forall2_subword.

  Variable (X : Type) (R : X -> X -> Prop).

  Fact Forall2_intercalate w1 w2 lw1 lw2 : Forall2 R w1 w2 
                                        -> Forall2 (Forall2 R) lw1 lw2
                                        -> Forall2 R (intercalate w1 lw1) (intercalate w2 lw2).
  Proof.
    intros H.
    revert H lw1 lw2.
    induction 1 as [ | x y w1 w2 H1 H2 H3 ]; intros lw1 lw2;
    induction 1 as [ | a b lw1 lw2 G1 G2 G3 ]; simpl.
    constructor 1.
    apply Forall2_app; auto.
    constructor 2; auto.
    apply Forall2_app; auto.
  Qed.

  Fact subword_intercalate w1 w2 lw1 lw2 : Forall2 R w1 w2 
                                        -> Forall2 (subword R) lw1 lw2
                                        -> subword R (intercalate w1 lw1) (intercalate w2 lw2).
  Proof.
    intros H.
    revert H lw1 lw2.
    induction 1 as [ | x y w1 w2 H1 H2 H3 ]; intros lw1 lw2;
    induction 1 as [ | a b lw1 lw2 G1 G2 G3 ]; simpl.
    constructor 1.
    apply subword_app_comp; auto.
    constructor 3; auto.
    apply subword_Forall2; auto.
    apply subword_app_comp; auto.
    constructor 3; auto.
  Qed.

End intercalate_Forall2_subword.

Section subword_not_intercalate.

  Variable (X : Type) (R : X -> X -> Prop).

  Fact subword_not_app w1 w w2 : 
        (forall x y, In x w1 -> In y w -> ~ R x y)
     -> subword R w1 (w++w2) -> subword R w1 w2.
  Proof.
    revert w w2.
    induction w1 as [ | x w1 IH ]; intros [ | y w ] w2 H; simpl; auto.
    intros; apply subword_nil_true.
    intros H1; apply subword_cons_inv in H1.
    destruct H1 as ([ | c l ] & b & r & H1 & H2 & H3).
    injection H1; clear H1; intros ? ?; subst.
    contradict H2; apply H; left; auto.
    injection H1; clear H1; intros H1 ?; subst.
    apply app_split in H1.
    destruct H1 as [ (m & H5 & H6) | (m & H5 & H6) ]; subst.
    apply subword_app; constructor 3; auto.
    destruct m as [ | d m ].
    simpl in H6; subst w2; constructor 3; auto.
    injection H6; clear H6; intros ? ?; subst.
    contradict H2; apply H.
    left; auto.
    right; apply in_or_app; right; left; auto.
  Qed.

  Fact subword_not_flat_map l lw : 
        Forall2 (fun x w => forall y, In y w -> ~ R x y) l lw 
     -> subword R l (flat_map (fun x => x) lw) -> length l = 0.
  Proof.
    induction 1 as [ | x w l lw H1 H2 H3 ]; simpl; auto.
    intros H; exfalso.
    apply subword_cons_inv in H.
    destruct H as (l1 & b & l2 & H5 & H6 & H7).
    apply app_split in H5.
    destruct H5 as [ (m & H8 & H9) | (m & H8 & H9) ].

    spec_all H3.
    rewrite H9; apply subword_app; constructor 2; auto.
    destruct l; try discriminate H3.
    destruct lw; try inversion H2.
    simpl in H9.
    destruct m; discriminate H9.
    
    destruct m as [ | a m ].
    
    simpl in H9.
    spec_all H3.
    rewrite <- H9; constructor 2; auto.
    destruct l; try discriminate H3.
    destruct lw; try inversion H2.
    simpl in H9; discriminate H9.
    
    injection H9; clear H9; intros H9 ?; subst a w.
    revert H6; apply H1, in_or_app; right; left; auto.
  Qed.
  
  Fact subword_not_intercalate l lw : 
       Forall2 (fun x w => forall y, In y w -> ~ R x y) l lw 
    -> forall w, length w < length l -> ~ subword R l (intercalate w lw).
  Proof.
    induction 1 as [ | x w l lw H1 H2 H3 ]; simpl.
    intros; omega.
    
    intros [ | y v ]; simpl.
    intros _ H.
    apply (@subword_not_flat_map (x::l) (w::lw)) in H.
    discriminate H.
    constructor; auto.
    
    intros Hv H.
    apply lt_S_n in Hv.
    apply subword_cons_inv in H.
    destruct H as (l1 & b & l2 & H5 & H6 & H7).
    apply app_split in H5.
    destruct H5 as [ (m & H8 & H9) | (m & H8 & H9) ].
    
    destruct m as [ | a m ]; simpl in H9.
    
    rewrite <- app_nil_end in H8.
    injection H9; clear H9; intros H9 ?; subst l1 b.
    subst l2; revert H7; apply H3; auto.
    
    injection H9; clear H9; intros H9 ?; subst a.
    apply H3 with (1 := Hv); rewrite H9.
    apply subword_app; constructor 2; auto.
    
    destruct m as [ | a m ]; simpl in H9.
    
    rewrite <- app_nil_end in H8.
    injection H9; clear H9; intros H9 ?; subst l1 b.
    subst l2; revert H7; apply H3; auto.
    
    injection H9; clear H9; intros H9 ?; subst a l2.
    subst w; revert H6; apply H1.
    apply in_or_app; right; left; auto.
  Qed.
 
  Fact subword_not_intercalate' w1 w2 lw : 
        (forall x y w, In x w1 -> In y w -> In w lw -> ~ R x y) 
     -> subword R w1 (intercalate w2 lw)
     -> subword R w1 w2.
  Proof.
    revert w1 lw.
    induction w2 as [ | y w2 IH ]; intros w1 [ | w lw ]; simpl; intros H; auto.
    destruct w1 as [ | x w1 ].
    constructor.
    intros H1.
    apply subword_cons_inv in H1.
    destruct H1 as (l & b & r & H1 & H2 & H3).
    contradict H2.
    assert (H2 : In b (l++b::r)).
      apply in_or_app; right; left; auto.
    rewrite <- H1 in H2.
    apply in_app_or in H2.
    destruct H2 as [ H2 | H2 ].
    apply H with w; auto; left; auto.
    rewrite in_flat_map in H2.
    destruct H2 as (w' & H2 & H4).
    apply H with w'; auto; left; auto.
    intros H1.
    apply subword_not_app in H1.
    2: intros; apply H with w; auto.
    destruct w1 as [ | x w1 ].
    apply subword_nil_true.
    apply subword_cons_inv in H1.
    destruct H1 as ([ | c l ] & b & r & H1 & H2 & H3); simpl in H1.

    injection H1; clear H1; intros; subst.
    constructor 3; auto.
    revert H3; apply IH.
    intros u v h ? ? ?; apply (H u v h); simpl; tauto.

    injection H1; clear H1; intros H1 ?; subst c.
    assert (subword R (x::w1) (intercalate w2 lw)) as H4.
      rewrite H1.
      apply subword_app.
      constructor 3; auto.
    apply IH in H4.
    constructor 2; auto.
    intros u v h ? ? ?; apply (H u v h); simpl; tauto.
  Qed.

End subword_not_intercalate.    

Section higman_decompositions.

  Variable (X : Type) (R : X -> X -> Prop).

  (* a Higman lemma decomposition of l into n words interleaved with (n-1) letters

              lw_1++w_2::lw_2++....++w_n::lw_n 

  *)

  Definition hig_decomp n (l : list X) c := 
    match c with
      | (w,lw) => length lw = n /\ S (length w) = n /\ l = intercalate w lw 
    end.

  Fact hig_decomp_length n l w lw : n = length l -> hig_decomp n l (w,lw) -> 1 <= lsum (map (@length _) lw).
  Proof.
    intros Hn (H1 & H2 & H3).
    subst l.
    rewrite intercalate_length_eq in Hn.
    omega.
  Qed.    
    
  Fact hig_decomp_finite_t n l : finite_t (hig_decomp n l).
  Proof.
    exists (filter (fun c => beq_nat (S (length (fst c))) n) (intercalate_inv n l)).
    intros (w,lw); split.
    rewrite filter_In.
    intros (H1 & H2).
    simpl fst in H2.
    rewrite beq_nat_true_iff in H2.
    rewrite intercalate_inv_spec in H1.
    destruct H1.
    red; auto.
    intros (H1 & H2 & H3).
    rewrite filter_In; split.
    rewrite intercalate_inv_spec; auto.
    simpl fst; rewrite H2.
    rewrite <- beq_nat_refl; auto. 
  Qed.
  
  Fact hig_decomp_finite n l : finite (hig_decomp n l).
  Proof.
    apply finite_t_finite, hig_decomp_finite_t.
  Qed.    

  Fact hig_decomp_exists_t n l : n < length l -> { c | hig_decomp (S n) l c }.
  Proof.    
    revert l; induction n as [ | n IHn ]; intros [ | x l ]; simpl; intros H; try omega.
    exists (nil,(x::l)::nil); simpl; rewrite <- app_nil_end; auto.
    apply lt_S_n, IHn in H.
    destruct H as ((w,lw) & H1 & H2 & H3).
    exists (x::w,nil::lw); simpl.
    rewrite H1, H2, H3; auto.
  Qed.

  Fact exists_hig_decomp n l c : hig_decomp n l c -> n <= S (length l).
  Proof.
    destruct c as (w,lw); intros (H1 & H2 & H3).
    rewrite H3, intercalate_length_eq; simpl; omega.
  Qed.
  
  Fact hig_decomp_Forall2 lx ly : length lx = length ly 
                             -> (forall d, hig_decomp (length lx) ly d
                                        -> Exists2 (R o @In _) lx (snd d))
                             -> Forall2 R lx ly.
  Proof.
    rewrite list_length_eq_Forall2.
    induction 1 as [ | x y lx ly H1 H2 H3 ].
    constructor.
    intros H; constructor.
    rewrite <- list_length_eq_Forall2 in H2.
    specialize (H (ly,((y::nil)::map (fun _ => nil) (list_n (length lx))))).
    spec_all H; auto.
    red; repeat split; auto.
    simpl; rewrite map_length, list_n_length; auto.
    simpl; f_equal; auto.
    rewrite intercalate_head_map_nil; auto.
    apply Exists2_cons_inv in H.
    destruct H as [ (b & H4 & [ H5 | [] ]) | H ].
    subst; auto.
    apply Exists2_In in H.
    destruct H as (a & b & H4 & H5 & H6).
    rewrite in_map_iff in H5.
    destruct H5 as (i & ? & H5); subst b.
    destruct H6 as (? & _ & []).
    apply H3.
    intros (w,lw) (H4 & H5 & H6).
    specialize (H (y::w,nil::lw)).
    spec_all H.
    red; repeat split; simpl; f_equal; auto.
    apply Exists2_cons_inv in H.
    destruct H as [ (? & _ & []) | H ].
    assumption.
  Qed.

  (* this one uses a "combinatorial" principle *)

  Fact hig_decomp_subword lx ly : length lx <= length ly 
                              -> (forall d, hig_decomp (length lx) ly d
                                         -> Exists2 (R o @In _) lx (snd d))
                               -> subword R lx ly.
  Proof.
    revert lx.
    induction ly as [ | y ly ].
    intros [ | ? ? ]; simpl; try constructor; omega.
    
    simpl; intros lx H.
    destruct (le_lt_dec (S (length ly)) (length lx)) as [ H' | H' ].
    assert (length lx = S (length ly)) as H''.
      omega.
    clear H' H; rename H'' into Hlxy.   
    intros H.
    apply subword_Forall2.
    apply hig_decomp_Forall2; auto; simpl; f_equal; auto.
    
    clear H; rename H' into Hlxy.
    intros H.
    apply le_S_n in Hlxy.
        
    destruct lx as [ | x lx ].
    apply subword_nil_true.

    (* combinatorial principle *)

    assert (forall d, hig_decomp (length (x::lx)) ly d -> Exists2 (R o @In _) (x::lx) (snd d) \/ R x y) as H1.
      intros (w,[ | a lw ]) (H1 & H2 & H3); simpl.
      simpl in H1; omega.      
      specialize (H (w,(y::a)::lw)); spec_all H.
      red; repeat split; simpl; f_equal; auto.
      rewrite H3; destruct w; simpl; auto.
      simpl in H |- *.
      apply Exists2_cons_inv in H.
      destruct H as [ (b & H4 & [H5 | H5 ]) | H ].
      subst; right; auto.
      left; constructor 1.
      exists b; auto.
      injection H1; auto.
      left; constructor 2; auto.
    apply finite_decide in H1.
    2: apply hig_decomp_finite.
    
    (* now we can choose *)

    destruct H1 as [ H1 | (u & _ & H2) ].
    constructor 2; auto.
    constructor 3; auto.
    apply IHly.
    simpl in Hlxy; omega.
    clear u H2.
    intros (w,lw) (H1 & H2 & H3).
    simpl.
    specialize (H (y::w,nil::lw)); spec_all H.
    red; repeat split; simpl; f_equal; auto.
    simpl in H.
    apply Exists2_cons_inv in H.
    destruct H as [ (? & _ & []) | H ]; auto.
  Qed.

End higman_decompositions.

Section higman_splits.

  Variable (X : Type) (R : X -> X -> Prop).

  (* a Higman theorem decomposition of ll into ll = l++x::r *)

  Definition hig_split (l : list X) c := hig_decomp (length l) l c.

  Fact hig_split_length l w lw : hig_split l (w,lw) -> lsum (map (@length _) lw) = 1.
  Proof.
    intros (H1 & H2 & H3).
    rewrite H3, intercalate_length_eq in H2.
    omega.
  Qed.
    
  Fact hig_split_finite_t l : finite_t (hig_split l).
  Proof.
    apply hig_decomp_finite_t.
  Qed.
  
  Fact hig_split_finite l : finite (hig_split l).
  Proof.
    apply finite_t_finite, hig_split_finite_t.
  Qed.    

  Fact hig_split_exists_t l : 0 < length l -> { c | hig_split l c }.
  Proof.
    intros H.
    unfold hig_split.
    case_eq (length l).
    omega.
    intros n Hn.
    apply hig_decomp_exists_t.
    omega.
  Qed.
  
  Fact hig_split_Forall2 lx ly : length lx = length ly 
                             -> (forall d, hig_split ly d
                                        -> Exists2 (R o @In _) lx (snd d))
                             -> Forall2 R lx ly.
  Proof.
    intros H1.
    unfold hig_split.
    rewrite <- H1.
    apply (hig_decomp_Forall2 _ _ H1).
  Qed.

End higman_splits.


Section not_subword_intercalate.
 
  Variables (X : Type) (R : X -> X -> Prop) 
            (Rdec : forall x y, { R x y } + { ~ R x y }).

  Fact not_subword_hig_decomp l m : 
         length l <= S (length m)
      -> ~ subword R l m
      -> { d | hig_decomp (length l) m d 
            /\ Forall2 (fun x => Forall (fun z => ~ R x z)) l (snd d) }.
  Proof.
    intros H2 H3.
    destruct (le_lt_dec (length l) (length m)) as [ H4 | H4 ].

    (* case where length l <= length m *)

    assert (~ (forall d, hig_decomp (length l) m d -> Exists2 (R o @In _) l (snd d))) as C.
      contradict H3.
      apply hig_decomp_subword; auto.
    apply finite_t_not_build in C.

    destruct C as (d & G1 & G4).
    exists d; repeat split; auto; try omega.

    destruct d as (w,lw); destruct G1 as (G1 & G2 & G3).
    simpl in G4 |- *.
    symmetry in G1; rewrite list_length_eq_Forall2 in G1.
    clear H2 H3 H4 w m G2 G3.
    revert G1 G4.
    induction 1 as [ | a w l lw H1 H2 H3 ]; intros H0.
    constructor.
    apply Forall2_length in H2.
    assert (~ (R o @In _) a w) as H4.
      contradict H0; constructor 1; auto.
    assert (~ Exists2 (R o In (A:=X)) l lw) as H5.
      contradict H0; constructor 2; auto.
    clear H0.
    apply H3 in H5.
    constructor; auto.
    rewrite Forall_forall.
    intros x Hx; contradict H4; exists x; auto.

    apply hig_decomp_finite_t.

    intros (w,lw) _; simpl.
    apply Exists2_dec.
    intros x ll _ _ .
    destruct (list_dec_rec (R x) ll) as [ (y & G1 & G2) | G ].
    intros; apply Rdec.
    left; exists y; auto.
    right; intros (y & G1 & G2); apply G with (1 := G2); auto.
    
    (* case where length m < length l *)
 
    assert (length l = S (length m)) as H1.
      omega.
      
    exists (m, map (fun _ => nil) l).
    repeat split; auto.
    rewrite map_length; auto.
    apply intercalate_map_nil.
    simpl; rewrite Forall2_map_right, Forall2_Forall, Forall_forall.
    intros; constructor.
  Qed. 

  Fact not_subword_intercalate n l m : 
            length l = S n 
         -> n <= length m 
         -> ~ subword R l m 
         -> { w : _ & 
               { lw |  m = intercalate w lw 
                    /\ length w = n
                    /\ length lw = S n
                    /\ Forall2 (fun x y => Forall (fun z => ~ R x z) y) l lw } }.
  Proof.
    intros H1 H2 H3.
    destruct not_subword_hig_decomp with (2 := H3)
      as ((w,lw) & (G1 & G2 & G3) & G4).
      
    omega.
    exists w ,lw; repeat split; auto; omega.
  Qed.

End not_subword_intercalate.



    
    
