(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Relations.
Require Import Wellfounded.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.
Require Import continuity.

Set Implicit Arguments.

Section Acc_utils.

  Variable (X : Type) (R S : X -> X -> Prop) (P : X -> Prop).
  
  Hypothesis H1 : forall x y, S x y -> P y -> R x y.
  Hypothesis H2 : forall x y, R x y -> P x.
  
  Theorem Acc_restr l : P l -> Acc R l -> Acc S l.
  Proof.
    intros H3 H4; revert H3.
    induction H4 as [ l Hl IHl ]; intros H3.
    constructor; intros m Hm.
    assert (R m l) as H4.
      apply H1; auto.
    apply IHl; auto.
    apply H2 in H4; tauto.
  Qed.

End Acc_utils.

Section lex_prod.

  Variables (X Y Z : Type) (f : X -> Y -> Z) (R : relation X) (S : relation Y).

  Hypothesis f_inj : forall x1 x2 y1 y2, f x1 y1 = f x2 y2 -> x1 = x2 /\ y1 = y2.
  
  Inductive lex_prod : Z -> Z -> Prop :=
    | in_lp0 : forall x1 x2 y1 y2, R x1 x2 -> lex_prod (f x1 y1) (f x2 y2)
    | in_lp1 : forall x y1 y2,     S y1 y2 -> lex_prod (f x  y1) (f x  y2).
    
  Fact lex_prod_inv z1 z2 : lex_prod z1 z2 -> (exists x1 x2 y1 y2, R x1 x2 /\ z1 = f x1 y1 /\ z2 = f x2 y2)
                                           \/ (exists x     y1 y2, S y1 y2 /\ z1 = f x  y1 /\ z2 = f x  y2).
  Proof.
    induction 1.
    left; exists x1, x2, y1, y2; auto.
    right; exists x, y1, y2; auto.
  Qed.

  Let K z1 z2 := exists x1 x2 y1 y2, lex_prod z1 z2 /\ z1 = f x1 y1 /\ z2 = f x2 y2 
                                  /\ Acc R x1 /\ Acc R x2
                                  /\ Acc S y1 /\ Acc S y2.   

  Fact lex_prod_Acc x y : Acc R x -> Acc S y -> Acc K (f x y).
  Proof.
    intros H; revert y; induction H as [ x Hx IHx ].
    induction 1 as [ y Hy IHy ].
    constructor; intros z (x1 & x2 & y1 & y2 & Hz & G1 & G2 & G3 & G4 & G5 & G6).
    apply lex_prod_inv in Hz.
    destruct Hz as [ (a1 & a2 & b1 & b2 & H1 & H2 & H3)
                   | (a1      & b1 & b2 & H1 & H2 & H3) ];
    apply f_inj in H3; destruct H3; subst;
    apply f_inj in G1; destruct G1; subst;
    apply f_inj in G2; destruct G2; subst.
    apply IHx; auto.
    apply IHy; auto.
  Qed.

  Variable HR : well_founded R.
  Variable HS : well_founded S.

  Fact lex_prod_wf x y : Acc lex_prod (f x y).
  Proof.
    generalize (lex_prod_Acc (HR x) (HS y)).
    apply Acc_restr with (P := fun z => exists x y, z = f x y); try clear x y.

    induction 1.
    intros (x' & y' & H1); apply f_inj in H1; destruct H1; subst.
    red.
    exists x1, x', y1, y'; split.
    constructor 1; auto.
    split; auto.
    intros (x' & y' & H1); apply f_inj in H1; destruct H1; subst.
    red.
    exists x', x', y1, y'; split.
    constructor 2; auto.
    split; auto.

    intros x y (_ & _ & _ & _ & H & _).
    induction H.
    exists x1, y1; auto.
    exists x, y1; auto.

    exists x, y; auto.
  Qed.

End lex_prod.

Section lex.

  Variables (X Y : Type) (R : X -> X -> Prop) (S : Y -> Y -> Prop).
    
  Notation Acc_bar := (fun (c : X*Y) => Acc R (fst c) /\ Acc S (snd c)).

  Definition lex (c1 c2 : X*Y) : Prop := 
    match c1, c2 with
      | (g,t), (f,s) => R g f \/ (g = f /\ S t s)
    end.

  Let K (c1 c2 : X*Y) := lex c1 c2 /\ Acc_bar c1 /\ Acc_bar c2.

  Fact lex_Acc_bar_Acc c : Acc_bar c -> Acc K c.
  Proof.
    destruct c as (x,y); simpl; intros [ H1 H2 ].
    revert y H2; induction H1 as [ x Hx IHx ].
    induction 1 as [ y Hy IHy ].
    constructor.
    intros (a,b) (H1 & (H2 & H3) & (H4 & H5)).
    destruct H1 as [ H1 | [H0 H1] ].
    apply IHx; auto.
    subst a.
    apply IHy; auto.
  Qed.

  Section lex_wf.

    Hypothesis R_wf : well_founded R.
    Hypothesis S_wf : well_founded S.
    
    Fact lex_wf : well_founded lex.
    Proof.
      intros (x,y).
      apply Acc_dec with K.
      clear x y.
      intros (x1,y1) (x2,y2); split; simpl; auto.
      apply lex_Acc_bar_Acc.
      split; auto.
    Qed.

  End lex_wf.
  
End lex.

Section lex_props.

  Variables (X Y : Type).

  Fact lex_mono_left S : op_monotonic (fun R => @lex X Y R S).
  Proof.
    intros ? ? ? [] [] [ | [] ]; [ left | right ]; auto.
  Qed.
  
  Fact lex_mono_right R : op_monotonic (fun S => @lex X Y R S).
  Proof.
    intros ? ? ? [] [] [ | [] ]; [ left | right ]; auto.
  Qed.
  
  Fact lex_w_cont_left S : op_w_continuous (fun R => @lex X Y R S).
  Proof.
    intros K HK (x1,y1) (x2,y2) [ (n & H1) | (H1 & H2) ].
    exists n; left; auto.
    exists 0; right; auto.
  Qed.
  
  Fact lex_w_cont_right R : op_w_continuous (fun S => @lex X Y R S).
  Proof.
    intros K HK (x1,y1) (x2,y2) [ H1 | (H1 & n & H2) ].
    exists 0; left; auto.
    exists n; right; auto.
  Qed.
    
End lex_props.

Section lex_seq.

  Variable (X : Type) (R : X -> X -> Prop) (HR : well_founded R).

  Definition lex_any (s t : nat -> X) := exists i, R (s i) (t i) /\ forall j, i < j -> s j = t j.

  Definition lex_seq k (s t : nat -> X) := exists i, i < k /\ R (s i) (t i) /\ forall j, i < j -> s j = t j.

  Fact lex_seq_any k s t : lex_seq k s t -> lex_any s t.
  Proof.
    intros (i & _ & ?); exists i; auto.
  Qed.
  
  Fact lex_any_seq k s t : (forall j, k < j -> s j = t j) -> lex_any s t -> exists k', lex_seq k' s t.
  Proof.
    intros H (i & H1 & H2).
    exists (i+k+1).
    exists i; repeat split; auto.
    omega.
  Qed.  
    
  Fact lex_seq_0 s t : ~ lex_seq 0 s t.
  Proof.
    intros (i & H & _); omega.
  Qed.

  Fact lex_seq_S k s t : lex_seq (S k) s t <-> ( R (s k) (t k) /\ (forall j, k < j -> s j = t j)
                                              \/ lex_seq k s t ).
  Proof.
    split.

    intros (i & H1 & H2 & H3).
    destruct (le_lt_dec k i) as [ H6 | H6 ].
    assert (i = k). omega.
    subst i.
    left; auto.
    right.
    exists i; auto.
    
    intros [ (H1 & H2) | (i & H1 & H2 & H3) ].
    exists k; auto.
    exists i; auto.
  Qed.

  Fact lex_seq_eq k s t : lex_seq k s t -> forall i, k <= i -> s i = t i.
  Proof.
    intros (i & ? & _ & Hi).
    intros; apply Hi; omega.
  Qed.
      
  Theorem wf_lex_seq k : well_founded (lex_seq k).
  Proof.
    induction k as [ | k IHk ].
    
    intros s.
    constructor 1; intros t Ht.
    destruct (lex_seq_0 Ht).
    
    set (h (s : nat -> X) := (s k,s)).
    generalize (wf_inverse_image _ _ _ h (lex_wf HR IHk)).
    apply wf_incl.
    intros s t H.
    apply lex_seq_S in H.
    destruct H as [ (H & _) | H ].
    left; auto.
    right; split; auto.
    apply lex_seq_eq with (1 := H).
    auto.
  Qed.

End lex_seq.

Section lex_map.

  Variables (X Y : Type) (RX : X -> X -> Prop) (RY : X -> Y -> Y -> Prop) (f : Y -> X).
  Hypothesis HRX : well_founded RX.
  Hypothesis HRY : forall x, well_founded (RY x).
  
  Definition lex_map (y1 y2 : Y) := RX (f y1) (f y2) \/ (f y1 = f y2 /\ RY (f y1) y1 y2).
  
  Let wf_lex_map_rec x : forall y, f y = x -> Acc lex_map y. 
  Proof.
    induction x as [ x IHx ] using (well_founded_induction HRX).
    intros y Hy.
    induction y as [ y IHy ] using (well_founded_induction (HRY x)).
    constructor.
    intros z [ H1 | (H1 & H2) ].
    apply IHx with (f z); subst; auto.
    rewrite Hy in H1.
    rewrite H1 in H2.
    apply IHy; auto.
  Qed.
  
  Fact wf_lex_map : well_founded lex_map.
  Proof.
    intros y; apply wf_lex_map_rec with (1 := eq_refl).
  Qed.

End lex_map.    
