(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Set Implicit Arguments.

Section list_insert.

  Variable X : Type.

  Fixpoint list_insert n (x:X) ll :=
    match n, ll with
      | 0, ll      => x::ll
      | S _, nil   => x::nil
      | S n, y::ll => y::list_insert n x ll
    end.

  Fact list_insert_length n x ll : length (list_insert n x ll) = S (length ll).
  Proof.
    revert ll; induction n; intros []; simpl; auto.
  Qed.

  Fact In_list_insert y n x ll : In y (list_insert n x ll) <-> x = y \/ In y ll.
  Proof.
    split.
    revert ll; induction n; simpl; auto.
    intros [ | z ll ].
    intros [ | [] ]; auto.
    intros [ H | H ].
    right; left; auto.
    destruct (IHn _ H); auto.
    right; right; auto.
    
    intros [ H | H ]; subst.
    
    revert ll; induction n; simpl; auto.
    intros [ | z ll ]; simpl; auto.
    
    revert ll H; induction n; simpl; auto.
    intros [ | z ll ] Hll; simpl; auto.
    destruct Hll; auto.
  Qed.

  Fixpoint list_remove n (ll : list X) := 
    match n, ll with
      | _, nil     => nil
      | 0,   _::ll => ll
      | S n, x::ll => x::list_remove n ll
    end.

  Fact list_remove_length n ll : n < length ll -> S (length (list_remove n ll)) = length ll.
  Proof.
    revert n; induction ll as [ | x ll IH ].
    intro; simpl; omega.
    intros [ | n ] Hn; simpl in Hn |- *; auto.
    f_equal; apply IH; omega.
  Qed.
  
  Fact list_remove_In n ll x : In x (list_remove n ll) -> In x ll.
  Proof.
    revert n x; induction ll as [ | x ll IH ]; intros [ | n ] y; simpl; auto.
    intros [ H | H ]; auto.
    right; revert H; apply IH.
  Qed.
    
  Fact list_remove_In_t n ll x : In_t x (list_remove n ll) -> In_t x ll.
  Proof.
    revert n x; induction ll as [ | x ll IH ]; intros [ | n ] y; simpl; auto.
    intros [ H | H ]; auto.
    right; revert H; apply IH.
  Defined.

  Fixpoint list_extract (ll : list X) n { struct ll } : n < length ll -> X.
  Proof.
    refine(match ll with
      | nil     => fun Hn => _
      | x::mm   => fun Hn => _
    end); simpl in Hn.
    exfalso; revert Hn; apply lt_n_0.
    destruct n as [ | n ].
    apply x.
    apply (list_extract mm n).
    apply lt_S_n, Hn.
  Defined.

  Fact list_extract_pirr ll n H1 H2 : @list_extract ll n H1 = @list_extract ll n H2.
  Proof.
    revert n H1 H2.
    induction ll as [ | x ll IH ]; intros [ | n ] H1 H2; simpl in H1, H2 |- *; auto; omega.
  Qed.    

  Fact list_extract_In ll n Hn : In (@list_extract ll n Hn) ll.
  Proof.
    revert n Hn.
    induction ll as [ | x ll IH ].
    intros n; simpl; intro; omega.
    intros [ | n ] Hn; simpl in Hn |- *.
    left; auto.
    right; auto.
  Qed.

  Fact list_extract_In_t ll n Hn : In_t (@list_extract ll n Hn) ll.
  Proof.
    revert n Hn.
    induction ll as [ | x ll IH ].
    intros n; simpl; intro; omega.
    intros [ | n ] Hn; simpl in Hn |- *.
    left; auto.
    right; auto.
  Defined.

  Fact list_extract_prop ll n (Hn : n < length ll) : ll = list_insert n (list_extract ll Hn) (list_remove n ll).
  
  Proof.
    revert n Hn.
    induction ll as [ | x ll IH ].
    intros n Hn; simpl in Hn; omega.
    intros [ | n ] Hn; simpl; auto.
    f_equal; auto.
  Qed.

  Fact list_extract_insert ll x n Hn : @list_extract (list_insert n x ll) n Hn = x.
  Proof.
    revert n Hn.
    induction ll; intros [ | n ] Hn; simpl; auto.
    simpl in Hn; omega.
  Qed.

  Fact list_remove_insert ll x n : n <= length ll -> @list_remove n (list_insert n x ll) = ll.
  Proof.
    revert ll; induction n as [ | n IHn ]; intros [ | y ll ]; simpl; auto.
    omega.
    intros H; f_equal; apply IHn; auto.
    omega.
  Qed.

  Let inj11 k : { i | i < S k } := exist _ 0 (lt_0_Sn k).

  Let inj12 k (x : { i | i < k }) : { i | i < S k } := 
    match x with exist n Hn => exist _ (S n) (lt_n_S n k Hn) end.

  Let inj31 (x : X) ll : (forall v, In_t v ll -> In_t v (x::ll)).
  Proof. right; auto. Defined.
  
  Let inj32 (x : X) ll mm : (forall v, In_t v mm -> In_t v ll) -> (forall v, In_t v (x::mm) -> In_t v (x::ll)).
  Proof.
    intros H v [ Hv | Hv ].
    left; auto.
    right; apply H, Hv.
  Defined.
  
  Fixpoint list_positions n : list { x | x < n }.
  Proof.
    refine (match n with 
              | 0   => nil
              | S u => (exist _ 0 _)::_
            end).
    apply lt_0_Sn.
    apply (map (@inj12 u)), list_positions.
  Defined.
  
  Fact list_positions_length n : length (list_positions n) = n.
  Proof.
    induction n; simpl; f_equal; auto.
    rewrite map_length; auto.
  Qed.
  
  Fact list_positions_spec n x : In x (list_positions n) <-> proj1_sig x < n.
  Proof.
    split.
    
    revert x; induction n as [ | n IH ].
    intros ? [].
    intros ([|x] & Hx); simpl.
    omega.
    intros [ H | H ].
    discriminate H.
    apply in_map_iff in H.
    destruct H as ((i & Hi) & H1 & H2).
    simpl in H1.
    injection H1; intro; subst i; auto.

    destruct x as (x & Hx).
    revert x Hx; induction n as [ | n IH ].
    intros ? ?; omega.
    intros [ | x ] Hx.
    simpl; left; f_equal; apply lt_pirr.
    simpl; right.
    apply in_map_iff.
    exists (exist _ x (lt_S_n _ _ Hx)); split.
    simpl; f_equal; apply lt_pirr.
    apply IH.
    simpl.
    apply lt_S_n; auto.
  Qed.
  
  Fact list_pos_extract ll : map (fun c => match c with exist n Hn => list_extract ll Hn end) (list_positions (length ll)) = ll.
  Proof.
    induction ll as [ | x ll IH ]; simpl; f_equal.
    rewrite <- IH at 3.
    rewrite map_map.
    apply map_ext.
    intros (a & Ha); simpl.
    apply list_extract_pirr.
  Qed.

  Definition list_pos_In_t ll k : k < length ll -> { x : X & In_t x ll }.
  Proof.
    intros Hk.
    exists (list_extract ll Hk).
    apply list_extract_In_t.
  Defined.

  Fixpoint In_t_list_pos ll (x : X) { struct ll } : In_t x ll -> nat.
  Proof.
    destruct ll as [ | y ll ].
    intros [].
    intros [ H | H ].
    exact 0.
    exact (S (In_t_list_pos ll x H)).
  Defined.

  Fact split_In_t_pos_eq l x r : In_t_list_pos _ _ (split_In_t l x r) = length l.
  Proof.
    induction l; simpl; auto.
  Qed.  

  Fact In_t_list_pos_prop ll x H : In_t_list_pos ll x H < length ll.
  Proof.
    induction ll as [ | y ll ].
    destruct H.
    destruct H as [ H | H ]; simpl.
    omega.
    apply lt_n_S, IHll.
  Qed.
  
  Fact list_extract_In_t_prop ll x H : list_extract ll (In_t_list_pos_prop ll x H) = x.
  Proof.
    induction ll as [ | y ll IH ].
    destruct H.
    destruct H as [ H | H ]; simpl; auto.
    specialize (IH H).
    replace_with IH.
    apply list_extract_pirr.
  Qed.  
   
  Definition list_insert_inv (ll : list X) : { mm : list ({ n | n < length ll } 
                                           * { x : _ & In_t x ll } 
                                           * { mm : list X & forall v, In_t v mm -> In_t v ll } ) 
                                             | forall n x l, n < length ll ->
                                                  ( (exists Hn Hx Hl, In (exist _ n Hn, existT _ x Hx, existT _ l Hl) mm)
                                               <-> ll = list_insert n x l) }.
  Proof.
    induction ll as [ | x ll (mm & Hmm) ].
    exists nil.
    intros n x l Hn; exfalso; simpl in Hn; abstract omega.

    simpl. 
    exists ((inj11 (length ll),
             existT _ x (inl eq_refl),
             existT _ ll (inj31 x ll)
            )::
            map (fun c => match c with 
                   (n,existT y Hy,existT l Hl) => (inj12 n,
                                                   existT _ y (inr Hy),
                                                   existT _ (x::l) (inj32 ll Hl))  
                          end) mm).
    intros n y l Hn0; split.

    intros (Hn & Hx & Hl & [ H | H ]).
    subst; injection H; intros; subst; auto.
    apply in_map_iff in H.
    destruct H as ((((i,Hi),(z,Hz)),(k,Hk)) & H1 & H2).
    injection H1; clear H1. 
    intros H3 ? H4 ? ?. subst.
    simpl; f_equal. 
    apply Hmm with (1 := Hi); auto.
    exists Hi, Hz, Hk; auto.
    
    intros H.
    destruct n as [ | n ].
    simpl in H; injection H; clear H; intros; subst.
    exists (lt_0_Sn (length l)), (inl eq_refl), (inj31 y l).
    left; auto.
    destruct l as [ | u l ].
    simpl in H; injection H; clear H; intros; subst.
    simpl in Hn0; abstract omega.
    simpl in H; injection H; clear H.
    intros H1 ?; subst u.
    apply Hmm with (1 := lt_S_n _ _ Hn0) in H1.
    destruct H1 as (Hn & Hy & Hl & H1).
    exists (lt_n_S _ _ Hn), (inr Hy), (inj32 ll Hl).
    right; apply in_map_iff.
    exists (exist _ n Hn,existT _ y Hy,existT _ l Hl); auto. 
  Defined.
   
  Fixpoint list_n_map n : (forall i, i < n -> X) -> list X.
  Proof.
    refine (match n with
      | 0   => fun _ => nil
      | S m => fun f => _
    end).
    
    apply (cons (f _ (lt_0_Sn _))).
    apply (list_n_map m).
    intros i Hi; apply (f (S i)), lt_n_S, Hi.
  Defined.

  Fact list_n_map_length n f : length (@list_n_map n f) = n.
  Proof. induction n; simpl; auto. Qed.

  Variable Y : Type.

  Fixpoint In_t_insert (a : X) ll (H : In_t a ll) (y : Y) (mm : list Y) { struct ll } : list Y.
  Proof.
    destruct ll as [ | b ll ].
    destruct H.
    destruct H as [ H | H ].
    apply (y::mm).
    destruct mm as [ | v mm ].
    apply (y::nil).
    apply (cons v).
    apply (In_t_insert _ _ H y mm).
  Defined.

  Fact In_t_insert_left a b ll H y mm : @In_t_insert a (b::ll) (inl H) y mm = y::mm.
  Proof. auto. Qed.
  
  Fact In_t_insert_right0 a b ll H y : @In_t_insert a (b::ll) (inr H) y nil = y::nil.
  Proof. auto. Qed.
  
  Fact In_t_insert_right1 a b ll H y u mm : @In_t_insert a (b::ll) (inr H) y (u::mm) = u::In_t_insert a ll H y mm.
  Proof. auto. Qed.

  Fact In_t_insert_length a ll H y mm : length (In_t_insert a ll H y mm) = S (length mm).
  Proof.
    revert mm.
    induction ll as [ | b ll ].
    destruct H.
    destruct H as [ H | H ].
    intros mm; rewrite In_t_insert_left; auto.
    intros [ | v mm ].
    rewrite In_t_insert_right0; auto.
    rewrite In_t_insert_right1; simpl; f_equal; auto.
  Qed.

  Fixpoint In_t_remove (a : X) ll (H : In_t a ll) (mm : list Y) { struct ll } : list Y.
  Proof.
    destruct ll as [ | b ll ].
    destruct H.
    destruct H as [ H | H ].

    destruct mm as [ | v mm ].
    apply nil.
    apply mm.
    
    destruct mm as [ | v mm ].
    apply nil.
    apply (v::In_t_remove _ _ H mm).
  Defined.
  
  Fact In_t_remove_left0 a b ll H : @In_t_remove a (b::ll) (inl H) nil = nil.
  Proof. auto. Qed.

  Fact In_t_remove_left1 a b ll H y mm : @In_t_remove a (b::ll) (inl H) (y::mm) = mm.
  Proof. auto. Qed.
  
  Fact In_t_remove_right0 a b ll H : @In_t_remove a (b::ll) (inr H) nil = nil.
  Proof. auto. Qed.

  Fact In_t_remove_right1 a b ll H y mm : @In_t_remove a (b::ll) (inr H) (y::mm) = y::In_t_remove a ll H mm.
  Proof. auto. Qed.
  
  Fact In_t_remove_length a ll H mm : length ll <= length mm -> S (length (In_t_remove a ll H mm)) = length mm.
  Proof.
    revert mm; induction ll as [ | b ll ].
    destruct H.
    destruct H as [ H | H ]; intros [ | v mm ] Hmm.
    simpl in Hmm; omega.
    rewrite In_t_remove_left1; auto.
    simpl in Hmm; omega.
    rewrite In_t_remove_right1; simpl; f_equal.
    apply IHll; simpl in Hmm; omega.
  Qed.

  Fixpoint In_t_extract x (ll : list X) (p : In_t x ll) (mm : list Y) : option Y.
  Proof.
    destruct ll as [ | b ll ].
    destruct p.
    destruct mm as [ | u mm ].
    apply None.
    destruct p as [ H | p ].
    apply Some, u.
    apply (In_t_extract x ll p mm).
  Defined. 
 
  Fact In_t_extract_left0 a b ll H : In_t_extract a (b::ll) (inl H) nil = None.
  Proof. auto. Qed.
  
  Fact In_t_extract_left1 a b ll H u mm : In_t_extract a (b::ll) (inl H) (u::mm) = Some u.
  Proof. auto. Qed.

  Fact In_t_extract_right0 a b ll H : In_t_extract a (b::ll) (inr H) nil = None.
  Proof. auto. Qed.
  
  Fact In_t_extract_right a b ll H u mm : In_t_extract a (b::ll) (inr H) (u::mm) = In_t_extract a ll H mm.
  Proof. auto. Qed.

  Fixpoint nat_In_t ll : forall k, k < length ll -> { x : X & In_t x ll }.
  Proof.
    refine (match ll with 
      | nil   => fun _ => _
      | x::mm => fun k => match k with 0 => _ | S l => _ end
    end).
    simpl; intro H; exfalso; revert H; apply lt_n_0.
    intros _; exists x; left; reflexivity.
    simpl; intros H.
    apply lt_S_n in H.
    destruct (nat_In_t mm l H) as (y & Hy).
    exists y; right; assumption.
  Defined.

  Fact nat_In_t_fix0 x ll H : @nat_In_t (x::ll) 0 H = existT _ x (inl eq_refl).
  Proof. auto. Qed.
  
  Fact nat_In_t_fix1 x ll k H : @nat_In_t (x::ll) (S k) H 
                              = let (x,Hx) := nat_In_t ll (lt_S_n _ _ H) in existT _ x (inr Hx).
  Proof. auto. Qed.
  
  Fixpoint In_t_nat ll (x : X) { struct ll } : In_t x ll -> { k | k < length ll }.
  Proof.
    refine (match ll with
      | nil   => _
      | y::mm => _
     end).
     intros [].
     intros [ H | H ].
     subst; exists 0; simpl; apply lt_0_Sn.
     destruct (In_t_nat _ _ H) as (k & Hk).
     exists (S k); apply lt_n_S, Hk.
   Defined.

End list_insert.

Fact list_remove_map X Y (f : X -> Y) k ll : map f (list_remove k ll) = list_remove k (map f ll).
Proof.
  revert k; induction ll; intros [ | ? ]; simpl; f_equal; auto.
Qed.

Fact list_insert_Forall2 X Y (R : X -> Y -> Prop) k x y l m : R x y -> Forall2 R l m -> Forall2 R (list_insert k x l) (list_insert k y m).
Proof.
  intros Hxy H; revert H k.
  induction 1; intros [|]; simpl; constructor; auto.
Qed.

Fact list_extract_Forall2 X Y (R : X -> Y -> Prop) l m : Forall2 R l m <-> length l = length m
                                                                        /\ forall k Hl, exists Hm, R (@list_extract _ l k Hl) (@list_extract _ m  k Hm).
Proof.
  split.
  
  intros H; split; revert H.
  apply Forall2_length.
  induction 1 as [ | x y l m H1 H2 H3 ].
  intros k Hl; exfalso; simpl in Hl; omega.
  intros [ | k ] Hl. 
  simpl; exists (lt_0_Sn _); auto.
  simpl in Hl.
  specialize (H3 _ (lt_S_n _ _ Hl)).
  destruct H3 as (Hm & H3).
  exists (lt_n_S _ _ Hm); simpl.
  apply Forall2_length in H2.
  generalize Hl; intros Hl0.
  replace_with H3; apply list_extract_pirr.
  
  revert m; induction l as [ | x l IH ]; intros [ | y m ] (H1 & H2).
  constructor.
  discriminate H1.
  discriminate H1.
  simpl in H1; injection H1; clear H1; intros H1.
  constructor.
  destruct (H2 _ (lt_0_Sn _)) as (Hm & H3); auto.
  apply IH; split; auto.
  intros k Hl.
  destruct (H2 _ (lt_n_S _ _ Hl)) as (Hm & H3).
  exists (lt_S_n _ _ Hm).
  replace_with H3; simpl; apply list_extract_pirr.
Qed.