(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.
Require Import Wellfounded.

Require Omega.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Set Implicit Arguments.

(* this is an implementation of trees as an actual list of
   tree (nested induction). Since the sons form a list of trees,
   one can use all the results of the List.v library (map etc ...)

   We derive a very short induction principle tree_rect with
   the natural fixpoint equation, assuming the extensionality
   of the transformation. As of now, Coq is not able to 
   generate such induction principle itself as it does on
   nat or list for instance. 

   With this machinery, reasoning inductively on tree can be
   done directly without any detour through the size or the
   height of trees which makes recursion on trees very painful ...
*)

Section trees.

  Variable (X : Type).

  (* we do not want the too weak Coq generated induction principles *)

  Unset Elimination Schemes.

  Inductive tree : Type := in_tree : X -> list tree -> tree.

  Set Elimination Schemes.

  (* let us define our own induction principles *)

  Section tree_rect.

    Variable P : tree -> Type.
    Hypothesis f : forall a ll, (forall x, In_t x ll -> P x) -> P (in_tree a ll).

    Fixpoint tree_rect t : P t :=
      match t with
        | in_tree a ll => f a _ (map_t _ tree_rect  _) 
      end.

    Hypothesis f_ext : forall a ll f1 f2, (forall x Hx, f1 x Hx = f2 x Hx) -> f a ll f1 = f a ll f2.

    (* Coq recursive type-checking does not allow such a definition 
       but it is possible to prove this identity
    *)

    Fact tree_rect_fix a ll : tree_rect (in_tree a ll) = f a ll (fun t _ => tree_rect t).
    Proof.
      simpl.
      apply f_ext.
      clear a; intros a Ha.
      apply map_t_prop.
    Qed.

  End tree_rect.

  Definition tree_rec (P : tree -> Set) := tree_rect P.

  Section tree_ind.

    Variable P : tree -> Prop.
    Hypothesis HP : forall a ll, (forall x, In x ll -> P x) -> P (in_tree a ll).

    Definition tree_ind : forall t, P t.
    Proof.
      apply tree_rect.
      intros; apply HP.
      rewrite <- In_In_t; auto.
    Qed.

  End tree_ind.

  Section tree_recursion.

    (* the particular case when the output type does not depend on the tree *)

    Variables (Y : Type) (f : X -> list tree -> list Y -> Y).

    Definition tree_recursion : tree -> Y.
    Proof.
      apply tree_rect.
      intros x ll IH.
      apply (f x ll (list_In_t_map _ IH)).
    Defined.

    Fact tree_recursion_fix x ll : tree_recursion (in_tree x ll) = f x ll (map tree_recursion ll).
    Proof.
      unfold tree_recursion at 1.
      rewrite tree_rect_fix.
      f_equal.
      rewrite <- list_In_t_map_t.
      apply list_In_t_map_ext.
      clear x.
      intros x Hx; simpl.
      rewrite map_t_prop; auto.
      clear x ll.
      intros a ll u v Hfg.
      f_equal.
      apply list_In_t_map_ext; auto.
    Qed.
  
  End tree_recursion.

  Definition tree_size (t : tree) : nat.
  Proof.
    induction t as [ x _ ll ] using tree_recursion.
    apply (S (lsum ll)).
  Defined.

  Fact tree_size_fix x ll : tree_size (in_tree x ll) = S (lsum (map tree_size ll)).
  Proof.
    apply tree_recursion_fix.    
  Qed.

  (* this is the subtree relation, aka inclusion *)   

  Reserved Notation "x '<st' y" (at level 70, no associativity).

  Inductive sub_tree : tree -> tree -> Prop :=
    | in_subtree_0 : forall t, t <st t
    | in_subtree_1 : forall s t x ll, In t ll -> s <st t -> s <st in_tree x ll 
  where "x <st y" := (sub_tree x y).

  (* this is the strict subtree relation *)

  Reserved Notation "x '<<st' y" (at level 70, no associativity).

  Inductive ssub_tree : tree -> tree -> Prop :=
    | in_ssubtree_0 : forall s   x ll, In s ll -> s <<st in_tree x ll
    | in_ssubtree_1 : forall s t x ll, In t ll -> s <<st t -> s <<st in_tree x ll 
  where "x <<st y" := (ssub_tree x y).

  (* the immediate subtree relation *)

  Definition imsub_tree s t := exists x ll, In s ll /\ t = in_tree x ll.
  
  Infix "<ist" := imsub_tree (at level 70).

  Fact imsub_tree_wf : well_founded imsub_tree.
  Proof.
    intros t; induction t as [ x ll IH ].
    constructor 1.
    intros y (z & mm & H1 & H2).
    injection H2; intros; subst.
    apply IH; auto.
  Qed.

  Fact in_subtree_0' s t : s = t -> s <st t.
  Proof. intro; subst; constructor. Qed.
  
  Fact imsub_ssub_tree s t : s <ist t -> s <<st t.
  Proof.
    intros (x & ll & H1 & H2); subst.
    constructor 1; auto.
  Qed.
  
  Fact ssub_sub_tree s t : s <<st t -> s <st t.
  Proof.
    induction 1 as [ s | s t ].
    constructor 2 with s; auto.
    constructor 1.
    constructor 2 with t; auto.
  Qed.
  
  Fact ssub_tree_trans r s t : r <<st s -> s <<st t -> r <<st t.
  Proof.
    intro; induction 1 as [ s | ? t ]; auto.
    constructor 2 with s; auto.
    constructor 2 with t; auto.
  Qed.

  Fact ssub_imsub_tree : ssub_tree ~eq2 clos_trans _ imsub_tree.
  Proof.
    split; intros s t.

    induction 1 as [ s x ll | s t x ll H1 H2 H3 ].
    constructor 1.
    exists x, ll; auto.
    constructor 2 with (1 := H3).
    constructor 1.
    exists x, ll; auto.
    
    induction 1 as [ | ? s ].
    apply imsub_ssub_tree; auto.
    apply ssub_tree_trans with s; auto.
  Qed.
  
  Fact ssub_tree_wf : well_founded ssub_tree.
  Proof.
    generalize imsub_tree_wf; intros H.
    apply wf_clos_trans in H.
    revert H; apply wf_incl.
    intros ? ?; apply ssub_imsub_tree.
  Qed.

  Fact sub_tree_refl s : s <st s.
  Proof.
    constructor 1.
  Qed.

  Fact sub_tree_trans r s t : r <st s -> s <st t -> r <st t.
  Proof.
    intro; induction 1 as [ | ? t ]; auto.
    constructor 2 with t; auto.
  Qed.

  Fact ssub_tree_inv_left s t : s <<st t <-> exists s', s <ist s' /\ s' <st t.
  Proof.
    split.
    
    induction 1 as [ s x ll | s t x ll H1 H2 (s' & H3 & H4) ].
    exists (in_tree x ll); split.
    exists x, ll; auto.
    constructor 1.
    exists s'; split; auto.
    constructor 2 with t; auto.
    
    intros (s' & H1 & H2).
    revert s H1.
    induction H2 as [ s' | s' t x ll H1 H2 H3 ].
    intro; apply imsub_ssub_tree.
    intros s Hs.
    constructor 2 with t; auto.
  Qed.

  Fact ssub_tree_inv_right s t : s <<st t <-> exists s', s <st s' /\ s' <ist t.
  Proof.
    split.

    intros H.
    apply ssub_imsub_tree in H.
    apply clos_trans_tn1_iff in H.
    induction H as [ | t u H1 H2 (s' & H3 & H4) ].
    exists s; split; auto; constructor.
    exists t; split; auto.
    apply sub_tree_trans with (1 := H3).
    apply ssub_sub_tree, imsub_ssub_tree, H4.
    
    intros (s' & H1 & x & ll & H2 & H3); subst.
    revert x ll H2.
    induction H1 as [ s x ll | s t x ll H1 H2 IH ]; intros y ll' H3.
    constructor 1; auto.
    constructor 2 with (in_tree x ll); auto.
  Qed.
  
  Fact ssub_tree_inv s x ll : s <<st in_tree x ll <-> exists t, s <st t /\ In t ll.
  Proof.
    rewrite ssub_tree_inv_right.
    split.
    intros (t & H1 & a & mm & H2 & H3).
    injection H3; clear H3; intros; subst.
    exists t; auto.
    intros (t & H1 & H2).
    exists t; split; auto.
    exists x, ll; auto.
  Qed.

  Fact sub_ssub_tree_trans r s t : r <st s -> s <<st t -> r <<st t.
  Proof.
    intros H1 H2.
    rewrite ssub_tree_inv_right in H2.
    destruct H2 as (s' & H2 & H3).
    rewrite ssub_tree_inv_right.
    exists s'; split; auto.
    apply sub_tree_trans with (1 := H1); auto.
  Qed.
  
  Fact ssub_sub_tree_trans r s t : r <<st s -> s <st t -> r <<st t.
  Proof.
    intros H1 H2.
    rewrite ssub_tree_inv_left in H1.
    destruct H1 as (s' & H1 & H3).
    rewrite ssub_tree_inv_left.
    exists s'; split; auto.
    apply sub_tree_trans with (1 := H3); auto.
  Qed.

  Fact sub_tree_inv s t : s <st t <-> s = t \/ s <<st t.
  Proof.
    split.

    induction 1 as [ | s t x ll H1 H2 [ H3 | H3 ]]; subst; auto; right.
    constructor 1; auto.
    constructor 2 with t; auto.
    
    intros [ H | H ].
    subst; constructor.
    apply ssub_sub_tree; auto.
  Qed.

  Fact sub_in_tree_inv s x ll : s <st in_tree x ll -> s = in_tree x ll \/ exists t, s <st t /\ In t ll.
  Proof.
    intros H.
    apply sub_tree_inv in H.
    destruct H as [ H | H ]; auto; right.
    apply ssub_tree_inv_right in H.
    destruct H as (t & ? & r & mm & ? & ?).
    injection H1; clear H1; intros; subst.
    exists t; auto.
  Qed.

  Definition tree_leaf t : { x | in_tree x nil <st t }.
  Proof.
    induction t as [ x [ | t ll ] IH ] using tree_rect.
    
    exists x; constructor 1.
    destruct (IH t) as (y & Hy).
    left; auto.
    exists y; constructor 2 with t; simpl; auto.
  Qed.

  Section sub_trees.

    Definition sub_trees : tree -> list tree.
    Proof.
      induction 1 as [ a ll IH ] using tree_recursion.
      apply cons.
      apply (in_tree a ll).
      apply list_flatten, IH.
    Defined.

    Fact sub_trees_fix x ll : sub_trees (in_tree x ll) = in_tree x ll::list_flatten (map sub_trees ll).
    Proof.
      apply tree_recursion_fix.
    Qed.
    
    Fact sub_trees_sub_tree_eq t x : In x (sub_trees t) <-> x <st t.
    Proof.
      split.

      revert x; induction t as [ y ll IH ]; intros x Hx.
      rewrite sub_trees_fix in Hx.
      destruct Hx as [ Hx | Hx ].
      subst x; constructor 1.
      rewrite list_flatten_spec in Hx.
      destruct Hx as (mm & H1 & H2).
      apply in_map_iff in H2.
      destruct H2 as (t & H2 & H3); subst mm.
      constructor 2 with t; auto.
      
      induction 1 as [ [ x ll ] | s t x ll H1 H2 H3 ]; rewrite sub_trees_fix.
      left; auto.
      right.
      apply list_flatten_spec.
      exists (sub_trees t); split; auto.
      apply in_map_iff.
      exists t; auto.
    Qed.

  End sub_trees.

  (* finite quantification over the nodes of trees *)

  Section tree_fall_exst.

    Variable P : X -> list tree -> Prop.

    Definition tree_fall (t : tree) : Prop.
    Proof.
      induction t as [ a ll IH ] using tree_recursion.
      apply (and (P a ll) (fold_right and True IH)).
    Defined.

    Let conj_eq_prop (A B B' : Prop) : (B <-> B') -> (A /\ B <-> A /\ B').
    Proof. tauto. Qed.

    (* this is how we would like tree_fall to be recursively defined but this would
       not be well-formed in Coq
    *)

    Fact tree_fall_fix x ll : tree_fall (in_tree x ll) <-> P x ll /\ forall t, In t ll -> tree_fall t. 
    Proof.
      unfold tree_fall at 1.
      rewrite tree_recursion_fix.
      apply conj_eq_prop.
      rewrite fold_right_conj_prop.
      split.
      intros H t Ht; apply H; rewrite in_map_iff; exists t; auto.
      intros H y; rewrite in_map_iff; intros (t & H1 & H2); subst; apply H; auto.
    Qed.

    Fact tree_fall_sub_tree t : tree_fall t <-> forall x ll, in_tree x ll <st t -> P x ll.
    Proof.
      split.

      induction t as [ x ll IH ]; intros Ht.
      rewrite tree_fall_fix in Ht.
      destruct Ht as [ Hx Hll ].
      intros y mm H1.
      apply sub_in_tree_inv in H1.
      destruct H1 as [ H1 | (t & H1 & H2) ].
      injection H1; clear H1; intros; subst; auto.
      apply IH with t; auto.
      
      induction t as [ x ll IH ]; intros Ht.
      rewrite tree_fall_fix; split.
      apply Ht; constructor 1.
      intros u Hu; apply IH; auto.
      intros y mm Hmm; apply Ht.
      constructor 2 with u; auto.
    Qed.

    Definition tree_exst (t : tree) : Prop.
    Proof.
      induction t as [ a ll IH ] using tree_recursion.
      apply (or (P a ll) (fold_right or False IH)).
    Defined.

    Let disj_eq_prop (A B B' : Prop) : (B <-> B') -> (A \/ B <-> A \/ B').
    Proof. tauto. Qed.

    (* this is how we would like tree_exst to be recursively defined but this would
       not be well-formed in Coq
    *)

    Fact tree_exst_fix x ll : tree_exst (in_tree x ll) <-> P x ll \/ exists t, In t ll /\ tree_exst t. 
    Proof.
      unfold tree_exst at 1.
      rewrite tree_recursion_fix.
      apply disj_eq_prop.
      rewrite fold_right_disj_prop.
      split.
      intros (y & Hy1 & Hy2); rewrite in_map_iff in Hy1. 
      destruct Hy1 as (t & Hy1 & Hy3); subst; exists t; auto.
      intros (t & Ht1 & Ht2); exists (tree_exst t); rewrite in_map_iff; 
        split; auto; exists t; auto.
    Qed.

    Fact tree_exst_sub_tree t : tree_exst t <-> exists x ll, in_tree x ll <st t /\ P x ll.
    Proof.
      split.

      induction t as [ x ll IH ]; intros Ht.
      rewrite tree_exst_fix in Ht.
      destruct Ht as [ Ht | (t & H1 & H2) ].
      exists x, ll; split; auto; constructor.
      destruct (IH _ H1 H2) as (y & mm & H3 & H4).
      exists y, mm; split; auto.
      constructor 2 with t; auto.
      
      intros (x & ll & H1 & H2); revert x ll H1 H2.
      induction t as [ x ll IH ]; intros y mm H1 H2.
      rewrite tree_exst_fix.
      apply sub_in_tree_inv in H1.
      destruct H1 as [ H1 | (t & H1 & H3) ].
      left; injection H1; intros; subst; auto.
      right; exists t; split; auto.
      apply IH with (2 := H1); auto.
    Qed.

    Fact sub_tree_fall t1 t2 : t1 <st t2 -> tree_fall t2 -> tree_fall t1.
    Proof.
      induction 1; auto.
      rewrite tree_fall_fix.
      intros (? & ?); auto.
    Qed.

    Fact sub_tree_exst t1 t2 : t1 <st t2 -> tree_exst t1 -> tree_exst t2.
    Proof.
      induction 1 as [ | ? t ]; auto.
      rewrite tree_exst_fix.
      right; exists t; auto.
    Qed.

  End tree_fall_exst.

  Fact tree_fall_inc (P Q : _ -> _ -> Prop) : P inc2 Q -> tree_fall P inc1 tree_fall Q.
  Proof.
    intros H t; induction t as [ x ll IH ].
    repeat rewrite tree_fall_fix.
    intros []; split; auto.
  Qed.

  Fact tree_exst_inc (P Q : _ -> _ -> Prop) : P inc2 Q -> tree_exst P inc1 tree_exst Q.
  Proof.
    intros H t; induction t as [ x ll IH ].
    repeat rewrite tree_exst_fix.
    intros [| (t & ? & ?)]; [ left | right ]; auto; exists t; auto.
  Qed. 

  Section tree_fall_exst_dec.

    Variable (P Q : X -> list tree -> Prop).

    Hypothesis PQ_incomp : forall x ll, P x ll -> Q x ll -> False.
    
    Fact tree_fall_exst_incomp t : tree_fall P t -> tree_exst Q t -> False.
    Proof.
      induction t as [ x ll IH ].
      rewrite tree_fall_fix, tree_exst_fix.
      intros [ H1 H2 ] [ H3 | (t & H3 & H4) ].
      
      apply PQ_incomp with (1 := H1); auto.
      apply IH with (1 := H3); auto.
    Qed.      

    Hypothesis PQ_dec : forall x ll, { P x ll } + { Q x ll }.    
    
    Fact tree_fall_exst_dec t : { tree_fall P t } + { tree_exst Q t }.
    Proof.
      induction t as [ x ll IH ].
      destruct (list_dec_rect (tree_exst Q) (tree_fall P) ll) as [ (t & H1 & H2) | H1 ].
      intros z Hz; specialize (IH _ Hz); tauto.
      
      right.
      apply In_t_In in H1.
      apply tree_exst_fix.
      right; exists t; auto.
      
      rewrite In_In_t in H1.
      destruct (PQ_dec x ll) as [ | H2 ].
      
      left; apply tree_fall_fix; auto.
      
      right.
      apply tree_exst_fix.
      left; auto.
    Qed.
    
  End tree_fall_exst_dec.

  Section tree_fall_dec.
  
    Variable (P : X -> list tree -> Prop).

    Hypothesis PQ_dec : forall x ll, { P x ll } + { ~ P x ll }.
    
    Fact tree_fall_dec t : { tree_fall P t } + { ~ tree_fall P t }.
    Proof.
      destruct (tree_fall_exst_dec _ _ PQ_dec t) as [ | C ].
      tauto.
      right; intros H.
      apply tree_fall_exst_incomp with (2 := H) (3 := C).
      intros; tauto.
    Qed.
    
    Fact tree_exst_dec t : { tree_exst P t } + { ~ tree_exst P t }.
    Proof.
      destruct (tree_fall_exst_dec (fun x ll => ~ P x ll) P) with (t := t) as [ C | ].
      intros x ll; specialize (PQ_dec x ll); tauto.
      right; intros H.
      apply tree_fall_exst_incomp with (2 := C) (3 := H).
      intros; tauto.
      left; auto.
    Qed.

  End tree_fall_dec.

End trees.

Infix "<st" := (@sub_tree _) (at level 70, no associativity).
Infix "<<st" := (@ssub_tree _) (at level 70, no associativity).
Infix "<ist" := (@imsub_tree _) (at level 70, no associativity).

Section tree_map.

  Variables (X Y : Type) (f : X -> Y).
  
  Definition tree_map : tree X -> tree Y.
  Proof.
    induction 1 as [ x ts IH ] using tree_recursion.
    apply (in_tree (f x)), IH.
  Defined.

  Fact tree_map_fix x ts : tree_map (in_tree x ts) = in_tree (f x) (map tree_map ts).
  Proof.
    apply tree_recursion_fix.
  Qed.

End tree_map.

Section wfptree.

  (* well-formed trees given a predicate that allows only a subset P n of nodes
     for the arity n *)

  Variable (X : Type) (P : nat -> X -> Prop).

  Definition wfptree := tree_fall (fun x ll => P (length ll) x).
  
  Fact wfptree_fix x ll : wfptree (in_tree x ll) <-> P (length ll) x /\ forall t, In t ll -> wfptree t.
  Proof.
    apply tree_fall_fix.
  Qed.

  Section wfptree_rect.
  
    Variable (Q : tree X -> Type).
  
    Hypothesis HQ : forall x ll, wfptree (in_tree x ll) -> (forall t, In_t t ll -> Q t) -> Q (in_tree x ll).
  
   Theorem wfptree_rect t : wfptree t -> Q t.
    Proof.
      induction t as [ x ll IH ]; intros H.
      apply HQ; auto.
      rewrite wfptree_fix in H; destruct H as (H1 & H2).
      intros t Ht; apply IH; auto; apply H2, In_t_In; auto.
    Qed.

  End wfptree_rect.
  
  Section wfptree_ind.
  
    Variable (Q : tree X -> Prop).
  
    Hypothesis HQ : forall x ll, wfptree (in_tree x ll) -> (forall t, In t ll -> Q t) -> Q (in_tree x ll).
  
    Theorem wfptree_ind : forall t, wfptree t -> Q t.
    Proof.
      apply wfptree_rect.
      intros x ll H1 H2.
      rewrite In_In_t in H2; auto.
    Qed.
 
  End wfptree_ind.

  Hypothesis Pdec : forall n x, { P n x } + { ~ P n x }.

  Definition wfptree_dec t : { wfptree t } + { ~ wfptree t }.
  Proof.
    apply tree_fall_dec.
    intros ? ?; apply Pdec.
  Qed.  

End wfptree.

Section tree_with_signature.

  (* each symbol in X can be used exactly one given arity *)

  Variable (X : Type) (ar : X -> nat).

  Definition stree := tree_fall (fun x ll => ar x = length ll).
  
  Fact stree_fix x ll : stree (in_tree x ll) <-> ar x = length ll /\ forall t, In t ll -> stree t.
  Proof.
    apply tree_fall_fix.
  Qed.
  
  Fact stree_dec t : { stree t } + { ~ stree t }.
  Proof.
    apply tree_fall_dec.
    intros; apply eq_nat_dec.
  Defined.

End tree_with_signature.

Section bounded_breadth_tree.

  (* trees of bounded breadth *)
  
  Variable X : Type. 

  Definition btree k := @tree_fall X (fun x ll => length ll < k).

  Fact btree_fix k x ll : btree k (in_tree x ll) <-> length ll < k /\ forall t, In t ll -> btree k t.
  Proof.
    apply tree_fall_fix.
  Qed.

  Definition btree_dec k t : { btree k t } + { ~ btree k t }.
  Proof.
    apply tree_fall_dec.
    intros; apply lt_dec.
  Qed.

  Fact btree_inc n m : n <= m -> btree n inc1 btree m.
  Proof.
    intros H t; induction t as [ ? ? IH ].
    do 2 rewrite btree_fix.
    intros []; split; auto; omega.
  Qed.

  Definition bound_tree (t : tree X) : nat.
  Proof.
    induction t as [ t ll IH ] using tree_recursion.
    apply (length ll + lsum IH).
  Defined. 

  Fact bound_tree_fix x ll : bound_tree (in_tree x ll) = length ll + lsum (map bound_tree ll).
  Proof.
    apply tree_recursion_fix.
  Qed.
 
  (* we can compute a bound on the breadth of every tree *)

  Fact bound_tree_prop t : btree (S (bound_tree t)) t.
  Proof.
    induction t as [ x ll IH].
    rewrite btree_fix, bound_tree_fix; split.
    omega.
    intros t Ht; generalize (IH t Ht).
    apply btree_inc.
    apply in_map with (f := bound_tree) in Ht.
    apply lsum_prop in Ht; omega.
  Qed.

End bounded_breadth_tree. 


