(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import Wf.
Require Import Relations.

Require Omega.

Require Import tacs.

Require Import reverse.
Require Import rel_utils.

Set Implicit Arguments.

(*

   Recall the well_founded predicate definition (with implicit 
   arguments):

   Definition well_founded R := 
          forall Q, (forall a, (forall b, R b a -> Q b) -> Q a) 
                  -> forall a, Q a.

   Hence, while proving Q a, you can assume Q b for any b such 
   that R b a.

   Let f be a map (that potentially "forgets" some information).
   Basically, with a "well-founded upto f" relation R you can prove
   a predicate Q by induction using R provided that Q expresses no 
   property about the information that is forgotten.

   Definition well_founded_upto f R := 
          forall Q, (forall a, (forall b, R b a -> Q (f b)) -> Q (f a)) 
                  -> forall a, Q (f a).

   But in that case, R is not necessarily well-founded (it can even
   contain x such that R x x !!) 

   With a "well-founded upto" relation, you can use "hidden" information
   in the inductive steps of the proofs. But you cannot prove any predicate
   Q that way: 
      1/ Q must be extensionnal (not really a constraint)
      2/ Q cannot expresses properties about the "hidden" information

   Most importantly, the well-founded upto predicate is stable under 
   lexicographic products.

   That concept solved very nicely the problem of finding an outermost 
   induction principle suitable for the logical/prop-bounded version of 
   both Higman's theorem and Kruskal's theorem.

*)

Section well_founded_upto.

  Variable (X Y : Type).
  
  Implicit Type (f : X -> Y) (P : X -> Prop) (R S : relation X) (Q : Y -> Prop).

  (* Thank you again Napoleon !!! Napoleon is a horse and it is the
     second time I have a bright idea while being with him.
     The first was to use a "universe" to prove Higman & Kruskal
     theorems
  *)

  Definition well_founded_upto f P R := 
          forall Q, (forall a, P a -> (forall b, P b -> R b a -> Q (f b)) -> Q (f a)) 
                  -> forall a, P a -> Q (f a).

  Fact wf_upto_incl f P R S : R inc2 S 
                           -> well_founded_upto f P S 
                           -> well_founded_upto f P R.
  Proof.
     intros H1 H2 Q IH a.
     apply H2; auto.
  Qed.
   
   Fact wf_upto_ext f g P R : (forall x, f x = g x) -> well_founded_upto f P R -> well_founded_upto g P R.
   Proof.
     intros H1 H2 Q IH.
     intros a; rewrite <- H1; revert a.     
     apply H2.
     intros a H3 H4; rewrite H1.
     apply IH; auto.
     intros ?; rewrite <- H1; auto.
   Qed.

   Definition well_founded_upto_ext (ext : _ -> Prop) f P R := forall Q, ext Q 
                                                 -> (forall a, P a -> (forall b, P b -> R b a -> Q (f b)) -> Q (f a)) 
                                                 ->  forall a, P a -> Q (f a).
                                                 
   Fact wf_upto_ext_incl ext f P R S : R inc2 S 
                                    -> well_founded_upto_ext ext f P S 
                                    -> well_founded_upto_ext ext f P R.
   Proof.
     intros H1 H2 Q Qext IH a.
     apply H2; auto.
  Qed.
   
End well_founded_upto.

Fact wf_upto_wf X (R : relation X) : well_founded_upto (fun x => x) (fun _ => True) R <-> well_founded R.
Proof.
  split.
  intros H a.
  apply H; constructor; auto.
  intros H Q H1 a _.
  apply (well_founded_ind H); auto.
Qed.

Section wf_upto_map.

  Variable (X Y X' Y' : Type) (hX : X -> X') (hY : Y -> Y') 
           (f  : X -> Y)   (P : X -> Prop)   (R : relation X) 
           (f' : X' -> Y') (P' : X' -> Prop) (R' : relation X').
  
  Hypothesis Hh : forall x, f' (hX x) = hY (f x).
  Hypothesis Hh' : forall x' y, f' x' = hY y -> exists x, x' = hX x /\ y = f x.
  Hypothesis HP : forall x, P' (hX x) <-> P x.
  Hypothesis HRR' : forall x y, R x y -> R' (hX x) (hX y).
  
  Fact wf_upto_map : well_founded_upto f' P' R' -> well_founded_upto f P R.
  Proof.
    intros H Q IHQ.    
    set (Q' y' := forall y, y' = hY y -> Q y).
    assert (forall x', P' x' -> Q' (f' x')) as H'.
      apply (H Q').
      intros a Ha IHa.
      unfold Q'.
      intros y Hy.
      destruct (Hh' Hy) as (x & H1 & H2).
      subst y.
      apply IHQ.
      apply HP; subst a; auto.
      intros b Hb1 Hb2.
      apply (IHa (hX b)); auto.
      apply HP; auto.
      subst a; apply HRR'; auto.
    intros a Ha.
    apply (H' (hX a)); auto.
    apply HP; auto.
  Qed.

End wf_upto_map.

Section wf_upto_lex_dep_prod.

  Variables (X1 Y1 X2 Y2 : Type) (f1 : X1 -> Y1) (f2 : X2 -> Y2) 
            (P1 : X1 -> Prop) (R1 : relation X1) 
            (P2 : X2 -> Prop) (R2 : X1 -> X2 -> X2 -> Prop).
  
  Definition lex_dep_prod c1 c2 :=
    match c1   , c2 with
      | (a1,b1), (a2,b2) => R1 a1 a2 \/ a1 = a2 /\ R2 a1 b1 b2
    end.    

  Fact wf_upto_lex_dep_prod : 
                     well_founded_upto f1 P1 R1
                  -> (forall x, well_founded_upto f2 P2 (R2 x))
                  -> well_founded_upto (fun c => (f1 (fst c), f2 (snd c))) 
                                       (fun c => P1 (fst c) /\ P2 (snd c))
                                       lex_dep_prod.
  Proof.
    intros H1 H2.
    intros Q HQ (a,b) (H3 & H4); simpl in H3, H4. 
    revert a H3 b H4.

    specialize (H1 (fun a => forall b, P2 b -> Q (a,f2 b))); simpl in H1 |- *.
    intros a Ha b; apply H1; auto; clear a Ha b H1.
    
    intros a Ha IHa.
    
    specialize (H2 a (fun b => Q (f1 a,b))); simpl in H2.
    intros b; apply H2; clear b H2.
    
    intros b Hb IHb.
    
    apply HQ with (a := (a,b)); auto.
    intros (a',b') (Ha' & Hb') [ H | (? & H) ]; simpl.

    apply IHa; auto.
    
    subst a'.
    apply IHb; auto.
  Qed.

End wf_upto_lex_dep_prod.

Section wf_upto_lex_dep_prod_ext'.

  Variables (X1 Y1 X2 Y2 : Type) (f1 : X1 -> Y1) (f2 : X2 -> Y2) 
            (P1 : X1 -> Prop) (R1 : relation X1) 
            (P2 : Y1 -> X2 -> Prop) (R2 : X1 -> X2 -> X2 -> Prop).

  Fact wf_upto_lex_dep_prod_ext' ext : 
                     well_founded_upto f1 P1 R1
                  -> (forall x, well_founded_upto_ext ext f2 (P2 (f1 x)) (R2 x))
                  -> well_founded_upto_ext (fun Q => forall x1, ext (fun x2 => Q (x1,x2))) 
                                       (fun c => (f1 (fst c), f2 (snd c))) 
                                       (fun c => P1 (fst c) /\ P2 (f1 (fst c)) (snd c))
                                       (lex_dep_prod R1 R2).
  Proof.
    intros H1 H2.
    intros Q Qext HQ (a,b) (H3 & H4); simpl in H3, H4 |- *. 
    revert a H3 b H4.

    specialize (H1 (fun a => forall b, P2 a b -> Q (a,f2 b))); simpl in H1 |- *.
    intros a Ha b; apply H1; auto; clear a Ha b H1.
    
    intros a Ha IHa.
    
    specialize (H2 a (fun b => Q (f1 a,b))); simpl in H2.
    intros b; apply H2; clear b H2.
    
    apply Qext.
    
    intros b Hb IHb.
    
    apply HQ with (a := (a,b)); auto.
    intros (a',b') (Ha' & Hb') [ H | (? & H) ]; simpl.

    apply IHa; auto.
    
    subst a'.
    apply IHb; auto.
  Qed.

End wf_upto_lex_dep_prod_ext'.

Section wf_upto_lex_dep_prod_ext.

  Variables (X1 Y1 X2 Y2 : Type) (f1 : X1 -> Y1) (f2 : X2 -> Y2) 
            (P1 : X1 -> Prop) (R1 : relation X1) 
            (P2 : X2 -> Prop) (R2 : X1 -> X2 -> X2 -> Prop).

  Fact wf_upto_lex_dep_prod_ext ext : 
                     well_founded_upto f1 P1 R1
                  -> (forall x, well_founded_upto_ext ext f2 P2 (R2 x))
                  -> well_founded_upto_ext (fun Q => forall x1, ext (fun x2 => Q (x1,x2))) 
                                       (fun c => (f1 (fst c), f2 (snd c))) 
                                       (fun c => P1 (fst c) /\ P2 (snd c))
                                       (lex_dep_prod R1 R2).
  Proof.
    intros H1 H2.
    intros Q Qext HQ (a,b) (H3 & H4); simpl in H3, H4 |- *. 
    revert a H3 b H4.

    specialize (H1 (fun a => forall b, P2 b -> Q (a,f2 b))); simpl in H1 |- *.
    intros a Ha b; apply H1; auto; clear a Ha b H1.
    
    intros a Ha IHa.
    
    specialize (H2 a (fun b => Q (f1 a,b))); simpl in H2.
    intros b; apply H2; clear b H2.
    
    apply Qext.
    
    intros b Hb IHb.
    
    apply HQ with (a := (a,b)); auto.
    intros (a',b') (Ha' & Hb') [ H | (? & H) ]; simpl.

    apply IHa; auto.
    
    subst a'.
    apply IHb; auto.
  Qed.

End wf_upto_lex_dep_prod_ext.

Definition well_founded_ext X (ext : _ -> Prop) R := 
  forall Q : X -> Prop, ext Q 
                    -> (forall a, (forall b, R b a -> Q b) -> Q a) 
                    -> forall a, Q a.

Section wf_lex_seq.

  Variable (X Y : Type).

  Implicit Type (R : nat -> relation X).

  Definition lex_nat_seq_rev k R s1 s2 := 
      exists i, i < k  
           /\  (forall j, j < i -> s1 j = s2 j)
           /\  R i (s1 i) (s2 i).


  Let cons U (x : U) s i := match i with 0 => x | S i => s i end.
  Let tail U (s : nat -> U) i := s (S i).

  Let cons_tail_eq U (s : _ -> U) i : cons (s 0) (tail s) i = s i.
  Proof.
    destruct i; cbv; auto.
  Qed.

  Fact wf_lex_nat_seq_rev k R : (forall i, well_founded (R i)) 
                          -> well_founded_ext ext (lex_nat_seq_rev k R). 
  Proof.
    revert R; induction k as [ | k IHk ]; intros R HR.
    
    intros P Pext HP a; apply HP.
    intros ? (i & C & _); omega.
    
    intros P Pext HP.
    assert (forall x s, P (cons x s)) as H.
      induction x as [ x Hx ] using (well_founded_ind (HR 0)).      
      apply (IHk (tail R)) with (Q := fun s => P (cons x s)).
      unfold tail; auto.
      unfold cons; intros ? ? ?; apply Pext.
      intros [|]; auto.
      intros a Ha.
      apply HP.
      intros b ([|i] & H1 & H2 & H3).
      apply Pext with (s := cons (b 0) (tail b)).
      intro; apply cons_tail_eq.
      apply Hx; auto.
      apply Pext with (s := cons x (tail b)).
      intros [|]; auto.
      rewrite H2; auto; omega.
      apply Ha.
      exists i.
      repeat split; auto.
      omega.
      unfold tail; intros ? ?; rewrite H2; auto; omega.
    
    intros a.
    apply Pext with (s := cons (a 0) (tail a)); auto.
  Qed.

  Variable (f : X -> Y).

  Fact wf_upto_lex_nat_seq_rev k P R : 
          (forall i, well_founded_upto f (P i) (R i)) 
        -> well_founded_upto_ext ext (fun s i => f (s i)) 
                                     (fun s => forall i, P i (s i))
                                     (lex_nat_seq_rev k R).
  Proof.
    revert P R; induction k as [ | k IHk ]; intros P R HR.
    
    intros Q Qext HQ a Ha; apply HQ; auto.
    intros ? ? (i & C & _); omega.
    
    intros Q Qext HQ.
    assert (forall x, P 0 x -> forall s, (forall i, P (S i) (s i)) 
                                        -> Q (cons (f x) (fun i => f (s i)))) as H.
      set (Q' y := forall s, (forall i, P (S i) (s i)) -> Q (cons y (fun i => f (s i)))).
      apply (HR 0 (fun y => forall s, (forall i, P (S i) (s i)) 
                                   -> Q (cons y (fun i => f (s i))))); intros x Hx IHx.      
      apply (IHk (tail P) (tail R)) with (Q := fun s => Q (cons (f x) s)).
      unfold tail; auto.
      unfold cons; intros ? ? ?; apply Qext.
      intros [|]; auto.
      intros a Ha IHa.
      apply Qext with (s := fun i => f (cons x a i)).
      intros [|]; auto.
      apply HQ.
      intros [|]; auto.
      intros b Hb ([|i] & H1 & H2 & H3).

      apply Qext with (s := cons (f (b 0)) (fun i => f (tail b i))).
      intros [|]; simpl; auto.
      apply IHx; auto.
      unfold tail; auto.
      
      apply Qext with (s := cons (f x) (fun i => f (tail b i))).
      intros [|]; auto.
      rewrite H2; auto; omega.
      apply IHa.
      unfold tail; auto.
      exists i.
      repeat split; auto.
      omega.
      unfold tail; intros ? ?; rewrite H2; auto; omega.
    
    intros a Ha.
    apply Qext with (s := cons (f (a 0)) (fun i => f (tail a i))); auto.
    intros [|]; simpl; auto.
    apply H; unfold tail; auto.
  Qed.

End wf_lex_seq.

Section wf_upto_lex_seq.

  Variables (X Y : Type) (k : nat) (f : X -> Y) (P : nat -> X -> Prop) 
            (R : nat -> relation X) (HR : forall i, well_founded_upto f (P i) (R i)).

  Definition lex_nat_seq s1 s2 := 
      exists i, i < k  
           /\  (forall j, i < j -> s1 j = s2 j)
           /\  R i (s1 i) (s2 i).
 
  Theorem wf_upto_lex_nat_seq : well_founded_upto_ext ext (fun s i => f (s i)) 
                                                          (fun s => forall i, P i (s i))
                                                          lex_nat_seq. 
  Proof.
    intros Q Qext HQ.
    intros a Ha.
    apply Qext with (s := fun i => f (rev k (rev k a) i)).
    intros; rewrite rev_inv; auto.
    assert (Ha' : forall i, P i (rev k (rev k a) i)).
      intro; rewrite rev_inv; auto.
    revert Ha'.    
    generalize (rev k a); clear a Ha.

    generalize (@wf_upto_lex_nat_seq_rev _ _ f k (rev k P) (rev k R)); intros H.
    spec_all H.
    intros; apply HR.
    red in H.
    intros ? ?; apply H with (Q := fun s => Q (rev k s)); clear H.
    
    1: intros ? ? ?; apply Qext; unfold rev; auto.
    2: unfold rev; intro i; rewrite <- (revk_inv k i) at 2; apply Ha'.   
    
    intros a Ha IHa.
    apply HQ.
    intros i; rewrite <- (revk_inv k i) at 1; apply Ha.    
    intros b Hb (i & H1 & H2 & H3).
    apply Qext with (s := rev k (fun i => f (rev k b i))).
    intros; unfold rev; simpl; rewrite revk_inv; auto.
    apply IHa.
    intro; unfold rev; auto.
    exists (revk k i).
    split; auto.
    apply revk_prop; auto.
    split.
    intros j Hj; unfold rev; rewrite H2.
    rewrite revk_inv; auto.
    revert Hj; unfold revk.
    destruct (le_lt_dec k i);
    destruct (le_lt_dec k j); intros; omega.
    unfold rev; rewrite revk_inv; auto.
  Qed.

End wf_upto_lex_seq.
