(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREER SOFTWARE LICENSE AGREEMENT         *)
(**************************************************************)

Require Import Arith.
Require Import List.
Require Import Relations.
Require Import Wf.
Require Import Wellfounded.

Require Import tacs.

Require Import rel_utils.
Require Import list_utils.

Require Import wf_utils.
Require Import lex.

Require Import good_base.

Set Implicit Arguments.

Section WFT.

  (* Well-founded trees of arity X *)

  (* The initial ideas of WFT come from "Stop when you are almost full" 

     The notion of stump is a generalization to Types from the
     one found in Wim Veldman's paper.
  *)

  Variables (X : Type).

  Inductive wft : Type :=
    | in_wft_zt : wft
    | in_wft_sup : (X -> wft) -> wft.

  Fixpoint af_securedby (R : X -> X -> Prop) t :=
    match t with
      | in_wft_zt    => forall x y, R x y
      | in_wft_sup f => forall x, af_securedby (R rlift x) (f x)
    end.

  Definition wft_lt x y := 
    match y with 
      | in_wft_zt    => False
      | in_wft_sup f => exists n, x = f n
    end.

  Fact in_wft_lt f x : wft_lt (f x) (in_wft_sup f).
  Proof.
    exists x; auto.
  Qed.

  Fact wft_lt_wf : well_founded wft_lt.
  Proof.
    intros s; induction s as [ | f IHf ].
    constructor; intros ? [].
    constructor; intros ? (? & ?); subst; auto.
  Qed.

  Fixpoint wft_seq_bound t (f : nat -> X) :=
    match t with
      | in_wft_zt    => 0
      | in_wft_sup g => S (wft_seq_bound (g (f 0)) (fun n => f (S n)))
    end.

  Fact af_sb_good_pfx t R f : af_securedby R t -> good R (pfx_rev f (2+wft_seq_bound t f)).
  Proof.
    revert R f; induction t as [ | h IHh ]; intros R f.
    simpl; intros; constructor 1 with (f 0); simpl; auto.
    intros H; simpl in H.
    simpl wft_seq_bound.
    generalize (IHh _ _ (fun n => f (S n)) (H (f 0))); intros H1.
    apply good_lift_rel in H1.
    replace_with H1.
    rewrite <- pfx_rev_S; auto.
  Qed.

End WFT.

Implicit Arguments in_wft_zt [ X ].

Section owft.

  Variable X : Type.

  (* wft and option wft represents stumps, ie inductivelly defined
     predicates (ie subsets) over list X which are decidable, see
     Wim Veldman page 219--221 *)

  Fixpoint wft_stump (t : wft X) : list X -> Prop :=
    match t with
      | in_wft_zt    => eq nil
      | in_wft_sup f => fun l => 
      match l with 
        | nil  => True
        | x::l => wft_stump (f x) l
      end
    end.

  Fact wft_stump_dec t l : { wft_stump t l } + { ~ wft_stump t l }.
  Proof.
    revert l; induction t as [ | f IH ]; intros [ | x l ].
    left; simpl; auto.
    right; simpl; discriminate.
    left; simpl; auto.
    destruct (IH x l); [ left | right ]; simpl; auto.
  Qed.

  Definition owft := option (wft X).

  Definition owft_stump (t : owft) : list X -> Prop :=
    match t with 
      | None   => fun _ => False
      | Some t => fun l => wft_stump t l
    end.

  Fact owft_stump_dec t l :  { owft_stump t l } + { ~ owft_stump t l }.
  Proof.
    destruct t as [ t | ]; simpl.
    apply wft_stump_dec.
    right; auto.
  Qed.

  Definition owft_lt (t1 t2 : owft) :=
    match t1, t2 with
      | _     , None   => False
      | None  , Some _ => True 
      | Some r, Some s => wft_lt r s
    end.
 
  Fact in_owft_lt0 : owft_lt None (Some in_wft_zt).
  Proof. simpl; auto. Qed.
  
  Fact in_owft_lt1 f x : owft_lt (Some (f x)) (Some (in_wft_sup f)).
  Proof. exists x; auto. Qed.
   
  Fact owft_lt_wf : well_founded owft_lt.
  Proof.
    intros [ t | ].
    revert t; apply well_founded_induction with (1 := @wft_lt_wf _).
    intros r Hr. 
    constructor; intros [ s | ] Hs.
    simpl in Hs; auto.
    constructor; intros [|] [].
    constructor; intros [|] [].
  Qed.

  Section easier.

    (* the "easier" well-founded relation over k-sequences of owft's *)

    Variable k : nat.

    Definition owft_easier := lex_seq owft_lt k.

    Fact owft_easier_eq s t : owft_easier s t 
                          <-> exists i, i < k /\ owft_lt (s i) (t i) /\ forall j, i < j -> s j = t j.
    Proof.
      split; auto.
    Qed.
    
    Fact owft_easier_wf : well_founded owft_easier.
    Proof.
      apply wf_lex_seq, owft_lt_wf.
    Qed.
    
    Section owft_rect.
  
      Variable P : (nat -> owft) -> Type.
      Hypothesis HP : forall t, (forall s, owft_easier s t -> P s) -> P t.
    
      Theorem owft_easier_rect : forall s, P s.
      Proof.
        apply well_founded_induction_type with (1 := owft_easier_wf); auto.
      Qed.

    End owft_rect.
      
  End easier.

  Section more_facile.
   
   (* the "more-facile" well-founded relation over non trivial stumps *)

    Definition nt_stump := { s : nat -> owft & { k : nat | s k <> None /\ forall i, k < i -> s i = None } }.
    
    Definition nts_seq (s : nt_stump) := projT1 s.
    
    Definition nts_char (x : nt_stump) : nat := 
      match x with
        | existT _ (exist k _) => k
      end.
      
    Fact nts_prop1 : forall s, nts_seq s (nts_char s) <> None.
    Proof.
      intros (? & ? & ? & ?); simpl; auto.
    Qed.
  
    Fact nts_prop2 : forall s i, nts_char s < i -> nts_seq s i = None.
    Proof.
      intros (? & ? & ? & ?); simpl; auto.
    Qed.

    (** The definition of page 221 with the following change
     **
     ** the condition nts_char t < nts_char s is NOT NEEDED to get a well founded relation 
     **)

    Definition nts_facile (s t : nt_stump) := 
           (owft_lt (nts_seq s (nts_char s)) (nts_seq t (nts_char t)))            
        \/ (nts_char t = nts_char s 
        /\ exists i, owft_lt (nts_seq s i) (nts_seq t i) 
                 /\ forall j, i < j -> nts_seq s j = nts_seq t j).

    (* we get a well_founded relation *)
  
    Fact wf_nts_facile : well_founded nts_facile.
    Proof.
      set (f s := (nts_char s,nts_seq s (nts_char s))).
      set (RX (c1 c2 : nat * _)  := owft_lt (snd c1) (snd c2)).
      set (RY (c : nat * owft) s t := lex_seq owft_lt (S (fst c)) (nts_seq s) (nts_seq t)).
      
      assert (wf_RX : well_founded RX).
        generalize owft_lt_wf; apply wf_inverse_image.
      
      assert (wf_RY : forall x, well_founded (RY x)).
        intros (x & y); unfold RY; simpl; clear y.
        generalize (wf_lex_seq owft_lt_wf (S x)).
        apply wf_inverse_image.
        
      generalize (wf_lex_map _ f wf_RX wf_RY).
      apply wf_incl.
      
      intros s t [ H2 | (H1 & H2) ].
      left; unfold f, RX; simpl; auto.
      
      destruct H2 as (i & H2 & H3).
      
      destruct (lt_eq_lt_dec i (nts_char s)) as [ [ H4 | H4 ] | H4 ].
      
      right; unfold f, RY; simpl; split.
      f_equal; auto.
      rewrite H1; apply H3; auto.
      exists i; repeat split; auto.
      
      left.
      unfold f, RX; simpl.
      replace_with H2; f_equal; omega.
      
      rewrite nts_prop2 in H2; try omega.
      rewrite nts_prop2 in H2; try omega.
      apply wf_irrefl with (1 := owft_lt_wf) in H2; destruct H2.
    Qed.

    (* we deduce the following induction principle *)

    Section nt_stump_rect.
  
      Variable P : nt_stump -> Type.
      Hypothesis HP : forall t, (forall s, nts_facile s t -> P s) -> P t.
    
      Theorem nt_stump_rect : forall s, P s.
      Proof.
        apply (well_founded_induction_type wf_nts_facile); auto.
      Qed.

    End nt_stump_rect.

  End more_facile.

End owft.

  
  






